/*
communicator.h 0.12.0 (2002-Nov-25-Mon)
Adam M. Costello  
http://www.nicemice.net/amc/

Provides an abstract interface for unreliable
datagrams and reliable byte streams.

This is ANSI C code that refers to POSIX file descriptors.

*/


#ifndef COMMUNICATOR_H
#define COMMUNICATOR_H


#include <stddef.h>

#include "scheduler.h"  /* version 0.6.* */


/*** Model ***/

  /* A communicator is something that can talk to other               */
  /* communicators.  Each communicator can have a number of inboxes,  */
  /* which come in two flavors: message inboxes and connection        */
  /* inboxes.  An unreliable datagram can be sent from any            */
  /* communicator to any message inbox.  A bidirectional reliable     */
  /* byte stream pipe can be set up between any communicator and any  */
  /* connection inbox.  Notice that there is no concept of outbox.    */

  /* Each communicator has a global address, and each inbox has an    */
  /* address relative to its communicator's address.  Therefore, in   */
  /* order to specify an inbox, two addresses are needed, one for the */
  /* communicator and one for the inbox itself.                       */

  /* When this interface is implemented on top of IP, communicator    */
  /* addresses map onto IP addresses, message inbox addresses map     */
  /* onto UDP destination ports, and connection inbox addresses map   */
  /* onto TCP destination ports.  This interface has nothing like     */
  /* source ports.                                                    */


/*** Address format ***/

  /* Every address is a sequence of bytes in which the first byte     */
  /* contains the number of subsequent bytes.  The format and meaning */
  /* of the subsequent bytes are explicitly unspecified.              */


/*** Incomplete types ***/

struct communicator;
    /* The entity that manages communication. */

struct message_inbox;
    /* An endpoint for receiving datagrams. */

struct connection_inbox;
    /* An endpoint for receiving incoming connection requests. */

struct configurator;
    /* See configurator.h. */


/*** Functions for using a communicator ***/

  /* In all functions below, unless otherwise noted, the caller       */
  /* retains ownership of any memory buffers passed as arguments; the */
  /* callee will not free them, and will not continue to access them  */
  /* after the function has returned.                                 */


struct communicator *communicator_new(
  const unsigned char *, struct scheduler *, struct configurator * );

    /* communicator_new(address,scheduler,configurator) returns   */
    /* a pointer to a new communicator with the given address,    */
    /* or a null pointer if one cannot be created or if the       */
    /* address is unavailable.  The address may be a null pointer */
    /* if no particular address is needed (one will be chosen     */
    /* automatically).  The communicator will use the given       */
    /* scheduler and configurator.                                */


void communicator_delete(struct communicator *);

    /* Deletes the communicator.  After it has been deleted, it  */
    /* must not be passed to any functions.  Inboxes created by  */
    /* the communicator will no longer be active, but will still */
    /* exist until they are deleted.                             */


const unsigned char *communicator_get_address(struct communicator *);

    /* Returns a pointer to the address of the communicator.   */
    /* The pointer is valid until the communicator is deleted. */


struct message_inbox *message_inbox_new(
  struct communicator *, const unsigned char * );

    /* message_inbox_new(communicator,address) returns a pointer to  */
    /* a new inbox for receiving messages.  The inbox will belong    */
    /* to the given communicator and have the given address.  If     */
    /* the address is a null pointer, an arbitrary address will be   */
    /* chosen.  The return value will be a null pointer if the inbox */
    /* could not be created (for example, if the requested address   */
    /* is invalid or already in use, or if the underlying network    */
    /* support is currently unavailable).                            */


void message_inbox_delete(struct message_inbox *);

    /* Deletes the inbox.  Afterwards, incoming      */
    /* messages addressed to that inbox are dropped. */


const unsigned char *message_inbox_get_address(struct message_inbox *);

    /* Returns a pointer to the address of the inbox.   */
    /* The pointer is valid until the inbox is deleted. */


/* The following action codes can be returned from a Message_Handler */
/* (see below) to indicate what should happen to the message.        */

enum message_action {
  message_keep = 0,   /* keep the message in the inbox */
  message_delete = 1  /* delete the message */
};

typedef enum message_action Message_Handler(
  const unsigned char *, size_t, void *, void * );

    /* A Message_Handler is a function that gets called              */
    /* when a message enters the inbox, and has the form             */
    /* handle(msg,size,who,what), where msg points to the message,   */
    /* size is its size in bytes, and who and what are the same      */
    /* values that were passed to message_inbox_set_handler() (see   */
    /* below).  The msg pointer is valid until the message or the    */
    /* inbox is deleted.  If the handler returns message_delete,     */
    /* the message is deleted immediately.  If the handler returns   */
    /* message_keep, the message continues to occupy the inbox until */
    /* it is deleted by a call to message_inbox_clear() (see below). */
    /* At most one message can be in the inbox at a time; subsequent */
    /* messages cannot enter the inbox until the current occupant    */
    /* is deleted.  Messages arriving at an occupied inbox might be  */
    /* queued, or might be dropped, depending on the implementation. */
    /* Messages are never truncated; if a message is too large to    */
    /* enter the inbox, it is discarded instead.                     */


void message_inbox_set_handler(
  struct message_inbox *, Message_Handler *, void *, void * );

    /* message_inbox_set_handler(inbox,handler,who,what) sets up a    */
    /* handler for message arrivals.  Whenever a message enters the   */
    /* inbox, the handler will be called.  The handler can be changed */
    /* by calling this function again.  If the handler argument is a  */
    /* null pointer then there is no handler, and arriving messages   */
    /* will be discarded.  That is the initial state of the inbox.    */


void message_inbox_clear(struct message_inbox *);

    /* message_inbox_clear(inbox) deletes the message occupying the */
    /* inbox, or does nothing if the inbox is unoccupied.  After a  */
    /* message has been deleted, the pointer to it is invalid and   */
    /* must not be used.                                            */


/* The following status codes can be returned from */
/* communicator_send() (see below)                 */

enum message_send_status {
  message_sent = 0,             /* sent (but won't necessarily arrive) */

  /* All remaining status codes mean the message */
  /* was not sent, and indicate the reason.      */

  message_unsent = 1,           /* unspecified reason */
  message_strange_address = 2,  /* address format is unknown or invalid */
  message_too_big = 3,          /* payload too big for this address format */
  message_too_fast = 4          /* queue full, resending later might work */
};


enum message_send_status communicator_send(
  struct communicator *, const unsigned char *,
  const unsigned char *, const unsigned char *, size_t ); 

    /* communicator_send(communicator, communicator_address,          */
    /* inbox_address, message, size) causes the communicator to send  */
    /* the given message of the given size to the message inbox       */
    /* indicated by the given addresses.  The message is unreliable   */
    /* (does not necessarily reach its destination).  There is no     */
    /* automatic return address; if one is needed, include it in the  */
    /* message itself.  There is no hard upper limit on the message   */
    /* size, but larger messages are more likely to run into trouble. */


struct connection_inbox *connection_inbox_new(
  struct communicator *, const unsigned char * );

    /* Returns a new inbox with the given address, for receiving   */
    /* incoming connections.  If the address is a null pointer, an */
    /* arbitrary address will be chosen.  The return value will be */
    /* a null pointer if the request could not be satisfied (for   */
    /* example, if the requested address is invalid or already     */
    /* in use, or if the underlying network support is currently   */
    /* unavailable).                                               */


void connection_inbox_delete(struct connection_inbox *);

    /* Deletes the inbox.  Afterwards, incoming         */
    /* connections addressed to that inbox are dropped. */


const unsigned char *
connection_inbox_get_address(struct connection_inbox *);

    /* Returns a pointer to the address of the inbox.   */
    /* The pointer is valid until the inbox is deleted. */


typedef void Connection_Handler(int, void *, void *);

    /* A Connection_Handler is a function that gets called when a */
    /* connection arrives, and has the form handle(fd,who,what),  */
    /* where who and what are the same values that were passed    */
    /* to connection_inbox_set_handler() (see below), and fd is   */
    /* a non-blocking file descriptor for the connection, to be   */
    /* used with read(), write(), and close().                    */


void connection_inbox_set_handler(
  struct connection_inbox *, Connection_Handler *, void *, void * );

    /* connection_inbox_set_handler(inbox,handler,who,what) requests */
    /* that the handler be called when connections arrive at the     */
    /* given inbox.  If the handler is a null pointer then no action */
    /* will be taken when connections arrive.                        */


int communicator_connect(
  struct communicator *, const unsigned char *, const unsigned char * );

    /* communicator_connect(communicator, communicator_address,     */
    /* inbox_address) returns a non-blocking file descriptor for a  */
    /* new connection from the communicator to the connection inbox */
    /* indicated by the given addresses, to be used with read(),    */
    /* write(), and close().  Returns -1 on failure.  Even if -1    */
    /* is not returned, the connection setup is not necessarily     */
    /* complete; it may be in progress or may yet fail.  The caller */
    /* will discover such conditions when it tries to read or write */
    /* the file descriptor.                                         */


int communicator_validate_addresses( struct communicator *,
  const unsigned char *, const unsigned char *, const unsigned char * );

    /* communicator_validate_addresses(communicator,                 */
    /* communicator_address, message_inbox_address,                  */
    /* connection_inbox_address) returns 0 if any address is invalid */
    /* or uses an unknown format, otherwise returns nonzero.  The    */
    /* communicator address must not be a null pointer, but either   */
    /* or both of the other two addresses may be a null pointer, in  */
    /* which case it is not checked.                                 */


  /* Eventually we might want functions for converting addresses */
  /* to/from customary text formats, like IPv4 dotted decimal,   */
  /* and decimal UDP/TCP port numbers.  We might also want to be */
  /* notified when the communicator goes offline or online, or   */
  /* when creating a new communicator is advisable.              */


#endif /* COMMUNICATOR_H */
