/*
delayer.c 0.0.3 (2003-Apr-11-Fri)
Adam M. Costello  
http://www.nicemice.net/amc/

Implements delayer.h 0.0.x.

This is ANSI C code.

*/


#include "delayer.h" /* version 0.0.x */

#include <assert.h>
#include <stddef.h>
#include <stdlib.h>

#include "complainer.h" /* version 5.>=1.x */
#include "rttdata.h" /* version 0.2.x */
#include "rng.h" /* version 1.>=1.x */


/*** Constants ***/

static const unsigned long randmax = 0x3FFFFFFF;

/*** Types ***/

#if UINT_MAX >= 0xFFFFFFFF
  typedef unsigned int uint32;
#else
  typedef unsigned long uint32;
#endif
    /* uint32 is an unsigned type with at least 32 bits. */

struct delayer {
  int guard;              /* see conventions */
  struct rng *rng;        /* see rng.h */
  double offset;          /* separation between first sample and 0% */
  unsigned int penult;    /* rttdata_sample_count - 2 */
  uint32 max_endpoints;   /* max number of endpoints we will allocate */
  uint32 used_endpoints;  /* number of endpoints in use */
  uint32 bitmap[1];       /* bitmap of endpoints in use */
                          /* array size determined at run-time */
};


/*** Functions ***/

/** Private functions **/

/* random_rtt(delayer,node1,node2) returns a random RTT between */
/* two nodes (which may be given in either order).  The node    */
/* indices must be distinct and less than rttdata_node_count.   */
/* The rttdata.h interface provides no assurance that the RTTs  */
/* are nonnegative, and neither does this function.             */

static double random_rtt(
  struct delayer *delayer, unsigned int node1, unsigned int node2 )
{
  double p, x, ym1, y0, y1, y2, xc, x2, x3, xc2, xc3;
  const double onesixth = 1.0 / 6, twothirds = 2.0 / 3;
  uint32 pair_index, lo, hi;
  const float *cdf;
  unsigned int i;

  if (node1 < node2) lo = node1, hi = node2;
  else lo = node2, hi = node1;
  pair_index = ((hi * (hi - 1)) >> 1) + lo;
  cdf = rttdata + rttdata_sample_count * pair_index;

  /* generate a uniform random percentile */
  p = delayer->offset + rng_float_as_double(delayer->rng, rttdata_sample_freq);

  /* find the two samples on either side */

  i = (unsigned int) p;
  i -= i == delayer->penult;
  ym1 = cdf[i-1];
  y0  = cdf[i  ];
  y1  = cdf[i+1];
  y2  = cdf[i+2];

  /* interpolate the RTT */

  x = p - i;
  xc = 1.0 - x;
  x2 = x*x;
  x3 = x2*x;
  xc2 = xc*xc;
  xc3 = xc2*xc;
  return onesixth * (xc3 * ym1 + x3 * y2) +
         (0.5 * x3  - x2  + twothirds) * y0 +
         (0.5 * xc3 - xc2 + twothirds) * y1 ;
}


/** Public functions **/

struct delayer *delayer_new(struct rng *rng)
{
  struct delayer *delayer;
  uint32 pairs, words, i;

  pairs = (uint32) rttdata_node_count * rttdata_node_count;
  words = (pairs + 31) >> 5;
  delayer = malloc( offsetof(struct delayer, bitmap) +
                    words * sizeof (uint32) );
  if (delayer == 0) return 0;
  delayer->guard = 0;  /* nothing is called, no need to set guard */
  delayer->rng = rng;
  delayer->offset = ((rttdata_sample_count - 1) - rttdata_sample_freq) * 0.5;
  assert(delayer->offset >= 1.0);
  delayer->penult = rttdata_sample_count - 2;
  i = pairs - rttdata_node_count;
  delayer->max_endpoints = i % 5 + i / 5 * 4;
  delayer->used_endpoints = 0;
  for (i = 0;  i < words;  ++i) delayer->bitmap[i] = 0;
  return delayer;
}


void delayer_delete(struct delayer *delayer)
{
  assert(delayer->guard == 0);
  /* no point in setting a guard only to free it */
  free(delayer);
}


int delayer_endpoint_alloc(
  struct delayer *delayer, struct delayer_endpoint *endpoint )
{
  unsigned int region, point;
  uint32 bit, word, shift;

  assert(delayer->guard == 0);
  if (delayer->used_endpoints > delayer->max_endpoints) return 0;
  delayer->guard = 1;

  for (;;) {
    region = rng_uint(delayer->rng, rttdata_node_count - 1);
    point = rng_uint(delayer->rng, rttdata_node_count - 1);
    if (region == point) continue;
    bit = (uint32) region * rttdata_node_count + point;
    word = bit >> 5;
    shift = bit & 0x1F;
    if ((delayer->bitmap[word] & (1 << shift)) == 0) break;
  }

  delayer->bitmap[word] |= (1 << shift);
  endpoint->region = region;
  endpoint->point = point;
  ++delayer->used_endpoints;

  delayer->guard = 0;
  return 1;
}


void delayer_endpoint_free(
  struct delayer *delayer, struct delayer_endpoint endpoint )
{
  uint32 bit, word, shift;

  assert(delayer->guard == 0);
  /* nothing is called, no need to set guard */

  bit = (uint32) endpoint.region * rttdata_node_count + endpoint.point;
  word = bit >> 5;
  shift = bit & 0x1F;
  assert((delayer->bitmap[word] & (1 << shift)) != 0);
  delayer->bitmap[word] &= ~(1 << shift);
  --delayer->used_endpoints;
}


double delayer_delay(
  struct delayer *delayer,
  struct delayer_endpoint src, struct delayer_endpoint dest )
{
  double rtt;

  assert(delayer->guard == 0);
  delayer->guard = 1;

  rtt = src.region == dest.region ?
    rttdata_region_scale[src.region] *
      random_rtt(delayer, src.point, dest.point)
  : random_rtt(delayer, src.region, dest.region) +
    rttdata_region_scale[src.region] *
      random_rtt(delayer, src.region, src.point) +
    rttdata_region_scale[dest.region] *
      random_rtt(delayer, dest.region, dest.point) ;

  /* We are obligated to return a nonnegative value. */

  if (rtt < 0.0) {
    warnf("negative RTT %g, forcing to 0\n", rtt);
    rtt = 0.0;
  }

  delayer->guard = 0;
  return rtt * 0.5;  /* RTT/2 is our best guess for the one-way delay */
}
