/*
list.h 0.0.4 (2002-Nov-04-Mon)
Adam M. Costello  
http://www.nicemice.net/amc/

This is not a module, and does not define an abstract type.  It is
merely a collection of list manipulation functions meant to be inlined
into other modules, and a structure type for the list pointers.

This is ANSI C code.

*/


#ifndef LIST_H
#define LIST_H


#ifdef INLINE
#define LIST_INLINE INLINE
#else
#define LIST_INLINE
#endif

/* If your compiler supports inline declarations, you can define     */
/* INLINE as inline (or __inline__, or whatever it takes).  Even if  */
/* INLINE is not defined, it might be possible to tell an optimizing */
/* compiler to inline functions of its own choosing.                 */


/* Any structure meant to be put into a linked list must contain      */
/* a member of type struct list_ptrs.  If it is the first member,     */
/* pointers to the larger structure and pointers to the first member  */
/* can be converted back and forth directly.  If the struct list_ptrs */
/* is not the first member, the conversion will be tricky in one      */
/* direction, requiring the offsetof() macro.                         */

struct list_ptrs {
  struct list_ptrs *next;   /* pointer to the next item */
  struct list_ptrs **prev;  /* pointer to the pointer to this item */
};


/* list_insert_item(&list, item) inserts the item at the head of  */
/* the list, so that list will then point to item.  Neither &list */
/* nor item may be null.                                          */

static LIST_INLINE void list_insert_item(
  struct list_ptrs **listptr, struct list_ptrs *item )
{
  struct list_ptrs *list = *listptr;

  (void) list_insert_item;  /* quiet function-not-used warnings */
  item->next = list;
  if (list != 0) list->prev = &item->next;
  item->prev = listptr;
  *listptr = item;
}


/* list_delete_item(item) deletes the item from the list.  */
/* It must not be a null pointer.  Only one pointer to the */
/* item will be updated (the one the item points back at), */
/* so don't create multiple pointers to an item.           */

static LIST_INLINE void list_delete_item(struct list_ptrs *item)
{
  (void) list_delete_item;  /* quiet function-not-used warnings */
  *item->prev = item->next;
  if (item->next != 0) item->next->prev = item->prev;
}


/* A struct list_ptrs can also be used as a queue.  The  */
/* list is circular, the next field is the head, and the */
/* prev field points into the tail item.                 */


/* list_init_queue(&queue) initializes a queue to be empty. */

static LIST_INLINE void list_init_queue(struct list_ptrs *queueptr)
{
  (void) list_init_queue;  /* quiet function-not-used warnings */
  queueptr->next = queueptr;
  queueptr->prev = &queueptr->next;
}


/* list_queue_is_empty(queue) returns 1 */
/* if the queue is empty, 0 otherwise.  */

static LIST_INLINE int list_queue_is_empty(struct list_ptrs queue)
{
  (void) list_queue_is_empty;  /* quiet function-not-used warnings */
  return queue.next == queue.next->next;
}


/* list_queue_head(queue) returns the item at the head of */
/* the queue, or a null pointer if the queue is empty.    */

static LIST_INLINE struct list_ptrs *list_queue_head(
  struct list_ptrs queue )
{
  (void) list_queue_head;  /* quiet function-not-used warnings */
  return queue.next == queue.next->next ? 0 : queue.next;
}


/* list_dequeue(queue) is like list_queue_head(queue), but it */
/* also deletes the item (if there is one) as a side effect.  */

static LIST_INLINE struct list_ptrs *list_dequeue(
  struct list_ptrs queue )
{
  struct list_ptrs *item = queue.next;

  (void) list_dequeue;  /* quiet function-not-used warnings */
  if (item == item->next) return 0;
  list_delete_item(item);
  return item;
}


/* list_enqueue(queue,item) inserts   */
/* the item at the tail of the queue. */

static LIST_INLINE void list_enqueue(
  struct list_ptrs queue, struct list_ptrs *item )
{
  (void) list_enqueue;  /* quiet function-not-used warnings */
  list_insert_item(queue.prev, item);
}


#undef LIST_INLINE

#endif /* LIST_H */
