/*
rng.c 1.1.1 (2003-Apr-12-Sat)
Adam M. Costello
http://www.nicemice.net/amc/

Implements rng.h 1.1.x.

This is ANSI C code that can optionally use a few non-ANSI functions if
certain preprocessor symbols are defined:

If USE_RAND_R is defined to 1, the thread-safe POSIX function rand_r()
will be used instead of the non-thread-safe ANSI function rand().
Note, however, that the quality of rand_r() is compromised by a flawed
interface design, and therefore this is not recommended.  Perhaps in the
future we will add an option to use the GNU functions random_r() and
friends.

If USE_GETTIMEOFDAY is defined to 1, the Single UNIX function
gettimeofday() will be used for the seed instead of the lower resolution
ANSI function time().

*/


#ifndef INLINE
#define INLINE
#endif

/* If your compiler supports inline declarations, you can define     */
/* INLINE as inline (or __inline__, or whatever it takes).  Even if  */
/* INLINE is not defined, it might be possible to tell an optimizing */
/* compiler to inline functions of its own choosing.                 */

#ifndef USE_RAND_R
#define USE_RAND_R 0
#endif

#ifndef USE_GETTIMEOFDAY
#define USE_GETTIMEOFDAY 0
#endif

#if USE_RAND_R
#define _POSIX_SOURCE 1
#define rng_rand(rng) rand_r(&rng->seed)
#else
#define rng_rand(rng) rand()
#endif


#include "rng.h" /* version 1.1.x */

#include <assert.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <stdlib.h>

#if USE_GETTIMEOFDAY
#include <sys/time.h>
#else
#include <time.h>
#endif


#define SIMPLE_BITS ( (RAND_MAX & (RAND_MAX >> 1)) == (RAND_MAX >> 1) )
/* SIMPLE_BITS is true iff RAND_MAX is one less than a power of two. */


struct rng {

  /* We don't bother with a guard because this module does    */
  /* not call any others, so it can't possibly be re-entered. */

  #if USE_RAND_R
  unsigned int seed;
  #endif

  int float_iterations;
  int double_iterations;
  float float_scale;
  double double_scale;
  double float_as_double_scale;

  int bits_per_rand;
  int bits_cache;
  int bits_count;

  /* The least significant bits_count bits of bits_cache are random, */
  /* and the rest are zero.  When the cache is empty, the lowest     */
  /* bits_per_rand bits of rng_rand(rng) are used to refill it.      */

  #if !SIMPLE_BITS
  int bits_rand_max;
  int bits_mask;
  #endif

  /* If RAND_MAX is not one less than a power of two, we need to     */
  /* define a smaller max that has all ones in the least significant */
  /* bits_per_rand places, and call rng_rand(rng) repeatedly until   */
  /* we get a value no greater than this smaller max.  The mask is   */
  /* bits_per_rand ones.                                             */

  double ln2;  /* log(2.0) */

};


#if !USE_RAND_R
static struct rng shared_rng;
static int srand_called = 0;
#endif


#define FLOAT_B ((float) RAND_MAX + 1)
#define DOUBLE_B ((double) RAND_MAX + 1)


struct rng *rng_new(void)
{
  unsigned int seed;
  struct rng *rng;
  float f, fscale;
  double d, dscale;
  int i, n;
  #if USE_GETTIMEOFDAY
    struct timeval tv;

    (void) gettimeofday(&tv,0);
    seed = tv.tv_sec + tv.tv_usec;
  #else
    seed = time(0);
  #endif

  #if USE_RAND_R
    rng = malloc(sizeof (struct rng));
    if (rng == 0) return 0;
    rng->seed = seed;
  #else
    rng = &shared_rng;

    if (!srand_called) {
      srand(seed);
      srand_called = 1;
    }
  #endif

  for ( f = 1.0f, fscale = 0.0f, i = 0;
        f > FLT_EPSILON;
        f /= FLOAT_B, fscale *= FLOAT_B, fscale += RAND_MAX, ++i );
  rng->float_iterations = i;
  rng->float_scale = fscale;

  for ( d = 1.0, dscale = 0.0, i = 0;
        d > DBL_EPSILON;
        d /= DOUBLE_B, dscale *= DOUBLE_B, dscale += RAND_MAX, ++i );
  rng->double_iterations = i;
  rng->double_scale = dscale;

  for ( d = 1.0, dscale = 0.0, i = 0;
        i < rng->float_iterations;  
        dscale *= DOUBLE_B, dscale += RAND_MAX, ++i );
  rng->float_as_double_scale = dscale;

  /* Eventually i will be rng->bits_per_rand. */
  /* It starts off as the width of RAND_MAX.  */

  for (n = RAND_MAX, i = 0;  n != 0;  n >>= 1, ++i);

  #if !SIMPLE_BITS

    /* The number of bits per rand is the number of ones at the    */
    /* low end of the max.  We can either use RAND_MAX as it is,   */
    /* or we can use a smaller max with more low ones, at the cost */
    /* of having to redo some rand calls.  The break-even point is */
    /* roughly when RAND_MAX has i - width(i) ones at the low end. */

    for (n = i;  n != 0;  n >>= 1, --i);  /* reduce i by its own width */
    n = RAND_MAX ^ (RAND_MAX >> i << i);
    /* Now n contains the least significant i bits of RAND_MAX. */

    if ( (n & (n >> 1)) != (n >> 1) ) {
      /* n is not simply i ones, so we use a smaller max. */
      rng->bits_rand_max = (RAND_MAX >> i << i) - 1;
    }
    else {
      /* n is simply i ones, so we don't use a smaller */
      /* max, and we might even be able to increase i. */
      rng->bits_rand_max = RAND_MAX;
      for (n = RAND_MAX >> i;  (n & 1) != 0;  n >>= 1, ++i);
      /* Now i is the number of ones at the low end of RAND_MAX. */
    }

    rng->bits_mask = (1 << i) - 1;
  #endif

  rng->bits_per_rand = i;
  rng->bits_cache = 0;
  rng->bits_count = 0;

  rng->ln2 = log(2.0);

  return rng;
}


void rng_delete(struct rng *rng)
{
  #if USE_RAND_R
    free(rng);
  #endif
}


/* For float and double it's easiest to just duplicate the code, but */
/* for uint and ulong the algorithm is sufficiently subtle that I'd  */
/* rather use a template.                                            */

#define rng_uinttype(rng,max) rng_uint(rng,max)
#define uinttype unsigned int
#define UINTTYPE_MAX UINT_MAX
#include "rnguint.c"

#define rng_uinttype(rng,max) rng_ulong(rng,max)
#define uinttype unsigned long
#define UINTTYPE_MAX ULONG_MAX
#include "rnguint.c"


float rng_float(struct rng *rng, float max)
{
  int i;
  float x;

  for ( x = rng_rand(rng), i = 1;
        i < rng->float_iterations;
        x *= FLOAT_B, x += rng_rand(rng), ++i );
  return x / rng->float_scale * max;
}


double rng_double(struct rng *rng, double max)
{
  int i;
  double x;

  for ( x = rng_rand(rng), i = 1;
        i < rng->double_iterations;
        x *= DOUBLE_B, x += rng_rand(rng), ++i );
  return x / rng->double_scale * max;
}


double rng_float_as_double(struct rng *rng, double max)
{
  int i;
  double x;

  for ( x = rng_rand(rng), i = 1;
        i < rng->float_iterations;
        x *= DOUBLE_B, x += rng_rand(rng), ++i );
  return x / rng->float_as_double_scale * max;
}


/* rng_bits(rng) returns a nonnegative number for which the least */
/* significant rng->bits_per_rand bits are random, and the others */
/* are zero.                                                      */

#if SIMPLE_BITS
#define rng_bits(rng) rng_rand(rng)
#else
static INLINE int rng_bits(struct rng *rng)
{
  int r;

  do r = rng_rand(rng);
  while (r > rng->bits_rand_max);
  return r & rng->bits_mask;
}
#endif


int rng_bit(struct rng *rng)
{
  int r;

  if (rng->bits_count == 0) {
    rng->bits_cache = rng_bits(rng);
    rng->bits_count = rng->bits_per_rand;
  }

  r = rng->bits_cache & 1;
  rng->bits_cache >>= 1;
  --rng->bits_count;
  return r;
}


int rng_bit_run(struct rng *rng)
{
  int n, cache, count;

  n = 0;
  cache = rng->bits_cache;
  count = rng->bits_count;

  while (cache == 0) {
    n += count;
    cache = rng_bits(rng);
    count = rng->bits_per_rand;
  }

  while ((cache & 1) == 0) {
    cache >>= 1;
    --count;
    ++n;
  }

  rng->bits_cache = cache >> 1;
  rng->bits_count = count - 1;
  return n;
}


int rng_bernoulli(struct rng *rng, double p)
{
  unsigned int r, cutoff;

  for (;;) {
    assert(p >= 0.0 && p <= 1.0);
    p *= DOUBLE_B;
    cutoff = (unsigned int) p;
    r = rng_rand(rng);
    if (r < cutoff) return 1;
    if (r > cutoff) return 0;
    p -= cutoff;
  }
}


double rng_exponential(struct rng *rng, double mu)
{
  return rng_bit_run(rng) * rng->ln2 - log(0.5 + rng_double(rng, 0.5));
}


unsigned char rng_crypto_byte(struct rng *rng)
{
  assert(0);  /* not implemented yet */
}
