/*
rng.h 1.1.0 (2003-Apr-11-Fri)
Adam M. Costello
http://www.nicemice.net/amc/

A random number generator (rng) that generates pseudo-random numbers of
various types.

WARNING:  Except where noted otherwise, the pseudo-random numbers
are not suitable for cryptography!  They are meant for things like
load-balancing and simulations.

This is ANSI C code.

*/


#ifndef RNG_H
#define RNG_H


/*** Incomplete types ***/

struct rng;
    /* The entity that keeps the state needed */
    /* to generate pseudo-random numbers.     */


/*** Functions for using a random number generator ***/

struct rng *rng_new(void);

    /* Returns a pointer to a new random number generator, */
    /* or a null pointer if one cannot be created.         */


void rng_delete(struct rng *);

    /* Deletes the random number generator.  After it has    */
    /* been deleted, it must not be passed to any functions. */


unsigned int rng_uint(struct rng *, unsigned int);
unsigned long rng_ulong(struct rng *, unsigned long);
float rng_float(struct rng *, float);
double rng_double(struct rng *, double);
double rng_float_as_double(struct rng*, double);

    /* Returns a random unsigned int or unsigned long or float     */
    /* or double having a uniform distribution between zero and    */
    /* the given maximum (inclusive).  The computational cost is   */
    /* likely to be greater for rng_double() than for rng_float(). */
    /* The cost of rng_float_as_double() is the same as the cost   */
    /* of rng_float() (or very nearly so), and the granularity     */
    /* is at least as fine, possibly finer.  For rng_uint() and    */
    /* rng_ulong(), the cost may be greater for wider intervals.   */


int rng_bit(struct rng *);

    /* Returns 0 or 1, with an equal probability of each.  This */
    /* is equivalent to rng_uint(rng,1), but might be faster.   */


int rng_bit_run(struct rng *);

    /* Generates random bits until a one occurs, and  */
    /* returns the number of zeros that preceeded it. */


int rng_bernoulli(struct rng *, double);

    /* rng_bernoulli(rng,p) returns 1 with probability   */
    /* p, 0 otherwise.  Must not be called with p < 0 or */
    /* p > 1.  This function is likely to be faster and  */
    /* more accurate than generating a uniform [0,1] and */
    /* comparing it with p.                              */


double rng_exponential(struct rng *, double);

    /* rng_exponential(rng,mu) returns a nonnegative value    */
    /* exponentially distributed with mean mu.  This function */
    /* is likely to be more accurate than simply taking the   */
    /* log of a uniform [0,1], and almost as fast.            */


unsigned char rng_crypto_byte(struct rng *);

    /* Returns a random byte.  This is the only function that    */
    /* returns a pseudo-random number suitable for cryptography. */


#endif /* RNG_H */
