/*
router.h 0.14.0 (2003-May-28-Wed)
Adam M. Costello  
http://www.nicemice.net/amc/

Provides an abstract distributed hash table that maps labels to
referrals.  A label is a byte string that identifies a resource.  A
referral consists of a shipping address from which the resource can be
retrieved, a POSIX timestamp indicating the version of the resource
available there, and a one-byte completeness measure indicating how much
of the resource is available: at least the first completeness/255 of it.

This is ANSI C code.

*/


#ifndef ROUTER_H
#define ROUTER_H


#include <time.h>


/*** Incomplete types ***/

struct router;
    /* The entity that coordinates with   */
    /* peers to implement the hash table. */

struct router_lookup;
    /* A reference to a lookup currently in progress. */

struct router_peer;
    /* A reference to a peer node. */

struct scheduler;
    /* See scheduler.h. */

struct configurator;
    /* See configurator.h. */

struct communicator;
    /* See communicator.h. */


/*** Functions for using a router ***/

  /* In all functions below, unless otherwise noted, the caller       */
  /* retains ownership of any memory buffers passed as arguments; the */
  /* callee will not free them, and will not continue to access them  */
  /* after the function has returned.                                 */


/* The following state codes can be returned from a */
/* Router_Referral_Handler (see below) to indicate  */
/* what state the lookup should enter.              */

enum router_lookup_state {
  router_lookup_active = 0,    /* seeking locations */
  router_lookup_inactive = 1,  /* dormant           */
  router_lookup_dead = 2       /* no longer exists  */
};

typedef enum router_lookup_state Router_Referral_Handler(
  const unsigned char *, unsigned int,
  const unsigned char *, const unsigned char *,
  time_t, unsigned char completeness, void *, void * );

    /* A Router_Referral_Handler is a function that gets called      */
    /* when a referral is found, and has the form handler(label,     */
    /* size, communicator_address, inbox_address, timestamp,         */
    /* completeness, who, what), where who and what are the same     */
    /* values that were passed to router_lookup_new().  The label    */
    /* size is given in bytes, while the addresses use the format    */
    /* described in communicator.h.  The label and addresses         */
    /* are valid only until the handler returns, so the callee       */
    /* might need to copy them if they are needed later.  The two    */
    /* addresses together make up the shipping address.  The same    */
    /* shipping address may be reported more than once.  A not-found */
    /* indication has null pointers for the addresses, a garbage     */
    /* value for the timestamp, and a completeness of zero if there  */
    /* was an exact match between the label and the responder's      */
    /* center of responsibility, nonzero otherwise.  Not-found       */
    /* indications are not definitive, and it may be fruitful to     */
    /* keep trying (but several such indications increase confidence */
    /* that the resource does not exist).  If the handler returns    */
    /* router_lookup_active, the router will seek more referrals     */
    /* (just as if router_lookup_resume() had been called).  If      */
    /* the handler returns router_lookup_inactive, the router will   */
    /* not seek more referrals (just as if router_lookup_suspend()   */
    /* had been called), but the calls to the handler will not       */
    /* necessarily cease immediately.  If the handler returns        */
    /* router_lookup_dead, the lookup will be deleted (just as if    */
    /* router_lookup_delete() had been called), and there will be no */
    /* more calls to the handler.                                    */
  

struct router *router_new(
  struct scheduler *, struct configurator *,
  struct communicator *, const unsigned char *, unsigned int );

    /* router_new(scheduler, configurator, communicator, group_id,   */
    /* size) returns a pointer to a new router, or a null pointer if */
    /* one cannot be created.  The group_id is a byte string used    */
    /* an an opaque identifier for the population of nodes to which  */
    /* this node belongs.  Its size is given in bytes, and might be  */
    /* subject to an upper limit.  Soon after the router is created, */
    /* router_set_shipping_address() should be called.  The router   */
    /* might not start talking to peers until then.                  */


void router_set_shipping_address(struct router *, const unsigned char *);

    /* router_set_shipping_address(router, inbox_address) informs     */
    /* the router of the inbox address of the shipper, to be used by  */
    /* other nodes to initiate file transfers to/from this node.  The */
    /* inbox_address must not be a null pointer, but the underlying   */
    /* protocol could specify a null address meaning that no shipper  */
    /* is present.  This function cannot be called twice for the same */
    /* router.                                                        */


void router_delete(struct router *);

    /* Deletes a router.  After it has been deleted, */
    /* it must not be passed to any functions.       */


struct router_lookup *router_lookup_new(
  struct router *, const unsigned char *,
  unsigned int, Router_Referral_Handler *, void *, void * );

    /* router_lookup_new(router,label,size,handler,who,what) starts a */
    /* new lookup for the given label of the given size (in bytes).   */
    /* Whenever a referral is found (or explicitly not found), the    */
    /* handler will be called.  The return value can be passed to     */
    /* the other router_lookup_* functions.  If the lookup cannot be  */
    /* attempted (for example, if the label is an unsupported size),  */
    /* a null pointer is returned.                                    */


void router_lookup_delete(struct router_lookup *);

    /* Terminates the lookup, causing the calls to the handler to     */
    /* cease, just as if the handler had returned router_lookup_dead. */
    /* After the lookup has been deleted it must not be passed to any */
    /* functions.                                                     */


void router_lookup_suspend(struct router_lookup *);

    /* Suspends the lookup, just as if the handler had returned */
    /* router_lookup_inactive.  Has no effect if the lookup was */
    /* already inactive.                                        */


void router_lookup_resume(struct router_lookup *);

    /* Resumes the lookup, just as if the handler */
    /* had returned router_lookup_active.  Has no */
    /* effect if the lookup was already active.   */


void router_set_bandwidth(struct router *, double);

    /* router_set_bandwidth(router,bandwidth) informs the router of   */
    /* the estimated bandwidth to the "core" of the network, which    */
    /* may be made known to other nodes.  A negative value means      */
    /* reset to the initial value, which is "unknown" if the protocol */
    /* can represent it, or some default value otherwise.             */


void router_set_space(struct router *, double);

    /* router_set_space(router,bandwidth) informs the router of the   */
    /* estimated unused cache space, which may be made known to other */
    /* nodes.  A negative value means reset to the initial value,     */
    /* which is "unknown" if the protocol can represent it, or some   */
    /* default value otherwise.                                       */


void router_advertise(
  struct router *, const unsigned char *,
  unsigned int, time_t, unsigned char, unsigned char );

    /* router_advertise(router, label, size, timestamp, completeness, */
    /* old) adds the <label,timestamp> pair to the set of such pairs  */
    /* to be advertised by the router, along with the associated      */
    /* old flag and completeness measure.  The old flag must be 1     */
    /* if it is known that this version of the resource is not the    */
    /* newest, 0 if it might be the newest.  The label size is given  */
    /* in bytes.  If the pair was already in the set, the set is not  */
    /* changed, but the old flag and completeness are updated, and    */
    /* the timing of the advertisements could be affected.  The set   */
    /* is empty when the router is created.                           */

void router_unadvertise(
  struct router *, const unsigned char *, unsigned int, time_t );

    /* router_unadvertise(router,label,size,timestamp) deletes the  */
    /* given <label,timestamp> pair from the set of labels to be    */
    /* advertised by the router.  The label size is given in bytes. */
    /* If the pair was not already in the set, nothing happens.     */


void router_report_newer(
  struct router *, Router_Referral_Handler *, void *, void * );

    /* Registers a handler to be called whenever an advertised    */
    /* resource is discovered to be possibly out of date, and     */
    /* the resource's old flag is 0.  The handler will be called  */
    /* with a referral to a newer version of the resource.  The   */
    /* return value of the handler is ignored.  There is at most  */
    /* one handler for this sort of notification.  This function  */
    /* can be called again to change the handler and/or the void* */
    /* arguments.  A null pointer can be passed as the handler,   */
    /* which stops the notifications.                             */


struct router_peer *router_get_peer(struct router *);

    /* Returns a reference to a random peer node, */
    /* or a null pointer if there are no peers.   */


void router_peer_delete(struct router *, struct router_peer *);

    /* Deletes the peer object.  After it has been      */
    /* deleted, it must not be passed to any functions. */


const unsigned char *router_peer_communicator_address(
  struct router *, struct router_peer * );

    /* Returns the communicator address of the peer. */
    /* The pointer is valid only as long as the peer */
    /* object exists, but not after the peer object  */
    /* is deleted.                                   */


const unsigned char *router_peer_shipping_inbox_address(
  struct router *, struct router_peer * );

    /* Returns the shipping inbox address of the */
    /* peer.  The pointer is valid only as long  */
    /* as the peer object exists, but not after  */
    /* the peer object is deleted.               */


double router_peer_rtt(struct router *, struct router_peer *);

    /* Returns an estimate of the round-trip time  */
    /* to the peer (in seconds), or -1 if unknown. */


double router_peer_uptime(struct router *, struct router_peer *);

    /* Returns an estimate of the time the peer has been up and */
    /* running continuously (in seconds), or -1 if unknown.     */


double router_peer_bandwidth(struct router *, struct router_peer *);

    /* Returns an estimate of the peer's bandwidth (in bits */
    /* per second) to the "core" of the network (not to the */
    /* local node), or -1 if unknown.                       */


double router_peer_space(struct router *, struct router_peer *);

    /* Returns an estimate of the current unused cache space */
    /* available at the peer (in bytes), or -1 if unknown.   */


#endif /* ROUTER_H */
