/*
scheduler.h 0.6.0 (2002-Sep-09-Mon)
Adam M. Costello  
http://www.nicemice.net/amc/

A scheduler allows triggers to be registered, which cause handler
functions to be called in response to events and/or timeouts.

This is ANSI C code.

*/


#ifndef SCHEDULER_H
#define SCHEDULER_H


/*** Incomplete types ***/

struct scheduler;
    /* The entity that keeps track of triggers, */
    /* watches for events, and calls handlers.  */

struct trigger;
    /* A binding between a future event and a handler. */


/*** Complete types ***/

typedef void Trigger_Handler(void *, void *);

    /* A Trigger_Handler is a function that gets called when   */
    /* a trigger fires, and has the form handler(who,what),    */
    /* where who and what are the same values that were passed */
    /* to trigger_set() (see below).  A handler should execute */
    /* quickly, so it should never use blocking I/O.  After    */
    /* a trigger fires it ceases to exist, so the handler is   */
    /* responsible for erasing any vestiges of it.             */


enum trigger_condition {
  trigger_none = 0,
  trigger_readable = 1,
  trigger_writable = 2,
  trigger_exception = 3
};


/*** Functions for using a scheduler ***/

double scheduler_uptime(struct scheduler *);

    /* Returns the number of seconds since the scheduler was created. */


struct trigger *trigger_set(
  struct scheduler *, enum trigger_condition,
  int, double, Trigger_Handler *, void *, void * );

    /* trigger_set(scheduler,condition,fd,timeout,handler,who,what)   */
    /* returns a pointer to a new trigger, or a null pointer if one   */
    /* could not be created.  If the condition is trigger_readable,   */
    /* trigger_writable, or trigger_exception, then handler(who,what) */
    /* will be called when the file descriptor fd satisfies the       */
    /* condition or when timeout seconds have elapsed, whichever      */
    /* comes first.  A negative timeout means infinity (never time    */
    /* out).  If the condition is trigger_none, then fd is ignored,   */
    /* and the timeout must be finite (not negative).  If multiple    */
    /* triggers fire at the same time, the handlers may be called in  */
    /* any order.                                                     */


void trigger_cancel(struct scheduler *, struct trigger *);

    /* Cancels a trigger that has not yet fired.  After a trigger is */
    /* canceled, it must not be passed to any functions.  A trigger  */
    /* that has already fired must not be canceled, because it no    */
    /* longer exists, and the pointer is no longer valid.  The       */
    /* handler should erase all the invalidated pointers.            */


/*** Functions for creating and managing a scheduler ***/

struct scheduler *scheduler_new(void);

    /* Returns a pointer to a new scheduler, or a */
    /* null pointer if one could not be created.  */


void scheduler_delete(struct scheduler *);

    /* Deletes a scheduler.  Must not be called while the scheduler */
    /* is running, nor when there are outstanding triggers.  After  */
    /* a scheduler has been deleted, it must not be passed to any   */
    /* functions.                                                   */


void scheduler_run(struct scheduler *);

    /* Runs the scheduler, which repeatedly waits for events and   */
    /* calls handlers.  Must not be called while the scheduler     */
    /* is already running.  The function returns when no pending   */
    /* triggers remain (which might never happen because handlers  */
    /* typically create new triggers) or scheduler_stop() has been */
    /* called.                                                     */


void scheduler_stop(struct scheduler *);

    /* Causes scheduler_run() to return the next time a handler */
    /* returns control to it.  Must only be called while the    */
    /* scheduler is running.                                    */


#endif /* SCHEDULER_H */
