/*
sim.c 0.6.0 (2003-Jul-19-Sat)
Adam M. Costello  
http://www.nicemice.net/amc/

Main simulator.  Defines and initializes the global variables (see
simglobals.h), creates the modules, runs the scheduler, and measures the
activity.

This is ANSI C code.

*/


#ifndef SIM_DEALLOCATE
#define SIM_DEALLOCATE 0
#endif

#ifndef SIM_TAG_LABELS
#define SIM_TAG_LABELS 0
#endif


#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "complainer.h" /* version 5.>=1.x */
#include "communicator.h" /* version 0.12.x */
#include "rng.h" /* version 1.>=1.x */
#include "router.h" /* version 0.13.x */
#include "scheduler.h" /* version 0.6.x */
#include "simcomm.h" /* version 0.2.x */
#include "simglobals.h" /* version 0.3.x */
#include "simrouter.h" /* version 0.1.x */
#include "simtangle.h" /* version 0.6.x */

#ifdef malloc
#include "malloc.h"
#endif


/*** Parameters ***/

static sim_index_t  /* -n */ nodes_n = 10,
                    /* -b */ relay_bug_step = 0,
                    /* -a */ resources_per_node = 5;  /* see below */
static double       /* -p */ lookup_period = 30.0,    /* see below */
                    /* -t */ lookup_timeout = 20.0,
                    /* -q */ query_period = 0.5,
                    /* -r */ run_time = 60.0 * 60;
static int          /* -v */ verbose_tally = 0;  /* boolean */

static double       /* -f */ freeze_time = -1.0;

static const char * /* -s */ lookup_style_plan = "d";  /* see below */

                    /* -l (see simcomm.h) */
static double       lossp_transit = 0.0,
                    lossp_send_init = 0.0,
                    lossp_recv_init = 0.0;
static unsigned int lossp_send_param = 0,
                    lossp_recv_param = 0;

                    /* -i (see simcomm.h, and see below) */
static sim_index_t  fault_step = 0,
                    fault_horizon_send = -1,
                    fault_horizon_recv = -1;
static double       fault_lossp_send = 0.0,
                    fault_lossp_recv = 0.0;

                    /* -d (for divide) (see below) */
static double       partition_fraction = 0.0,
                    partition_cut_time = -1.0,
                    partition_heal_time = -1.0;

                    /* -e (for enter/exit) (see below) */
static sim_index_t  transient_n = 0,
                    transient_first = 0;
static double       transient_start_time = 0.0,
                    transient_period = 0.0,
                    transient_duration = 300.0;

/***   These default values can be changed via command-line options.

*** Notes regarding -a and -p

-a is for "advertise", and sets resources_per_node.  If it is zero, node
labels are looked up rather than resource labels.

-p sets lookup_period.  If it is not positive, no lookups are performed.

*** Notes regarding -s lookup_style_plan

lookup_style_plan is a string in which each character stands for a lookup
style (see simtangle.h):

    d     simtangle_lookup_default
    t     simtangle_lookup_reduce_time
    h     simtangle_lookup_reduce_hops
    u     simtangle_lookup_uniform
    m     simtangle_lookup_min_time
    0..9  0..9
    A..Z  10..35

The lookups will use the lookup styles specified by lookup_style_plan
in a round-robin fashion.  The string may have any length greater than
zero.

*** Notes regarding -i fault_*

When fault_step is 0, the other fault_* variables have no effect.

The proper way to specify an unlimited horizon is -1; other large values
will not have the exact same effect, even if they are larger than
nodes_n, because any other value will cause the selected nodes to be
considered "horizon-limited", which has side effects (see below).

When fault_step > 1, faulty nodes are bypassed in the chain of initial
peers.  That is, if node i is faulty, then both it and node i+1 will
start out knowing node i-1.

Resources belonging to horizon-limited senders and receivers are not
looked up.

Horizon-limited receivers do not perform lookups (but horizon-limited
senders do, if they are not also horizon-limited receivers).

*** Notes regarding -d partition_*

The cut_time and heal_time indicate the times at which the network
transitions between being partitioned or not, and the fraction
influences the relative sizes of the two partitions (each node is
assigned to the first partition with a probability equal to fraction).
A negative time means no transition.  If both transitions happen, they
cannot be simultaneous.  Whichever transition comes first determines the
initial state of the network (but an initially partitioned network is
not likely to be interesting, given the linear shape of the initial peer
graph, because it would break the chain into many disconnected pieces).
If neither transition happens, the network is not partitioned.

*** Notes regarding -e transient_*

If transient_n is 0, the other transient_* variables have no effect.

The transient_n nodes starting at index transient_first are not created
at the beginning like other nodes, but are created as the simulation
runs.  The transient nodes are created in order at the rate of one
per transient_period starting at transient_start_time, and deleted
transient_duration later, never to be revived.  But the exact start
time and end time are independently shifted by a uniform random amount
between zero and transient_period, so that the phase is completely
random.  The transient_duration must be greater than double the
transient_period, to ensure that the deletion comes after the creation.
If transient_duration is negative, the nodes are never deleted (until
the simulation ends).  If freeze_time is nonnegative, creation and
deletion of transient nodes ceases when the freeze occurs.

If the transient nodes split the permanent nodes into two ranges, then
the two ranges are not initially connected; they form two separate
chains.  When each transient node is created, is has an initial peer
chosen at random from each range of permanent nodes (therefore it has
one or two initial peers, depending on the number of ranges).

Resources belonging to transient nodes are not looked up.

*/


/*** Constants ***/

static const sim_index_t horizon_inf = -1;

static const char
   * const dirname[2] = { "sent", "received" },
   * const classname[simtangle_msg_class_n] = { "peer", "ad", "lookup" },
   * const hybrid_styles = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";


/*** Types ***/

struct lookup_job {
  double start_time;
  struct router_lookup *lookup;
  struct trigger *expiration_trigger;
  char lookup_style_char;
};

#define ABS_TIME_FMT "%.1f"
#define INTERVAL_FMT "%.3f"
#define DBL_PARAM_FMT "%.16g"


/*** Global variables ***/

struct rng *sim_rng;
struct scheduler *sim_scheduler;
sim_index_t sim_nodes_n;
struct sim_node *sim_nodes;


/*** Static variables ***/

static unsigned char *resource_labels = 0;
static long outstanding_lookups = 0;

static int
  shutting_down = 0,
  warmed_up = 0,
  lookup_style_rotator = 0,
  lookup_styles_n,
  *lookup_styles = 0;

static double
  stop_or_freeze_time,
  total_tally[2][simtangle_msg_class_n];

static sim_index_t
  initial_peers[2],
  fault_remainder,
  relay_bug_remainder,
  transient_created = 0,
  transient_deleted = 0,
  transient_end;

static struct communicator **communicators;
static struct router **routers;

    /* communicators and routers were originally static,   */
    /* and were changed to be automatic within main(), but */
    /* then they got optimized away, which was a pain for  */
    /* debugging, so now they're static again.  And now    */
    /* create_node() and msg_tally() refer to them.        */


/*** Functions ***/

static void stop(void *who, void *what)
{
  (void) stop;  /* quiet warnings if this function is not used */
  scheduler_stop(sim_scheduler);
}


static void freeze(void *who, void *what)
{
  sim_index_t i;

  for (i = 0;  i < sim_nodes_n;  ++i) {
    if (routers[i] != 0) simtangle_freeze(routers[i]);
  }
}


static void set_partition(void *who, void *what)
{
  simcomm_partition = what == 0 ? 0.0 : partition_fraction;
}


static enum router_lookup_state handle_referral(
  const unsigned char *label, unsigned int size,
  const unsigned char *communicator_address,
  const unsigned char *inbox_address,
  time_t timestamp, unsigned char completeness,
  void *who, void *what )
{
  struct lookup_job *job = what;
  double now;

  if (resources_per_node > 0) {
    if (communicator_address == 0) return router_lookup_active;
  }
  else {
    assert(communicator_address == 0);
    if (completeness) return router_lookup_active;
  }

  now = scheduler_uptime(sim_scheduler);
  printf( ABS_TIME_FMT " " INTERVAL_FMT,
          job->start_time, now - job->start_time );
  printf(" %c\n", job->lookup_style_char);
  trigger_cancel(sim_scheduler, job->expiration_trigger);
  job->expiration_trigger = 0;
  free(job);
  --outstanding_lookups;
  if ( outstanding_lookups == 0 &&
       now >= run_time             ) scheduler_stop(sim_scheduler);
  return router_lookup_dead;
}


static void handle_expiration_trigger(void *who, void *what)
{
  struct router **routerp = who, *router = *routerp;
  struct lookup_job *job = what;
  double now;

  job->expiration_trigger = 0;
  now = scheduler_uptime(sim_scheduler);

  if (router != 0) {
    printf(ABS_TIME_FMT " " INTERVAL_FMT, job->start_time, lookup_timeout);
    printf(" %c\n", job->lookup_style_char);
    router_lookup_delete(job->lookup);
  }

  job->lookup = 0;
  free(job);
  --outstanding_lookups;
  if ( outstanding_lookups == 0 &&
       now >= run_time             ) scheduler_stop(sim_scheduler);
}


static void handle_lookup_trigger(void *who, void *what)
{
  struct router **routerp = who, *router = *routerp;
  sim_index_t i, j;
  const unsigned char *label;
  double now, delay;
  struct lookup_job *job;
  struct trigger *trigger;

  if (shutting_down || router == 0) return;

  do {
    i = rng_uint(sim_rng, sim_nodes_n - transient_n - 1);
    if (i >= transient_first) i += transient_n;
  } while ( fault_step > 0 &&
            ( fault_horizon_send != horizon_inf ||
              fault_horizon_recv != horizon_inf    ) &&
            i % fault_step == fault_remainder           );

  if (resources_per_node == 0) label = sim_nodes[i].label;
  else {
    j = rng_uint(sim_rng, resources_per_node - 1);
    label = resource_labels + (i * resources_per_node + j) * sim_label_size;
  }

  now = scheduler_uptime(sim_scheduler);

  if (now >= run_time) {
    if (outstanding_lookups == 0) scheduler_stop(sim_scheduler);
    return;
  }

  job = malloc_or_fail(sizeof (struct lookup_job));
  job->start_time = now;

  job->lookup_style_char = lookup_style_plan[lookup_style_rotator];
  simtangle_set_lookup_style(router, lookup_styles[lookup_style_rotator]);
  if (++lookup_style_rotator == lookup_styles_n) lookup_style_rotator = 0;

  job->lookup = router_lookup_new(
    router, label, sim_label_size, handle_referral, 0, job );
  if (job->lookup == 0) failf("unable to start a new lookup\n");
  job->expiration_trigger = trigger_set( sim_scheduler, trigger_none, -1,
    lookup_timeout, handle_expiration_trigger, routerp, job );
  if (job->expiration_trigger == 0) failf(
    "unable to create expiration trigger\n" );
  ++outstanding_lookups;
  delay = lookup_period / 2.0 + rng_float_as_double(sim_rng, lookup_period);
  trigger = trigger_set( sim_scheduler, trigger_none, -1, delay,
                         handle_lookup_trigger, routerp, 0 );
  if (trigger == 0) failf("unable to schedule next lookup\n");
}


static void create_node(sim_index_t i)
{
  /* simrouter_initial_peers and _n need to be set before calling this. */
  /* Warning: Don't create the same i twice. */

  struct sim_node *node;
  const unsigned char
    shipping_address[1 + sim_con_addr_len] = { sim_con_addr_len };
  unsigned char *start, *end, *p;
  sim_index_t j;
  double delay;
  struct trigger *trigger;

  /* Create the communicator and router. */

  node = sim_nodes + i;
  communicators[i] = communicator_new(node->nmc_addr, sim_scheduler, 0);
  if (communicators[i] == 0) failf("unable to create communicator %d\n", i);
  simtangle_relay_bug =
    relay_bug_step > 0 && i % relay_bug_step == relay_bug_remainder;
  routers[i] = router_new(sim_scheduler, 0, communicators[i], 0, 0);
  if (routers[i] == 0) failf("unable to create router %d\n", i);
  router_set_shipping_address(routers[i], shipping_address);

  /* Set up the advertisements. */

  if (resources_per_node > 0) {
    start = resource_labels + i * resources_per_node * sim_label_size;
    end = start + resources_per_node * sim_label_size;
    for (p = start;  p < end;  ++p) *p = rng_uint(sim_rng, 255);

    for (p = start, j = 0;  j < resources_per_node;  ++j) {
      #if SIM_TAG_LABELS
        /* tag the low bits of the label for debugging: */
        p[sim_label_size - 1] = j & 255;
        p[sim_label_size - 2] = j >> 8;
        p[sim_label_size - 3] = i & 255;
        p[sim_label_size - 4] = i >> 8;
      #endif /* SIM_TAG_LABELS */
      router_advertise(routers[i], p, sim_label_size, 0, 255, 0);
      p += sim_label_size;
    }
  }

  /* Set up the lookup trigger. */

  if ( lookup_period > 0.0 &&
       ( fault_step == 0 || fault_horizon_recv == horizon_inf ||
         i % fault_step != fault_remainder ) ) {
    delay = rng_float_as_double(sim_rng, lookup_period);
    trigger = trigger_set( sim_scheduler, trigger_none, -1, delay,
                           handle_lookup_trigger, routers + i, 0 );
    if (trigger == 0) failf("unable to schedule lookup for router %d\n", i);
  }
}


static void delete_node(sim_index_t i)
{
  /* Warning: Don't delete a non-existent node. */

  router_delete(routers[i]);
  routers[i] = 0;
  communicator_delete(communicators[i]);
  communicators[i] = 0;
}


static Trigger_Handler handle_create_trigger;

static void schedule_next_create(void)
{
  /* Warning: Don't call this if transient_nodes_n == 0. */

  double when, delay;
  struct trigger *trigger;

  when = transient_start_time + transient_period * transient_created +
         rng_double(sim_rng, transient_period);
  if (when >= stop_or_freeze_time) return;
  delay = when - scheduler_uptime(sim_scheduler);
  assert(delay >= 0.0);
  trigger = trigger_set(
    sim_scheduler, trigger_none, -1, delay, handle_create_trigger, 0, 0 );
  if (trigger == 0) failf("unable to schedule next transient node creation\n");
}


static Trigger_Handler handle_delete_trigger;

static void schedule_next_delete(void)
{
  /* Warning: Don't call this if transient_nodes_n == 0. */

  double when, delay;
  struct trigger *trigger;

  when = transient_start_time + transient_duration +
         transient_period * transient_deleted +
         rng_double(sim_rng, transient_period);
  if (when >= stop_or_freeze_time) return;
  delay = when - scheduler_uptime(sim_scheduler);
  assert(delay >= 0.0);
  trigger = trigger_set(
    sim_scheduler, trigger_none, -1, delay, handle_delete_trigger, 0, 0 );
  if (trigger == 0) failf("unable to schedule next transient node deletion\n");
}


static void handle_create_trigger(void *who, void *what)
{
  simrouter_initial_peers_n = 0;

  if (transient_first > 0) {
    initial_peers[simrouter_initial_peers_n++] =
      rng_uint(sim_rng, transient_first - 1);
  }

  if (transient_end < sim_nodes_n) {
    initial_peers[simrouter_initial_peers_n++] =
      transient_end + rng_uint(sim_rng, sim_nodes_n - transient_end - 1);
  }

  create_node(transient_first + transient_created++);
  if (transient_created < transient_n) schedule_next_create();
}


static void router_msg_tally(int i)
{
  int direction, class;
  unsigned long tally;

  for (direction = 0;  direction < 2;  ++direction) {
    for (class = 0;  class < simtangle_msg_class_n;  ++class) {
      if ( (resources_per_node == 0 && class == simtangle_msg_ad) ||
           (lookup_period <= 0.0 && class == simtangle_msg_lookup)
         ) continue;
      tally = simtangle_get_msg_tally(routers[i], direction, class);
      if (verbose_tally) printf( "# node %u %s %s = %lu\n",
        i, dirname[direction], classname[class], tally );
      total_tally[direction][class] += tally;
    }
  }
}


static void msg_tally(void)
{
  sim_index_t i;
  int direction, class;

  for (i = 0;  i < nodes_n;  ++i) {
    if (routers[i] != 0) router_msg_tally(i);
  }

  for (direction = 0;  direction < 2;  ++direction) {
    for (class = 0;  class < simtangle_msg_class_n;  ++class) {
      if ( (resources_per_node == 0 && class == simtangle_msg_ad) ||
           (lookup_period <= 0.0 && class == simtangle_msg_lookup)
         ) continue;
      printf( "# total %s %s = %.0f\n",
              dirname[direction], classname[class],
              total_tally[direction][class] );
    }
  }
}


static void partial_msg_tally(void *who, void *what)
{
  double tmp[2][simtangle_msg_class_n];

  memcpy(tmp, total_tally, sizeof tmp);
  msg_tally();
  memcpy(total_tally, tmp, sizeof tmp);
}


static void warmup_callback(
  double now, struct router *router, const unsigned char *address )
{
  struct trigger *trigger;

  printf("# " ABS_TIME_FMT " warmed_up = %d\n", now, ++warmed_up);

  if (warmed_up == sim_nodes_n) {
    printf("# " ABS_TIME_FMT " converged\n", now);
    if (lookup_period <= 0.0) scheduler_stop(sim_scheduler);
    else {
      trigger = trigger_set(
        sim_scheduler, trigger_none, -1, 0.0, partial_msg_tally, 0, 0 );
      if (trigger == 0) failf("unable to create tally trigger");
    }
  }
}


static void handle_delete_trigger(void *who, void *what)
{
  int i;

  assert(transient_deleted < transient_created);
  i = transient_first + transient_deleted++;
  router_msg_tally(i);
  delete_node(i);
  if (transient_deleted < transient_n) schedule_next_delete();
}


int main(int argc, char **argv)
{
  char **argp, *arg, c, *pos;
  sim_index_t i, j;
  struct sim_node *node;
  unsigned char *p;
  struct trigger *trigger;
  void *non_null_pointer = &non_null_pointer;

  /* Parse the command-line options. */

  if (*argv != 0) {
    argp = argv;

    for (;;) {
      arg = *++argp;
      if (arg == 0) break;

      if ( arg[0] != '-' || arg[2] != 0 ||
           strchr("lidenbaptqrvfs", arg[1]) == 0 ) {
        failf("unknown option %s\n", arg);
      }

      c = arg[1];
      arg = *++argp;

      if (c == 'l') {
        if (arg == 0) goto parse_need_more_arguments;
        lossp_transit = atof(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        lossp_send_init = atof(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        lossp_recv_init = atof(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        lossp_send_param = atoi(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        lossp_recv_param = atoi(arg);
        continue;
      }

      if (c == 'i') {
        if (arg == 0) goto parse_need_more_arguments;
        fault_step = atoi(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        fault_horizon_send = atoi(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        fault_horizon_recv = atoi(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        fault_lossp_send = atof(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        fault_lossp_recv = atof(arg);
        continue;
      }

      if (c == 'd') {
        if (arg == 0) goto parse_need_more_arguments;
        partition_fraction = atof(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        partition_cut_time = atof(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        partition_heal_time = atof(arg);
        continue;
      }

      if (c == 'e') {
        if (arg == 0) goto parse_need_more_arguments;
        transient_n = atoi(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        transient_first = atoi(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        transient_start_time = atof(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        transient_period = atof(arg);
        arg = *++argp;
        if (arg == 0) goto parse_need_more_arguments;
        transient_duration = atof(arg);
        continue;
      }

      if (arg == 0) {
        failf("option -%c needs an argument\n", c);
      }

      if      (c == 'n') nodes_n = atoi(arg);
      else if (c == 'b') relay_bug_step = atoi(arg);
      else if (c == 'a') resources_per_node = atoi(arg);
      else if (c == 'p') lookup_period = atof(arg);
      else if (c == 't') lookup_timeout = atof(arg);
      else if (c == 'q') query_period = atof(arg);
      else if (c == 'r') run_time = atof(arg);
      else if (c == 'v') verbose_tally = atoi(arg);
      else if (c == 'f') freeze_time = atof(arg);
      else if (c == 's') lookup_style_plan = arg;
      else assert(0);
      continue;

      parse_need_more_arguments:
      failf("option -%c needs more arguments\n", c);
    }
  }

  if (lookup_timeout < 0.0) failf("-t <lookup_timeout> must be nonnegative\n");
  if (query_period <= 0.0) failf("-q <query_period> must be positive\n");
  if (run_time < 0.0) failf("-r <run_time> must be nonnegative\n");
  if (lookup_style_plan[0] == 0) failf(
    "-s <lookup_style_plan> must not be empty\n" );
  if ( partition_cut_time >= 0.0 &&
       partition_cut_time == partition_heal_time ) failf(
    "partition_cut_time must not equal partition_heal_time\n" );
  if (transient_first + transient_n > nodes_n) failf(
    "transient_first + transient_n must not exceed nodes_n\n" );
  if (transient_start_time < 0.0) failf(
    "transient_start_time must be nonnegative\n" );
  if ( transient_duration >= 0.0 &&
       transient_duration <= 2.0 * transient_period ) failf(
     "transient_duration must be greater than 2.0 * transient_period\n" );

  printf(  "# nodes_n = %u\n", nodes_n);
  if (relay_bug_step > 0) printf(
           "# relay_bug_step = %u\n", relay_bug_step );
  printf(  "# resources_per_node = %u\n", resources_per_node);
  printf("# lookup_period = " DBL_PARAM_FMT "\n", lookup_period);
  printf("# lookup_timeout = " DBL_PARAM_FMT "\n", lookup_timeout);

  printf(  "# query_period = " DBL_PARAM_FMT "\n", query_period);
  printf(  "# run_time = " DBL_PARAM_FMT "\n", run_time);
  printf(  "# verbose_tally = %d\n", verbose_tally);
  if (freeze_time >= 0.0) printf(
           "# freeze_time = " DBL_PARAM_FMT "\n", freeze_time );
  printf(  "# lookup_style_plan = %s\n", lookup_style_plan);

  if ( lossp_transit   != 0.0 ||
       lossp_send_init != 0.0 ||
       lossp_recv_init != 0.0    ) {
    printf("# lossp_transit = " DBL_PARAM_FMT "\n", lossp_transit);
    printf("# lossp_send_init = " DBL_PARAM_FMT "\n", lossp_send_init);
    printf("# lossp_recv_init = " DBL_PARAM_FMT "\n", lossp_recv_init);
    printf("# lossp_send_param = %u\n", lossp_send_param);
    printf("# lossp_recv_param = %u\n", lossp_recv_param);
  }

  if (fault_step > 0) {
    printf("# fault_step = %u\n", fault_step);
    printf("# fault_horizon_send = %u\n", fault_horizon_send);
    printf("# fault_horizon_recv = %u\n", fault_horizon_recv);
    printf("# fault_lossp_send = " DBL_PARAM_FMT "\n", fault_lossp_send);
    printf("# fault_lossp_recv = " DBL_PARAM_FMT "\n", fault_lossp_recv);
  }

  if (partition_fraction > 0.0) {
    printf("# partition_fraction = " DBL_PARAM_FMT "\n", partition_fraction);
    printf("# partition_cut_time = " DBL_PARAM_FMT "\n", partition_cut_time);
    printf("# partition_heal_time = " DBL_PARAM_FMT "\n", partition_heal_time);
  }

  if (transient_n > 0) {
    printf("# transient_n = %u\n", transient_n);
    printf("# transient_first = %u\n", transient_first);
    printf("# transient_start_time = " DBL_PARAM_FMT "\n",
           transient_start_time);
    printf("# transient_period = " DBL_PARAM_FMT "\n", transient_period);
    printf("# transient_duration = " DBL_PARAM_FMT "\n", transient_duration);
  }

  /* Allocate the global variables. */

  sim_nodes_n = nodes_n;
  sim_nodes = malloc_or_fail(sizeof (struct sim_node) * nodes_n);

  /* Allocate the basic modules. */

  sim_rng = rng_new();
  if (sim_rng == 0) failf("unable to create sim_rng\n");
  sim_scheduler = scheduler_new();
  if (sim_scheduler == 0) failf("unable to create sim_scheduler\n");

  /* Initialize the global variables. */

  simrouter_initial_peers = initial_peers;
  simrouter_warmup_callback = warmup_callback;

  simtangle_query_period = (float) query_period;

  simcomm_lossp_transit = lossp_transit;
  simcomm_lossp_send_init = lossp_send_init;
  simcomm_lossp_recv_init = lossp_recv_init;
  simcomm_lossp_send_param = lossp_send_param;
  simcomm_lossp_recv_param = lossp_recv_param;

  simcomm_fault_step = fault_step;
  simcomm_fault_horizon_send = fault_horizon_send;
  simcomm_fault_horizon_recv = fault_horizon_recv;
  simcomm_fault_lossp_send = fault_lossp_send;
  simcomm_fault_lossp_recv = fault_lossp_recv;

  if ( partition_heal_time >= 0.0 &&
       ( partition_cut_time < 0.0 ||
         partition_cut_time > partition_heal_time )
     ) simcomm_partition = partition_fraction;

  for (i = 0;  i < nodes_n;  ++i) {
    node = sim_nodes + i;

    for (j = 0;  j < sim_label_size;  ++j) {
      node->label[j] = rng_uint(sim_rng, 255);
    }

    p = node->nmc_addr;
    memset(p, 0, sim_nmc_addr_size);
    assert(i >> 8 <= 0xFF);
    p[0] = sim_node_addr_len;
    p[1] = i >> 8;
    p[2] = i & 0xFF;
    p += 1 + sim_node_addr_len;
    p[0] = sim_msg_addr_len;
    p += 1 + sim_msg_addr_len;
    p[0] = sim_con_addr_len;
  }

  /* Allocate the static variables. */

  communicators = malloc_or_fail(nodes_n * sizeof (struct communicator *));
  routers = malloc_or_fail(nodes_n * sizeof (struct router *));
  if (resources_per_node > 0) resource_labels = malloc_or_fail(
    nodes_n * resources_per_node * sim_label_size );
  lookup_styles_n = strlen(lookup_style_plan);
  lookup_styles = malloc_or_fail(lookup_styles_n * sizeof (int));

  /* Initialize the static variables (except for node creation). */

  stop_or_freeze_time = freeze_time >= 0.0 && freeze_time < run_time ?
                        freeze_time : run_time;

  if (transient_n > 0) transient_end = transient_first + transient_n;
  else transient_first = transient_end = -1;

  for (i = transient_first;  i < transient_end;  ++i) {
    routers[i] = 0;
    communicators[i] = 0;
  }

  simrouter_initial_peers_n = 0;
  initial_peers[0] = 0;
  if (relay_bug_step > 0) relay_bug_remainder =
    (nodes_n % relay_bug_step + relay_bug_step) / 2;
  if (fault_step > 0) fault_remainder =
    (nodes_n % fault_step + fault_step) / 2;

  for (i = 0;  i < lookup_styles_n;  ++i) {
    c = lookup_style_plan[i];
    if      (c == 'd') lookup_styles[i] = simtangle_lookup_default;
    else if (c == 't') lookup_styles[i] = simtangle_lookup_reduce_time;
    else if (c == 'h') lookup_styles[i] = simtangle_lookup_reduce_hops;
    else if (c == 'u') lookup_styles[i] = simtangle_lookup_uniform;
    else if (c == 'm') lookup_styles[i] = simtangle_lookup_min_time;
    else {
      pos = strchr(hybrid_styles, c);
      if (pos == 0) failf("unrecognized lookup style %c\n", c);
      lookup_styles[i] = pos - hybrid_styles;
    }
  }

  /* Create the nodes. */

  simcomm_startup();

  for (i = 0;  i < nodes_n;  ++i) {
    if (i == transient_first) {
      i = transient_end;
      if (i >= nodes_n) break;
      simrouter_initial_peers_n = 0;
      initial_peers[0] = i;
    }

    create_node(i);
    simrouter_initial_peers_n = 1;
    if ( fault_step <= 1 ||
         i % fault_step != fault_remainder ) initial_peers[0] = i;
  }

  /* Schedule the events. */

  if (lookup_period <= 0.0) {
    trigger = trigger_set(
      sim_scheduler, trigger_none, -1, run_time, stop, 0, 0 );
    if (trigger == 0) failf("unable to create stop trigger\n");
  }

  if (freeze_time >= 0.0 && freeze_time < run_time) {
    trigger = trigger_set(
      sim_scheduler, trigger_none, -1, freeze_time, freeze, 0, 0 );
    if (trigger == 0) failf("unable to create freeze trigger\n");
  }

  if (partition_cut_time >= 0.0 && partition_cut_time < run_time) {
    trigger = trigger_set( sim_scheduler, trigger_none, -1,
      partition_cut_time, set_partition, 0, non_null_pointer );
    if (trigger == 0) failf("unable to create partition_cut trigger\n");
  }

  if (partition_heal_time >= 0.0 && partition_heal_time < run_time) {
    trigger = trigger_set( sim_scheduler, trigger_none, -1,
      partition_heal_time, set_partition, 0, 0 );
    if (trigger == 0) failf("unable to create partition_heal trigger\n");
  }

  if (transient_n > 0) {
    schedule_next_create();
    if (transient_duration >= 0.0) schedule_next_delete();
  }

  /* Run the simulation. */

  #if 0
  for (;;) {
    unsigned char ch;
    warnf("outstanding lookups: %ld\n", outstanding_lookups);
    warnf("press enter to run another 30 seconds\n");
    do *(unsigned char *)&ch = getchar();
    while (ch != '\n');

    trigger =
      trigger_set(sim_scheduler, trigger_none, -1, 30.0, stop, 0, 0);
    if (trigger == 0) failf("unable to create stop trigger\n");
    scheduler_run(sim_scheduler);
  }
  #endif

  scheduler_run(sim_scheduler);

  /* Summarize. */

  msg_tally();

  #ifdef malloc
  malloc_histogram();
  #endif

  #if SIM_DEALLOCATE
  /* Deallocation is good for discovering    */
  /* memory leaks, but it takes a long time. */

  /* Delete the nodes and let the remaining events pass. */

  for (i = 0;  i < nodes_n;  ++i) {
    if (routers[i] != 0) delete_node(i);
  }

  shutting_down = 1;
  scheduler_run(sim_scheduler);

  /* Deallocate the other modules. */

  simcomm_shutdown();
  scheduler_delete(sim_scheduler);
  sim_scheduler = 0;
  rng_delete(sim_rng);
  sim_rng = 0;

  /* Deallocate the static variables. */

  if (resource_labels != 0) {
    free(resource_labels);
    resource_labels = 0;
  }

  free(routers);
  routers = 0;
  free(communicators);
  communicators = 0;

  /* Deallocate the global variables. */

  free(sim_nodes);
  sim_nodes = 0;

  #endif /* SIM_DEALLOCATE */

  return EXIT_SUCCESS;
}
