/*
simcomm.c 0.5.0 (2003-May-21-Wed)
Adam M. Costello  
http://www.nicemice.net/amc/

A simulated implementation of communicator.h 0.12.x, providing the
backdoor interface of simcomm.h 0.2.x.

This is ANSI C code.

*/


#ifndef INLINE
#define INLINE
#endif

/* If your compiler supports inline declarations, you can define     */
/* INLINE as inline (or __inline__, or whatever it takes).  Even if  */
/* INLINE is not defined, it might be possible to tell an optimizing */
/* compiler to inline functions of its own choosing.                 */


#include "communicator.h"  /* version 0.12.x */

#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "complainer.h"  /* version 5.>=1.x */
#include "delayer.h"  /* version 0.0.x */
#include "rng.h"  /* version 1.>=0.x */
#include "simcomm.h" /* version 0.2.x */
#include "simglobals.h"  /* version 0.3.x */


/*** Types ***/

struct message_inbox {
  Message_Handler *handler;  /* handler for incoming messages */
  void *who, *what;          /* arguments for handler */
  struct message *msg;       /* arrived message, or null pointer */
};

struct communicator {
  struct message_inbox mib;
  struct delayer_endpoint endpt;  /* location of node for delay purposes */
  double send_p;  /* probability packet is not dropped when it departs */
  double recv_p;  /* prob. not dropped in transit nor when it arrives  */
  double partition_key;  /* in range [0,1) */
  unsigned int valid : 1;         /* boolean: this communicator exists */
  unsigned int mib_valid : 1;     /* boolean: message inbox exists */
};

struct message {
  size_t size;           /* size of msg in bytes         */
  unsigned char msg[1];  /* actual size is given by size */
};


/*** Global variables ***/

double       simcomm_lossp_transit = 0.0,
             simcomm_lossp_send_init = 0.0,
             simcomm_lossp_recv_init = 0.0;
unsigned int simcomm_lossp_send_param = 0,
             simcomm_lossp_recv_param = 0;

sim_index_t  simcomm_fault_step = 0,
             simcomm_fault_horizon_send = -1,  /* -1 unsigned == max */
             simcomm_fault_horizon_recv = -1;  /* -1 unsigned == max */
double       simcomm_fault_lossp_send = 0.0,
             simcomm_fault_lossp_recv = 0.0;

double       simcomm_partition = 0.0;


/*** Static variables ***

I generally don't use non-constant static variables, because
under-the-table communication is dangerous.  But for the simulation,
it's expedient.

*/

static int guard = 0;  /* No need for a separate guard per node. */

static struct communicator *communicators = 0;
    /* Array of sim_nodes_n communicator structures. */

static sim_index_t valid_communicators = 0;
static unsigned long outstanding_packets = 0;

static struct delayer *delayer = 0;  /* see delayer.h */

static double fault_successp_send, fault_successp_recv;
static sim_index_t fault_remainder;


/*** Functions ***/

/** Private functions **/

/* node_index(addr) returns the node index corresponding to the */
/* given communicator address, or a value >= sim_nodes_n if the */
/* address is invalid.                                          */

static sim_index_t node_index(const unsigned char *addr)
{
  sim_index_t i;
  const unsigned char *p, *end;

  if ( addr == 0 || addr[0] != sim_node_addr_len) return sim_nodes_n;

  for (p = addr + 3, end = addr + 1 + sim_node_addr_len;  p < end;  ++p) {
    if (*p != 0) return sim_nodes_n;
  }

  i = (addr[1] << 8) | addr[2];
  return i;
}


/* mbox_index(addr) returns 0 if the message  */
/* inbox address is valid, nonzero otherwise. */

static sim_index_t mbox_index(const unsigned char *addr)
{
  const unsigned char *p, *end;

  if (addr == 0 || addr[0] != sim_msg_addr_len) return 1;

  for (p = addr + 1, end = p + sim_msg_addr_len;  p < end;  ++p) {
    if (*p != 0) return 1;
  }

  return 0;
}


/* cbox_index(addr) returns 0 if the connection */
/* inbox address is valid, -1 otherwise.        */

static sim_index_t cbox_index(const unsigned char *addr)
{
  const unsigned char *p, *end;

  if (addr == 0 || addr[0] != sim_con_addr_len) return 1;

  for (p = addr + 1, end = p + sim_con_addr_len;  p < end;  ++p) {
    if (*p != 0) return 1;
  }

  return 0;
}


/* delete_message(inbox) deletes the message from */
/* the message inbox, which must not be empty.    */

static INLINE void delete_message(struct message_inbox *inbox)
{
  free(inbox->msg);
  inbox->msg = 0;
}


/* deliver_message(inbox,msg) delivers the message to */
/* the inbox (or drops it if the inbox is occupied).  */

static void deliver_message(void *who, void *what)
{
  struct message_inbox *inbox = who;
  struct message *msg = what;
  enum message_action action;

  assert(guard == 0);

  if (inbox->handler == 0 || inbox->msg != 0) {
    free(msg);
    goto deliver_message_end;
  }

  guard = 1;

  inbox->msg = msg;
  action = inbox->handler(msg->msg, msg->size, inbox->who, inbox->what);
  if (action == message_delete) delete_message(inbox);
  else assert(action == message_keep);

deliver_message_end:
  --outstanding_packets;

  guard = 0;
}


/** Backdoor functions **

see simcomm.h for the interface documentation

*/


void simcomm_startup(void)
{
  struct communicator nullcom = {{0}};
  sim_index_t j;

  assert(communicators == 0);
  communicators = malloc_or_fail(sim_nodes_n * sizeof (struct communicator));

  /* We could just set the valid flags to zero, but let's */
  /* set everything to zero to help catch memory bugs:    */

  for (j = 0;  j < sim_nodes_n;  ++j) communicators[j] = nullcom;

  assert(delayer == 0);
  delayer = delayer_new(sim_rng);
  if (delayer == 0) failf("unable to create delayer\n");

  fault_successp_send = 1.0 - simcomm_fault_lossp_send;
  fault_successp_recv = 1.0 - simcomm_fault_lossp_recv;
  if (simcomm_fault_step > 0) fault_remainder =
    (sim_nodes_n % simcomm_fault_step + simcomm_fault_step) / 2;
}


void simcomm_shutdown(void)
{
  assert(valid_communicators == 0);
  assert(outstanding_packets == 0);

  if (communicators != 0) {
    free(communicators);
    communicators = 0;
  }

  if (delayer != 0) {
    delayer_delete(delayer);
    delayer = 0;
  }
}


/** Public functions **

see communicator.h for the interface documentation

*/


struct communicator *communicator_new( const unsigned char *address,
  struct scheduler *scheduler, struct configurator *configurator     )
{
  sim_index_t i;
  struct communicator *communicator;
  double p;
  int r;

  assert(guard == 0);
  guard = 1;

  (void) configurator;  /* quiet warnings */

  /* For the simulation, a valid address must be supplied. */
  i = node_index(address);
  if (i >= sim_nodes_n) return 0;

  assert(communicators != 0);
  communicator = communicators + i;
  if (communicator->valid) return 0;

  assert(delayer != 0);
  r = delayer_endpoint_alloc(delayer, &communicator->endpt);
  if (!r) return 0;

  for ( p = simcomm_lossp_send_init, i = 0;
        i < simcomm_lossp_send_param;
        p *= rng_double(sim_rng, 1.0), ++i );
  communicator->send_p = 1.0 - p;
  for ( p = simcomm_lossp_recv_init, i = 0;
        i < simcomm_lossp_recv_param;
        p *= rng_double(sim_rng, 1.0), ++i );
  communicator->recv_p = (1.0 - p) * (1.0 - simcomm_lossp_transit);
  do p = rng_double(sim_rng, 1.0);
  while (p == 1.0);
  communicator->partition_key = p;

  communicator->valid = 1;
  ++valid_communicators;

  guard = 0;
  return communicator;
}


void communicator_delete(struct communicator *communicator)
{
  assert(guard == 0);
  guard = 1;

  assert(communicator->valid);
  delayer_endpoint_free(delayer, communicator->endpt);
  communicator->valid = 0;
  --valid_communicators;

  guard = 0;
}


const unsigned char *communicator_get_address(
  struct communicator *communicator )
{
  struct sim_node *node = sim_nodes + (communicator - communicators);
  assert(guard == 0);
  /* nothing is called, no need to set guard */
  return node->nmc_addr;
}


struct message_inbox *message_inbox_new(
  struct communicator *communicator, const unsigned char *address )
{
  struct message_inbox *inbox;

  assert(guard == 0);
  /* nothing is called, no need to set guard */

  /* For the simulation, the address must be zeros, or not supplied. */
  if (address != 0 && mbox_index(address) != 0) return 0;

  assert(communicator->valid);
  if (communicator->mib_valid) return 0;
  inbox = &communicator->mib;
  inbox->handler = 0;
  inbox->who = 0;
  inbox->what = 0;
  inbox->msg = 0;
  communicator->mib_valid = 1;
  return &communicator->mib;
}


void message_inbox_delete(struct message_inbox *inbox)
{
  struct communicator *communicator = (struct communicator *) inbox;

  assert(guard == 0);
  guard = 1;

  assert(communicator->mib_valid);
  if (inbox->msg != 0) delete_message(inbox);

  /* Set pointers to null to help catch memory bugs: */
  inbox->handler = 0;
  inbox->who = 0;
  inbox->what = 0;

  communicator->mib_valid = 0;

  guard = 0;
}


const unsigned char *message_inbox_get_address(
  struct message_inbox *inbox )
{
  struct communicator *communicator = (struct communicator *) inbox;
  struct sim_node *node = sim_nodes + (communicator - communicators);
  const unsigned char *nmc_addr = node->nmc_addr;
  assert(guard == 0);
  /* nothing is called, no need to set guard */
  assert(communicator->mib_valid);
  return nmc_addr + 1 + nmc_addr[0];
}


void message_inbox_set_handler( struct message_inbox *inbox,
  Message_Handler *handler, void *who, void *what            )
{
  struct communicator *communicator = (struct communicator *) inbox;
  assert(guard == 0);
  /* nothing is called, no need to set guard */
  assert(communicator->mib_valid);
  inbox->handler = handler;
  inbox->who = who;
  inbox->what = what;
}


void message_inbox_clear(struct message_inbox *inbox)
{
  struct communicator *communicator = (struct communicator *) inbox;

  assert(guard == 0);
  guard = 1;

  assert(communicator->mib_valid);
  if (inbox->msg != 0) delete_message(inbox);

  guard = 0;
}


enum message_send_status communicator_send(
  struct communicator *communicator,
  const unsigned char *communicator_address,
  const unsigned char *inbox_address,
  const unsigned char *message, size_t size  ) 
{
  sim_index_t src, dest, diff;
  struct communicator *destcom;
  struct message_inbox *destbox;
  double successp, delay;
  struct message *msg;
  struct trigger *trigger;

  assert(guard == 0);
  guard = 1;

  dest = node_index(communicator_address);
  assert(dest < sim_nodes_n && mbox_index(inbox_address) == 0);

  /* In a real implementation, that would not be an           */
  /* assertion, we would just return message_strange_address. */

  destcom = communicators + dest;
  destbox = &destcom->mib;

  /* If message gets dropped, there is nothing to do. */

  if (simcomm_partition > 0.0) {
    if ( ( communicator->partition_key <  simcomm_partition &&
           destcom->partition_key >= simcomm_partition         ) ||
         ( communicator->partition_key >= simcomm_partition &&
           destcom->partition_key <  simcomm_partition         )
       ) goto communicator_send_end;
  }

  successp = communicator->send_p * destcom->recv_p;

  if (simcomm_fault_step > 0) {
    src = communicator - communicators;
    diff = src < dest ? dest - src : src - dest;

    if (src % simcomm_fault_step == fault_remainder) {
      if (diff > simcomm_fault_horizon_send) goto communicator_send_end;
      successp *= fault_successp_send;
    }

    if (dest % simcomm_fault_step == fault_remainder) {
      if (diff > simcomm_fault_horizon_recv) goto communicator_send_end;
      successp *= fault_successp_recv;
    }
  }

  if (!rng_bernoulli(sim_rng, successp)) goto communicator_send_end;

  msg = malloc( offsetof(struct message, msg)
                + size * sizeof (unsigned char) );
  assert(msg != 0);

  /* In a real implementation, that would not be an  */
  /* assertion, we would just return message_unsent. */

  msg->size = size;
  memcpy(msg->msg, message, size);
  delay = delayer_delay(delayer, communicator->endpt, destcom->endpt);
  trigger = trigger_set( sim_scheduler, trigger_none, -1,
                         delay, deliver_message, destbox, msg );
  assert(trigger != 0);
  ++outstanding_packets;

communicator_send_end:
  guard = 0;
  return message_sent;
}


struct connection_inbox *connection_inbox_new(
  struct communicator *communicator, const unsigned char *address )
{
  failf("connection_inbox_new() cannot be called in the simulation\n");
  return 0;  /* quiet warnings */
}


void connection_inbox_delete(struct connection_inbox *inbox)
{
  failf("connection_inbox_delete()"
        " cannot be called in the simulation\n");
}


const unsigned char *
connection_inbox_get_address(struct connection_inbox *inbox)
{
  failf("connection_inbox_get_address()"
        " cannot be called in the simulation\n");
  return 0;  /* quiet warnings */
}


void connection_inbox_set_handler( struct connection_inbox *inbox,
  Connection_Handler *handler, void *who, void *what               )
{
  failf("connection_inbox_set_handler()"
        " cannot be called in the simulation\n");
}


int communicator_connect(
  struct communicator *communicator,
  const unsigned char *communicator_address,
  const unsigned char *inbox_address         )
{
  failf("communicator_connect() cannot be called in the simulation\n");
  return 0;  /* quiet warnings */
}


int communicator_validate_addresses(
  struct communicator *communicator,
  const unsigned char *communicator_address,
  const unsigned char *message_inbox_address,
  const unsigned char *connection_inbox_address )
{
  int r;

  assert(guard == 0);
  guard = 1;

  (void) communicator;  /* quiet warnings */
  r = node_index(communicator_address) < sim_nodes_n &&
      ( message_inbox_address == 0 ||
        mbox_index(message_inbox_address) == 0 ) &&
      ( connection_inbox_address == 0 ||
        cbox_index(connection_inbox_address) == 0 );

  guard = 0;
  return r;
}
