/*
simcomm.h 0.2.0 (2003-May-21-Wed)
Adam M. Costello  
http://www.nicemice.net/amc/

This is the backdoor interface to simcomm.c, whose public interface is
communicator.h.

This is ANSI C code.

Generally, I don't use non-constant static data, because under-the-table
communication is dangerous.  But for the simulation, it's expedient.
The variables declared here are defined in simcomm.c, and their values
can be altered to affect the behavior of that module.

*/


#ifndef SIMCOMM_H
#define SIMCOMM_H


#include "simglobals.h" /* 0.3.* */


/*** Functions ***/

void simcomm_startup(void);

    /* Must be called before the communicator.h interface is used. */


void simcomm_shutdown(void);

    /* May be called after all communicators have been        */
    /* deleted and all in-flight packets have been delivered. */


/*** Packet loss model ***

The following variables are parameters for a packet loss model.  Each
communicator is assigned a loss probability for sending, and one for
receiving.  Each of those probabilities is the product of a fixed
initial value (*_init) and some number (*_param) of independent uniform
[0,1] random variables.  In addition, there is a single global loss
probability for packets in transit.  If the initial values are not
altered, no packets will ever be lost.

These variables may be changed only before simcomm_startup() is called.
After that, the variables must not be changed.

*/

extern double       simcomm_lossp_transit,
                    simcomm_lossp_send_init,
                    simcomm_lossp_recv_init;
extern unsigned int simcomm_lossp_send_param,
                    simcomm_lossp_recv_param;

/*** Packet delay model

The packet delay model is not parameterized, because it's not an
analytic model.  It uses measurements to imitate real delays.

*/

/*** Index-based fault model ***

The following variables are parameters for a fault model based
on node indices.  If simcomm_fault_step is not zero (which is
the initial value), then every node whose index has a particular
value modulo simcomm_fault_step will be governed by the other
simcomm_fault parameters.  The particular value is (sim_nodes_n %
simcomm_fault_step + simcomm_fault_step) / 2, which is desiged to
stay as far as possible from the first and last nodes.  An affected
node i can send only to nodes whose indices differ from i by no
more than simcomm_fault_send_horizon (in either direction), and
can receive only from nodes whose indices differ from i by no more
than simcomm_fault_recv_horizon.  The default value for both is the
largest possible index, so that the affected nodes are not actually
affected.  The affected nodes also have constant packet loss rates,
simcomm_fault_lossp_send and simcomm_fault_lossp_recv, initially
zero, which are independent of the simcomm_lossp_* model.  (In other
words, packets need to clear both hurdles in order to get through.
You probably want to use one or the other, not both.)

These variables may be changed only before simcomm_startup() is called.
After that, the variables must not be changed.

*/

extern sim_index_t simcomm_fault_step,
                   simcomm_fault_horizon_send,
                   simcomm_fault_horizon_recv;
extern double      simcomm_fault_lossp_send,
                   simcomm_fault_lossp_recv;

/*** Partition model ***

When each node is created, it is assigned a partition key chosen
uniformly at random from the range [0,1).  The network is partitioned
with all keys < simcomm_partition on one side and all keys >=
simcomm_partition on the other side, and packets sent from one side to
the other are dropped.  When simcomm_partition is <= 0.0 or >= 1.0, all
nodes are necessarily in the same partition.  The initial value is 0.0.

*/

extern double simcomm_partition;


#endif /* SIMCOMM_H */
