/*
simglobals.h 0.3.1 (2003-Aug-11-Mon)
Adam M. Costello  
http://www.nicemice.net/amc/

Global variables used for simulations.

This is ANSI C code.

Generally, I don't use non-constant static data, because under-the-table
communication is dangerous.  But for the simulation, it's expedient.
The variables declared here must be defined in a .c file and initialized
before the creation of any objects that might access them from outside
that .c file.  The variables must never be changed as long as any such
objects exist.  Therefore, the variables can be treated as constants
when viewed from outside the .c file that defines them.

The first two bytes of a node address (after the length byte) represent
a big-endian 16-bit unsigned integer, which can be used as an index into
a global array containing various information about the node, including
permanent copies of the addresses (so you don't have to use malloc to
copy them).  The remaining bytes (if any) are zero.

For message and connection addresses, all bytes (after the length byte)
(if any) are zero, because each node has at most one message inbox and
one connection inbox.

If we wanted maximal realism, node addresses would be four bytes (not
counting the length byte) (like IPv4 addresses), message and connection
addresses would be two bytes (not counting the length byte) (like UDP &
TCP port numbers), and labels would be 20 bytes (like SHA-1 hashes).

However, to save memory (to allow the simulation of more nodes), node
addresses are two bytes, message and connection addresses are zero
bytes, and labels are eight bytes.  Protocols running in simulation mode
might want to adjust their message size limits downward.

*/


#ifndef SIMGLOBALS_H
#define SIMGLOBALS_H


/*** Types ***/

typedef unsigned short sim_index_t;
    /* An unsigned type wide enough to represent    */
    /* the number of nodes or inboxes in any array. */

enum {
  /* Realistic:
  sim_label_size = 20,
  sim_node_addr_len = 4,
  sim_msg_addr_len = 2,
  sim_con_addr_len = 2,
  */

  sim_label_size = 8,
  sim_node_addr_len = 2,
  sim_msg_addr_len = 0,
  sim_con_addr_len = 0,

  sim_nmc_addr_size =
    1 + sim_node_addr_len +
    1 + sim_msg_addr_len +
    1 + sim_con_addr_len
};

/* The global information about a node  */
/* is contained in its struct sim_node. */

struct sim_node {
  unsigned char label[sim_label_size];  /* label of this node */
  unsigned char nmc_addr[sim_nmc_addr_size];
    /* concatenation of this node's node, message, connection addresses */
};


/*** Global variables ***/

/* The following variables must not be changed while there exist     */
/* objects that might access them from outside the defining .c file. */

extern struct rng *sim_rng;  /* see rng.h */
extern struct scheduler *sim_scheduler;  /* see scheduler.h */
extern sim_index_t sim_nodes_n;  /* number of nodes */
extern struct sim_node *sim_nodes;  /* sim_nodes[k] describes node k. */


#endif /* SIMGLOBALS_H */
