/*
simsched.c 0.0.5 (2002-Oct-15-Tue)
Adam M. Costello  
http://www.nicemice.net/amc/

Implements a simulation version of scheduler.h 0.6.x.  The only trigger
condition supported is trigger_none.

This is ANSI C code.

*/


#include "scheduler.h" /* version 0.6.x */

#include <assert.h>

#include "complainer.h" /* version 5.>=1.x */
#include "rng.h"  /* version 0.2.x */


struct scheduler {
  int guard;                     /* boolean                       */
  int running;                   /* boolean                       */
  int stop_requested;            /* boolean                       */
  struct trigger *next_trigger;  /* next to fire, or null pointer */
  double uptime;                 /* virtual time since creation   */
  struct rng *rng;               /* random number generator       */
};

struct trigger {
  struct trigger **parent;   /* my parent's pointer to me          */
  struct trigger *child[2];  /* descendents fire no sooner than me */
  double fire_time;          /* time I will fire                   */
  Trigger_Handler *handler;  /* handler to call when I fire        */
  void *who, *what;          /* handler arguments                  */
};

/* In a real implementation, triggers that never time out would have */
/* null parent pointers, and would not use the child and fire_time   */
/* members.  The tree would contain only the triggers with timeouts. */
/* Triggers whose condition is not trigger_none would belong to a    */
/* linked list.  There would be three arrays of lists (indexed by    */
/* file descriptor) for trigger_readable, trigger_writable, and      */
/* trigger_exception.  When a file descriptor event happened, all    */
/* triggers in the appropriate list would fire.  Whenever a trigger  */
/* fired or was canceled, it would need to be removed from both the  */
/* list (if it was in one) and the tree (if it was in the tree).     */


/*** Functions for manipulating scheduler state ***/

/* tree_insert_trigger(rng, &root, trigger) */
/* inserts the trigger into the tree.       */

static void tree_insert_trigger(
  struct rng *rng, struct trigger **rootptr, struct trigger *trigger )
{
  struct trigger *root, *child;
  int i;

  root = *rootptr;

  while (root != 0) {
    assert(root->parent == rootptr);

    /* If *trigger should preceed *root, swap them. */

    if (trigger->fire_time < root->fire_time) {
      trigger->parent = rootptr;
      *rootptr = trigger;

      for (i = 0;  i <= 1;  ++i) {
        child = root->child[i];
        trigger->child[i] = child;
        if (child != 0) child->parent = trigger->child + i;
      }

      trigger = root;
      root = *rootptr;
    }

    /* Descend randomly to either child. */
    rootptr = root->child + rng_bit(rng);
    root = *rootptr;
  }

  *rootptr = trigger;
  trigger->parent = rootptr;
  trigger->child[0] = 0;
  trigger->child[1] = 0;
}


/* tree_delete_trigger(trigger) deletes the trigger */
/* from the tree (but does not free the trigger).   */

static void tree_delete_trigger(struct trigger *trigger)
{
  struct trigger **rootptr, *child0, *child1, *tmp0, *tmp1;
  int i;

  rootptr = trigger->parent;
  child0 = trigger->child[0];
  child1 = trigger->child[1];

  for (;;) {
    /* If we have fewer than two children, we're done in one step. */

    if (child0 == 0) {
      *rootptr = child1;
      if (child1 != 0) child1->parent = rootptr;
      return;
    }

    if (child1 == 0) {
      *rootptr = child0;
      child0->parent = rootptr;
      return;
    }

    /* Move the earlier child into the void and descend. */

    i = child0->fire_time <= child1->fire_time;

    if (i) {
      tmp0 = child0;
      tmp1 = child1;
    }
    else {
      tmp0 = child1;
      tmp1 = child0;
    }

    /* Now tmp0 is the node that will move up into  */
    /* the void, and tmp1 will become its child[i], */
    /* and tmp0->child[!i] will be the new void.    */

    *rootptr = tmp0;
    tmp0->parent = rootptr;
    child0 = tmp0->child[0];
    child1 = tmp0->child[1];
    tmp0->child[i] = tmp1;
    if (tmp1 != 0) tmp1->parent = &tmp0->child[i];
    rootptr = &tmp0->child[!i];
  }
}


/*** Public functions ***/

double scheduler_uptime(struct scheduler *scheduler)
{
  return scheduler->uptime;
}


struct trigger *trigger_set( struct scheduler *scheduler,
  enum trigger_condition condition, int fd,
  double timeout, Trigger_Handler *handler, void *who, void *what )
{
  struct trigger *trigger;

  assert(scheduler->guard == 0);
  scheduler->guard = 1;

  assert(condition == trigger_none);
  assert(timeout >= 0.0);
  trigger = malloc(sizeof (struct trigger));
  if (trigger == 0) return 0;
  trigger->fire_time = scheduler->uptime + timeout;
  trigger->handler = handler;
  trigger->who = who;
  trigger->what = what;
  tree_insert_trigger(scheduler->rng, &scheduler->next_trigger, trigger);

  scheduler->guard = 0;
  return trigger;
}


void trigger_cancel(struct scheduler *scheduler, struct trigger *trigger)
{
  assert(scheduler->guard == 0);
  scheduler->guard = 1;

  tree_delete_trigger(trigger);
  free(trigger);

  scheduler->guard = 0;
}


struct scheduler *scheduler_new(void)
{
  struct scheduler *scheduler;

  scheduler = malloc(sizeof (struct scheduler));
  if (scheduler == 0) return 0;
  scheduler->guard = 1;

  scheduler->rng = rng_new();

  if (scheduler->rng == 0) {
    free(scheduler);
    return 0;
  }

  scheduler->running = 0;
  scheduler->stop_requested = 0;
  scheduler->next_trigger = 0;
  scheduler->uptime = 0.0;

  scheduler->guard = 0;
  return scheduler;
}


void scheduler_delete(struct scheduler *scheduler)
{
  assert(scheduler->guard == 0);
  scheduler->guard = 1;
  assert(scheduler->running == 0);
  assert(scheduler->next_trigger == 0);
  rng_delete(scheduler->rng);
  free(scheduler);
}


void scheduler_run(struct scheduler *scheduler)
{
  struct trigger *trigger;
  Trigger_Handler *handler;
  void *who, *what;

  assert(scheduler->guard == 0);
  assert(scheduler->running == 0);
  scheduler->running = 1;
  scheduler->stop_requested = 0;

  for ( trigger = scheduler->next_trigger;
        trigger != 0 && !scheduler->stop_requested;
        trigger = scheduler->next_trigger           ) {
    tree_delete_trigger(trigger);
    assert(trigger->fire_time >= scheduler->uptime);
    scheduler->uptime = trigger->fire_time;
    handler = trigger->handler;
    who = trigger->who;
    what = trigger->what;
    free(trigger);
    handler(who,what);
  }

  scheduler->running = 0;
}


void scheduler_stop(struct scheduler *scheduler)
{
  assert(scheduler->guard == 0);
  /* nothing is called, no need to set guard */
  assert(scheduler->running == 1);
  scheduler->stop_requested = 1;
}
