/*
tangle.c 0.12.3 (2003-Jul-30-Wed)
Adam M. Costello  
http://www.nicemice.net/amc/

Implements router.h 0.14.x using the Tangle protocol version 0.3.0.  For
simulations, also provides the backdoor interfaces of simrouter.h 0.1.x
and simtangle.h 0.6.x.

This is ANSI C code.

*/


#ifndef SIMULATION
#define SIMULATION 0  /* 1 enables shortcuts & backdoors for simulation. */
#endif

#ifndef INLINE
#define INLINE
#endif

/* If your compiler supports inline declarations, you can define     */
/* INLINE as inline (or __inline__, or whatever it takes).  Even if  */
/* INLINE is not defined, it might be possible to tell an optimizing */
/* compiler to inline functions of its own choosing.                 */


#include "router.h" /* version 0.14.x */

#include <assert.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "communicator.h" /* version 0.12.x */
#include "complainer.h" /* version 5.>=1.x */
#include "list.h" /* version 0.0.x */
#include "rng.h" /* version 1.>=0.x */
#include "scheduler.h" /* version 0.6.x */

#if SIMULATION
#include "simglobals.h" /* version 0.3.x */
#include "simrouter.h" /* version 0.1.x */
#include "simtangle.h" /* version 0.5.x */
#endif


/*** Units ***

All times are in seconds.
All data sizes are in bytes.
All data rates are in bits per second.

*/


/*** Constants ***/

enum {
  hash_size = 20,
      /* Tangle 0.3.0 does not yet specify the hash function, but  */
      /* SHA-1 is a likely candidate.  It is 160 bits == 20 bytes. */

  label_size =
    #if SIMULATION
      sim_label_size
    #else
      hash_size
    #endif
    ,

  message_type_size = 1,

  ping_cookie_size = 1 + sizeof (double) + hash_size,
      /* The cookie contains the current time and a keyed one-way  */
      /* hash of the timestamp and the peer's node and message     */
      /* addresses, so that we can verify that replies are really  */
      /* replies, and that the timestamp has not been altered.     */
      /* This is the size and format of cookies (including the     */
      /* length byte) that we create for outgoing pings and expect */
      /* in incoming pongs.  Cookies in incoming pings are not     */
      /* necessarily this size or format, but are copied blinding  */
      /* into outgoing pongs.                                      */

  query_cookie_size = 1,
      /* We put empty cookies (plus the length byte) in outgoing   */
      /* queries, and ignore cookies in incoming reports.  Cookies */
      /* in incoming queries are not necessarily empty, and are    */
      /* copied blinding into outgoing reports.                    */

  node_params_size = 2 + 2 + 2,
      /* Uptime, bandwidth, and free space, */
      /* each a usf (see tangle-spec).      */

  message_size_soft_limit = SIMULATION ? 526 : 1024
      /* We try to send messages no larger than this, but we will     */
      /* if necessary.  The limit is smaller for simulations to       */
      /* compensate for the smaller labels and addresses.  We can fit */
      /* 27 advertisements in a post message either way (assuming     */
      /* IPv4 addresses).  The number of locations that fit in a      */
      /* report is a little different, but most simulations won't use */
      /* large reports.                                               */

};


/*** Parameters ***/

/* Parameter types */

enum fwd_hint { /* forwarding hints defined in the Tangle spec */
  fwd_reduce_time = 0,  /* 00 */
  fwd_reduce_hops = 1,  /* 01 */
  fwd_uniform = 2,      /* 10 */
  fwd_min_time = 3      /* 11 */
};

/* Parameter values */

/* TODO: The parameters should be obtained from the Configurator. */

enum {
  phase2_countdown_init    = 5,
  fwd_hint_init            = fwd_min_time,
  unconfirmed_set_limit    = 20,
  confirmed_set_limit      = 20,
  historic_threshold       = 30,
  enough_locations         =  8,
  max_locations            = enough_locations * 8,
  max_location_list_length = max_locations > 255 ? 255 : max_locations,
    /* not a parameter, just a convenient function of max_locations */
  pong_list_length         =  2
};

/* For simulations, the advertisements are sent more */
/* often and time out sooner, by a factor of 60.     */

static const float
  ping_period_0             =  5.0,
  historic_period           = 60.0 * 5,
  #if !SIMULATION
    query_period            =  0.5,
  #endif
  confirmed_peer_age_factor =  1.0,
  local_resource_period     = (SIMULATION ? 1.0 : 60.0) * 120,
  ad_buffer_age_limit       = SIMULATION ? 10.0 : 30.0,
  location_map_timeout      = (SIMULATION ? 1.0 : 60.0) * 60 * 24,
  max_rtt                   = 10.0;  /* max believable round-trip time */


/*** Types ***/

/* The Tangle spec defines four forwarding hints, corresponding   */
/* to general flavors of distributions for choosing peers.  These */
/* general flavors need to be instantiated with particular        */
/* distributions, called skew styles.                             */

enum skew_style {
  skew_none,        /* The only possible instance of fwd_uniform.      */
                    /* Choose a peer uniformly at random.              */
  skew_sqrt_speed,  /* An instance of fwd_reduce_time.                 */
                    /* From n peers choose sqrt(n) uniformly at random */
                    /* and take the one that gains the most bits/rtt,  */
                    /* or the min rtt if the bit gain is unknown.      */
  skew_sqrt_bits,   /* An instance of fwd_reduce_hops.                 */
                    /* From n peers choose sqrt(n) uniformly at random */
                    /* and take the one that gains the most bits.      */
  skew_max_speed    /* An instance of fwd_min_time.                    */
                    /* Choose the peer with the max bits/rtt.          */
};

/* struct trie_ptrs is used for internal nodes in a trie, and      */
/* appears as the first member of the first member of a leaf node. */

struct trie_ptrs {
  struct trie_ptrs *child[2];  /* both are null iff this is a leaf */
};

/* struct trie_leaf appears as the first */
/* member of any leaf node in a trie.    */

struct trie_leaf {
  struct trie_ptrs ptrs;
  unsigned char label[label_size];
};

struct unconfirmed_set {
  int n;  /* number of elements in the set */
  const unsigned char *peers[1];
    /* The n items are unordered, and start at peers[0].  Each item */
    /* is a pointer to the concatenation of the peer's node address */
    /* and its message address, in that order.  The structure is    */
    /* allocated with enough space for the maximum number of items. */
};

struct peer_params {
  float rtt;                             /* seconds between us and peer */
  unsigned char node[node_params_size];  /* uptime, bandwidth, space */
};

struct confirmed_peer {
  const unsigned char *nmc_addr;    /* concatenation of the peer's node, */
                                    /* message, and connection addresses */
  unsigned char label[label_size];  /* computed from node address */
  struct peer_params params;        /* parameters of the peer */
  double confirmation_time;         /* time the peer was confirmed */
};

struct confirmed_set {
  int n;  /* number of elements in the set */
  struct confirmed_peer peers[1];
    /* The n items are unordered, and start at peers[0]. */
    /* The structure is allocated with enough space for  */
    /* the maximum number of items.                      */
};

struct resource_params {
  time_t timestamp;             /* POSIX timestamp of resource */
  unsigned char completeness;   /* first x/255 of resource is available */
  unsigned char old;            /* boolean: this is an old version */
};

struct location {
  const unsigned char *nmc_addr;  /* concatenation of the caching node's */
                                  /* node, message, and connection addresses */
  struct resource_params params;  /* parameters of the resource */
};

struct map_resource {
  struct trie_leaf leaf;
  struct list_ptrs *instances;
    /* Nonempty list of instances, each of which is a struct       */
    /* map_resource_instance with a different <timestamp, node     */
    /* address>.  The newest arrival is first, the oldest is last. */
};

struct map_resource_instance {
  struct list_ptrs ptrs;  /* for the instances of one resource */
  struct list_ptrs q_ptrs;
    /* for the circular queue of all resource instances in the map */
  struct map_resource *resource;  /* resource that this is an instance of */
  struct location location;
  double timeout_start;  /* start time for the timeout */
};

struct advertisement {
  struct list_ptrs ptrs;
  unsigned char label[label_size];
  struct location location;
};

struct ad_buffer {
  struct list_ptrs *head;         /* newest item in buffer */
  unsigned int length;            /* number of items in buffer */
  unsigned int size;              /* number of bytes to represent buffer */
  struct trigger *flush_trigger;  /* timer to flush buffer or null if empty */
};

struct local_resource {
  struct trie_leaf leaf;
  struct list_ptrs *versions;
    /* List of versions, each of which is a struct        */
    /* local_resource_version with a different timestamp. */
};

struct local_resource_version {
  struct list_ptrs ptrs;  /* for the versions of one local resource */
  struct list_ptrs q_ptrs;
    /* for the circular queue of all local resource versions */
  struct local_resource *resource;  /* resource that this is a version of */
  struct resource_params params;
};

struct router_lookup {
  struct list_ptrs ptrs;
  struct router *router;             /* associated router */
  unsigned char label[label_size];   /* resource label sought */
  Router_Referral_Handler *handler;  /* handler to call with referrals */
  void *who, *what;                  /* opaque caller state */
  struct trigger *query_trigger;     /* trigger to send next query, */
                                     /* or null pointer if dormant */
  enum fwd_hint fwd_hint;            /* forwarding hint */
  int phase2_countdown;
    /* Number of queries remaining in the second phase of hybrid */
    /* forwarding (fwd_reduce_time), or -1 if hybrid forwarding  */
    /* is not being used.                                        */
};

struct router_peer {
  const unsigned char *nmc_addr;    /* concatenation of the peer's node, */
                                    /* message, and connection addresses */
  struct peer_params params;        /* parameters of the peer */
};

struct router {
  int guard;                               /* see conventions */
  struct rng *rng;                         /* see rng.h */
  struct scheduler *scheduler;             /* see scheduler.h */
  struct configurator *configurator;       /* see configurator.h */
  struct communicator *communicator;       /* see communicator.h */
  unsigned char *group_id;                 /* opaque group identifier */
  unsigned int group_id_size;              /* size of group_id in bytes */
  struct message_inbox *inbox;             /* for receiving messages */
  const unsigned char *message;            /* message in inbox, 0 if empty */
  size_t message_size;                     /* size of message in bytes */
  struct trigger *message_trigger;         /* trigger for handling message */
  const unsigned char *addresses;
    /* concatenation of node and message addresses, */
    /* and connection address if present            */
  unsigned int nm_size;   /* size of node and message addresses */
  unsigned int nmc_size;  /* size of all three, or 0 if third not present */
  double birthtime;                        /* used to compute uptime */
  unsigned char params[node_params_size - 2];  /* remaining parameters */
  unsigned char label[label_size];         /* label of this node */
  int unconfirmed_set_limit;               /* max peers in unconfirmed set */
  int confirmed_set_limit;                 /* max peers in confirmed set */

  int historic_threshold;                  /* no removal if n < threshold */
  int historic_n;                          /* number of historic peers */
  const unsigned char **historic_peers;
    /* array of historic peers, oldest first, length is max(n,threshold) */

  struct unconfirmed_set *unconfirmed_peers[label_size * 8];
    /* sets of unconfirmed peers, initially null, allocated on demand */

  struct confirmed_set *confirmed_peers[label_size * 8];
    /* sets of confirmed peers, initially null, allocated on demand */

  struct trie_ptrs *location_map;  /* trie of lists of locations */

  struct list_ptrs location_map_queue;
    /* circular queue of the same locations, for timing them out */

  struct trigger *location_map_trigger;
    /* trigger to time out the oldest location */

  double location_map_trigger_time;
    /* time at which the next location map trigger is expected */

  struct ad_buffer ad_buffers[label_size * 8];   /* advertisement buffers */

  struct trie_ptrs *local_resources;
    /* trie of local resources, each of which is a list of versions */

  struct list_ptrs local_resource_queue;
    /* circular queue of local resource versions, for advertising them */

  int local_resource_queue_length;         /* number of items in the queue */
  struct trigger *local_resource_trigger;  /* trigger to advertise next item */

  struct list_ptrs *lookups;               /* list of lookups in progress */

  Router_Referral_Handler *newer_handler;  /* callback for out-of-date */
  void *newer_who, *newer_what;            /* local resources, and args */

  struct trigger *ping_trigger;      /* trigger to send next ping */
  struct trigger *historic_trigger;  /* trigger to add a historic peer */
  int depth;                         /* unconfirmed sets before 1st empty */
  int extra_deep;                    /* unconfirmed peers deeper than that */
  int last_ping_level;               /* which unconfirmed set got last ping */
  float ping_period_0;               /* average, when depth is 0 or 1 */
  float ping_period;                 /* _0 / sqrt(max(depth,1)) */
  float historic_period;             /* average */
  float local_resource_period;       /* average */
  float query_period;                /* average */
  float confirmed_peer_age_limit_0;  /* when depth is 0 or 1 */
  float confirmed_peer_age_limit;    /* _0 * sqrt(max(depth,1)) */
  unsigned int ping_size;            /* size of outgoing ping messages */
  unsigned int query_size;           /* size of outgoing query messages */
  unsigned char *msg_buf;            /* scratch space for outgoing messages */
  size_t msg_buf_size;               /* size of the scratch space in bytes */
  int pong_list_length;              /* number of peers in outgoing pongs */
  int fwd_hint_init;                 /* initializer for lookup->fwd_hint */
  int phase2_countdown_init;         /* init'er for lookup->phase2_countdown */
  enum skew_style skew_styles[4];    /* map from fwd hints to skew styles */
  struct confirmed_peer best_peer[1]; /* best confirmed peer at any level */

  #if SIMULATION
  unsigned long message_tally[2][simtangle_msg_class_n];
      /* message_tally[direction][class] messages of */
      /* the given class have been sent or received. */
  short level_target;
      /* Maximum proximity between this node's label and every other */
      /* node's label (in other words, the maximum level at which    */
      /* this node could ever have a confirmed peer), or -1 after a  */
      /* confirmed peer has been found at that level.                */
  char relay_bug;  /* boolean: fail to relay queries and reports */
  #endif

  char msg_lock;  /* boolean guard: message scratch space is in use */
};


#if SIMULATION

/*** Global and static variables ***

I generally don't use non-constant global/static variables, because
under-the-table communication is dangerous.  But for the simulation,
it's expedient.

*/

sim_index_t simrouter_initial_peers_n = 0;
sim_index_t *simrouter_initial_peers = 0;
simrouter_Warmup_Callback *simrouter_warmup_callback = 0;

float simtangle_query_period = 0.5;
int simtangle_relay_bug = 0;

static unsigned char shared_message_buffer[message_size_soft_limit];

    /* Since the nodes don't actually run concurrently, they */
    /* can all share the same scratch space for constructing */
    /* outgoing messages.                                    */


#endif /* SIMULATION */


/*** Functions ***/


/* quasi_period(router,average) returns a uniform    */
/* random value between 0.5*average and 1.5*average. */

static INLINE float quasi_period(struct router *router, float average)
{
  return average * 0.5f + rng_float(router->rng, average);
}


/** Functions for working with tries **/


/* trie_find_item(trie,label) returns the leaf matching the label */
/* if there is one, a null pointer otherwise.  The trie may be a  */
/* null pointer (it must be if the trie is empty).                */

static struct trie_leaf *trie_find_item(
  struct trie_ptrs *trie, const unsigned char *label )
{
  int depth, byte, shift, bit;
  struct trie_leaf *leaf;

  for (depth = 0;  ;  ++depth) {
    if (trie == 0) return 0;

    if (trie->child[0] == 0 && trie->child[1] == 0) {
      leaf = (struct trie_leaf *) trie;
      return memcmp(label, leaf->label, label_size) == 0 ? leaf : 0;
    }

    assert(depth < label_size << 3);
    shift = (depth & 7) ^ 7;
    byte = label[depth >> 3];
    bit = (byte >> shift) & 1;
    trie = trie->child[bit];
  }
}


/* trie_delete_item(&trie,label) is like trie_find_item,    */
/* but removes the leaf (if one is found) as a side effect. */

static struct trie_leaf *trie_delete_item(
  struct trie_ptrs **trieptr, const unsigned char *label )
{
  struct trie_ptrs **stack[label_size << 3], *trie;
  int depth, byte, shift, bit;
  struct trie_leaf *leaf;

  for (depth = 0;  ;  ++depth) {
    trie = *trieptr;
    if (trie == 0) return 0;

    if (trie->child[0] == 0 && trie->child[1] == 0) {
      leaf = (struct trie_leaf *) trie;
      if (memcmp(label, leaf->label, label_size) != 0) return 0;
      *trieptr = 0;
      break;
    }

    assert(depth < label_size << 3);
    stack[depth] = trieptr;
    shift = (depth & 7) ^ 7;
    byte = label[depth >> 3];
    bit = (byte >> shift) & 1;
    trieptr = trie->child + bit;
  }

  while (--depth >= 0) {
    trieptr = stack[depth];
    trie = *trieptr;
    if (trie->child[0] != 0 || trie->child[1] != 0) break;
    free(trie);
    *trieptr = 0;
  }

  return leaf;
}


/* trie_insert_item(&trie,leaf) sets the leaf's children to null */
/* pointers (to mark it as a leaf) and inserts it into the trie. */
/* Neither argument may be null.  The leaf's label absolutely    */
/* must not already exist in the trie, so use trie_find_item or  */
/* trie_delete_item() first.                                     */

static void trie_insert_item(
  struct trie_ptrs **trieptr, struct trie_leaf *leaf )
{
  struct trie_ptrs *trie;
  int depth, byte, shift, bit, bit2;
  struct trie_leaf *leaf2;
  const unsigned char *label = leaf->label;

  leaf->ptrs.child[0] = 0;
  leaf->ptrs.child[1] = 0;

  for (depth = 0;  ;  ++depth) {
    trie = *trieptr;

    if (trie == 0) {
      /* No leaf to split, just graft the new leaf in. */
      *trieptr = &leaf->ptrs;
      return;
    }

    if (trie->child[0] == 0 && trie->child[1] == 0) {
      leaf2 = (struct trie_leaf *) trie;
      assert (memcmp(leaf2->label, label, label_size) != 0);
      break;
    }

    assert(depth < label_size << 3);
    shift = (depth & 7) ^ 7;
    byte = label[depth >> 3];
    bit = (byte >> shift) & 1;
    trieptr = trie->child + bit;
  }

  /* We need to split a leaf.  Create a chain of new internal nodes. */

  for (;  ;  ++depth) {
    *trieptr = malloc_or_fail(sizeof (struct trie_ptrs));
    trie = *trieptr;
    assert(depth < label_size << 3);
    shift = (depth & 7) ^ 7;
    byte = label[depth >> 3];
    bit = (byte >> shift) & 1;
    byte = leaf2->label[depth >> 3];
    bit2 = (byte >> shift) & 1;
    if (bit != bit2) break;
    trie->child[!bit] = 0;
    trieptr = trie->child + bit;
  }

  trie->child[bit] = &leaf->ptrs;
  trie->child[bit2] = &leaf2->ptrs;
}


/** Functions for working with addresses **/


/* sizeof_address(addr,k) returns the total size of the k-part */
/* address starting at addr, where each part is a length byte  */
/* followed by that many bytes.                                */

static INLINE unsigned int sizeof_address(const unsigned char *addr, int k)
{
  const unsigned char *p = addr;
  while (k-- > 0) p += 1 + *p;
  return p - addr;
}


/* same_address(addr1,addr2,k) returns 1 if the    */
/* two k-part addresses are the same, 0 otherwise. */

static INLINE int same_address(
  const unsigned char *addr1, const unsigned char *addr2, int k )
{
  unsigned int size1, size2;

  #if SIMULATION
    if (addr1 == addr2) return 1;

    /* For non-simulation, that's not likely enough to     */
    /* bother checking.  The other checks will still work. */

  #endif

  size1 = sizeof_address(addr1,k);
  size2 = sizeof_address(addr2,k);
  return size1 == size2 && 0 == memcmp(addr1,addr2,size1);
}


/* copy_address(addr,k) returns a copy of the k-part address   */
/* starting at addr.  If SIMULATION, then the first part must  */
/* be a node address, the other parts (if any) must be message */
/* and/or connection addresses, and k must not exceed 3.       */

static INLINE const unsigned char *copy_address(
  const unsigned char *addr, int k )
{
  unsigned char *copy;

  #if !SIMULATION
    const unsigned int size = sizeof_address(addr,k);
    copy = malloc_or_fail(size);
    memcpy(copy,addr,size);
  #else
    assert(k <= 3);
    copy = sim_nodes[(addr[1] << 8) | addr[2]].nmc_addr;
    assert(copy == addr || 0 == memcmp(copy, addr, sizeof_address(addr,k)));
  #endif

  return copy;
}


/* delete_address(addr) deletes an     */
/* address returned by copy_address(). */

static INLINE void delete_address(const unsigned char *addr)
{
  #if !SIMULATION
    free((void *) addr);
  #endif
}


/** Functions for working with labels **/


/* compute_label(buffer, node_address) computes the label             */
/* corresponding to the given node address and writes it into the     */
/* buffer, which must have space for label_size bytes.  Since a label */
/* is a sequence of bits, we need to specify the bit order within     */
/* each byte; it is big-endian (most significant first), consistent   */
/* with the SHA-1 spec.                                               */

static void compute_label(
  unsigned char *buffer, const unsigned char *node_address )
{
  #if !SIMULATION
    /* TODO: compute SHA-1 hash of node address */
    failf("compute_label() is not yet implemented\n");
  #else
    int i = (node_address[1] << 8) | node_address[2];
    memcpy(buffer, &sim_nodes[i].label, label_size);
  #endif
}


/* label_proximity(label1,label2) returns the number   */
/* of initial bits that the two labels have in common. */

static int label_proximity(
  const unsigned char *label1, const unsigned char *label2 )
{
  int i, n;
  unsigned char b1, b2;

  for (i = n = 0;  i < label_size && label1[i] == label2[i];  ++i);
  n = i * 8;
  b1 = label1[i];
  b2 = label2[i];
  for (i = 1;  b1 >> i != b2 >> i;  ++i);
  return n + 8 - i;
}


/** Functions for working with messages **/


/* start_message(router,size) returns a pointer to scratch space */
/* for constructing a message.  The space must be released using */
/* finish_message() before start_message() can be called again,  */
/* because the space is usually common.                          */

static INLINE unsigned char *start_message(
  struct router *router, unsigned int size )
{
  assert(router->msg_lock == 0);
  router->msg_lock = 1;
  return size <= router->msg_buf_size ? router->msg_buf
                                      : malloc_or_fail(size) ;
}


/* finish_message(router, msg, size, position, nm_addr) writes the  */
/* group identifier at position, asserts that it ends at the end    */
/* of the message, sends the message to the address, and releases   */
/* the scratch space.  If nm_addr is a null pointer, the message is */
/* neither sent nor released (and therefore should not have been    */
/* allocated with start_message()).                                 */

static void finish_message(
  struct router *router, unsigned char *msg, unsigned int size,
  unsigned char *position, const unsigned char *nm_addr        )
{
  unsigned int field_size = router->group_id_size;
  const unsigned char *node_addr, *msg_addr;

  memcpy(position, router->group_id, field_size);
  position += field_size;
  *position++ = field_size;
  assert(position == msg + size);

  if (nm_addr == 0) return;

  assert(router->msg_lock == 1);
  router->msg_lock = 0;
  node_addr = nm_addr;
  msg_addr = nm_addr + 1 + nm_addr[0];
  (void) communicator_send(
    router->communicator, node_addr, msg_addr, msg, size );
  if (msg != router->msg_buf) free(msg);
}


/* usf_encode(p,x) writes the value of x as two bytes in p[0..1] in   */
/* unsigned short float format (see tangle-spec).  If x is negative,  */
/* zero is written.  If x is too large to be represented, the maximum */
/* value is written.                                                  */

static void usf_encode(unsigned char *p, double x)
{
  #if SIMULATION
    p[0] = p[1] = 0;
  #else
    /* TODO */
    failf("usf_encode() is not yet implemented\n");
  #endif
}


/* usf_decode(p) returns the value of the unsigned */
/* short float in p[0..1]. (see tangle-spec).      */

static double usf_decode(const unsigned char *p)
{
  #if SIMULATION
    return 0.0;
  #else
    /* TODO */
    failf("usf_decode() is not yet implemented\n");
    return 0.0;  /* quiet warnings */
  #endif
}


/* write_address(addr,k,buffer) writes the k-part address to the */
/* buffer, which must be large enough, and returns a pointer to  */
/* the next byte waiting to be written in the buffer.            */

static INLINE unsigned char *write_address(
  const unsigned char *addr, int k, unsigned char *buffer )
{
  unsigned int size = sizeof_address(addr,k);
  memcpy(buffer,addr,size);
  return buffer + size;
}


/* read_address(k,buffer,end) returns a pointer to the byte following */
/* the k-part address in the buffer, or a null pointer if the buffer  */
/* was syntactically invalid.  The end pointer must point just beyond */
/* the end of the buffer.                                             */

static INLINE const unsigned char *read_address(
  int k, const unsigned char *buffer, const unsigned char *end )
{
  unsigned int n;
  const unsigned char *p = buffer;

  while (k-- > 0) {
    n = *p + 1;
    if (end - p < n) return 0;
    p += n;
  }

  return p;
}


/* write_cookie(cookie,buffer) writes an opaque cookie to the buffer, */
/* which must be large enough, and returns a pointer to the next byte */
/* to be written in the buffer.                                       */

static INLINE unsigned char *write_cookie(
  const unsigned char *cookie, unsigned char *buffer )
{
  unsigned int size = 1 + cookie[0];
  memcpy(buffer,cookie,size);
  return buffer + size;
}


/* read_cookie(buffer,end) returns a pointer to the byte following    */
/* the cookie in the buffer, or a null pointer if the buffer was      */
/* syntactically invalid.  The end pointer must point just beyond the */
/* end of the buffer.                                                 */

static INLINE const unsigned char *read_cookie(
  const unsigned char *buffer, const unsigned char *end )
{
  unsigned int size;

  if (end - buffer < 1) return 0;
  size = 1 + buffer[0];
  if (end - buffer < size) return 0;
  return buffer + size;
}


/* write_ping_cookie(router,buffer) writes a ping cookie (which  */
/* is also used for two-cookie pongs) to the buffer, which must  */
/* be large enough, and returns a pointer to the next byte to be */
/* written in the buffer.                                        */

static INLINE unsigned char *write_ping_cookie(
  struct router *router, unsigned char *buffer )
{
  unsigned char *p = buffer;
  double now = scheduler_uptime(router->scheduler);
  *p++ = ping_cookie_size - 1;
  memcpy(p, &now, sizeof (double));
  p += sizeof (double);
  /* TODO: Compute the hash.  For now just use zeros. */
  memset(p, 0, hash_size);
  p += hash_size;
  assert(p == buffer + ping_cookie_size);
  return p;
}


/* location_field_size(location) returns the      */
/* field size needed for a location in a message. */

static INLINE unsigned int location_field_size(struct location *location)
{
  return sizeof_address(location->nmc_addr, 3) + 6;
}


/* write_location(location,buffer) encodes the location as a */
/* sequence of bytes and writes them into the buffer, which  */
/* must be large enough.  Returns a pointer to the next byte */
/* waiting to be written in the buffer.                      */

static INLINE unsigned char *write_location(
  struct location *location, unsigned char *buffer )
{
  unsigned char *p = buffer;
  time_t timestamp;

  /* node, message, and connection addresses */
  p = write_address(location->nmc_addr, 3, p);

  /* timestamp_flag */
  timestamp = location->params.timestamp;
  assert(timestamp >= 0);
  p[4] = timestamp & 0xFF;
  timestamp >>= 8;
  p[3] = timestamp & 0xFF;
  timestamp >>= 8;
  p[2] = timestamp & 0xFF;
  timestamp >>= 8;
  p[1] = timestamp & 0xFF;
  timestamp >>= 8;
  assert(timestamp <= 0x7F);
  p[0] = (location->params.old ? 0x80 : 0) | timestamp;
  p += 5;

  /* completeness */
  *p++ = location->params.completeness;

  return p;
}


/* read_location(&nmc_addr, &c_addr, &params, buffer, end) decodes  */
/* the buffer, overwrites nmc_addr and c_addr to point to the       */
/* addresses in the buffer, and overwrites the resource parameters. */
/* Returns a pointer to the next byte to be read, or a null pointer */
/* if the buffer was syntactically invalid.  The end pointer must   */
/* point just beyond the end of the buffer.                         */

static INLINE const unsigned char *read_location(
  const unsigned char **nmc_addr, const unsigned char **c_addr,
  struct resource_params *params,
  const unsigned char *buffer, const unsigned char *end         )
{
  const unsigned char *p = buffer;
  time_t timestamp, t;

  /* node, message, and connection addresses */
  *nmc_addr = p;
  p = read_address(2,p,end);
  if (p == 0) return 0;
  *c_addr = p;
  p = read_address(1,p,end);
  if (p == 0) return 0;

  /* size of remainder is known */
  if (end - p < 6) return 0;

  /* timestamp_flag */
  params->old = *p >> 7;
  timestamp = *p++ & 0x7F;
  t = timestamp << 7;
  timestamp <<= 8;
  if (timestamp < t) return 0;
  timestamp |= *p++;
  t = timestamp << 7;
  timestamp <<= 8;
  if (timestamp < t) return 0;
  timestamp |= *p++;
  t = timestamp << 7;
  timestamp <<= 8;
  if (timestamp < t) return 0;
  timestamp |= *p++;
  t = timestamp << 7;
  timestamp <<= 8;
  if (timestamp < t) return 0;
  timestamp |= *p++;
  params->timestamp = timestamp;

  /* completeness */
  params->completeness = *p++;

  return p;
}


/** Functions for reading router state **/


/* check_confirmed_peer(set, i, min_time) checks whether              */
/* set->peers[i] (which must be in range) has a confirmation time >=  */
/* min_time.  If so, a pointer to the peer is returned, otherwise the */
/* peer is deleted and a null pointer is returned.  The set must not  */
/* be a null pointer.  This is in fact a modification of the router   */
/* state, but we consider those peers to be already gone, and we are  */
/* just doing lazy house-cleaning, so logically this is a read-only   */
/* operation.                                                         */

static struct confirmed_peer *check_confirmed_peer(
  struct confirmed_set *set, int i, double min_time )
{
  struct confirmed_peer *peer;

  peer = set->peers + i;
  if (peer->confirmation_time >= min_time) return peer;
  delete_address(peer->nmc_addr);
  *peer = set->peers[--set->n];
  return 0;
}


/* choose_uniform_confirmed_peer(rng, set, min_time) chooses       */
/* uniformly at random a peer from the set that has a confirmation */
/* time >= min_time, and returns a pointer to the peer, or a null  */
/* pointer if there are none.  The set must not be a null pointer. */
/* As a side effect, older peers may be purged.                    */

static struct confirmed_peer *choose_uniform_confirmed_peer(
  struct rng *rng, struct confirmed_set *set, double min_time )
{
  int i;
  struct confirmed_peer *peer;

  while (set->n > 0) {
    i = rng_uint(rng, set->n - 1);
    peer = check_confirmed_peer(set, i, min_time);
    if (peer != 0) return peer;
  }

  return 0;
}


/* confirmed_peer_min_time(router) returns the minimum confirmation */
/* time that any confirmed peer is currently allowed to have.       */

static INLINE double confirmed_peer_min_time(struct router *router)
{
  #if SIMULATION
    /* side effect of halt_peer_discovery() */
    if (router->ping_trigger == 0) return 0.0;
    else
  #endif
      return scheduler_uptime(router->scheduler)
             - router->confirmed_peer_age_limit;
}


/* choose_confirmed_peer(router, level, label, skew_style) returns    */
/* a peer chosen from router->confirmed_peers[level], or a null       */
/* pointer if the set is empty (or if the level is out of range).     */
/* As a side effect, some peers in the set that have exceeded the     */
/* age limit may be removed from the set (which means a set that was  */
/* nonempty might become empty).  This is in fact a modification of   */
/* the router state, but we consider those peers to be already gone,  */
/* and we are just doing lazy house-cleaning, so logically this is a  */
/* read-only operation.  Depending on skew_style, the choice may be   */
/* uniform random (skew_none), nonuniform random (skew_sqrt_speed,    */
/* skew_sqrt_bits), or deterministic (skew_max_speed).  If skew_style */
/* is *_speed, the choice will favor peers with smaller round-trip    */
/* times, and if label is not a null pointer, then the round-trip     */
/* times are divided by the number of bits of common prefix gained    */
/* by the peer compared to ourself.  For skew_sqrt_bits, the choice   */
/* favors peers with more gained bits regardless of the round-trip    */
/* times (but if the label is null, then there are no gained bits,    */
/* and the behavior degenerates into skew_none).  The label is        */
/* ignored for skew_none.                                             */

static struct confirmed_peer *choose_confirmed_peer(
  struct router *router, int level,
  const unsigned char *label, enum skew_style skew_style )
{
  struct confirmed_set *set;
  double min_time, cost, min_cost;
  struct confirmed_peer *peer, *best_peer;
  int proximity, i;

  if (level < 0 || level >= label_size * 8) return 0;
  set = router->confirmed_peers[level];
  if (set == 0) return 0;
  min_time = confirmed_peer_min_time(router);

  if ( skew_style == skew_none ||
       (skew_style == skew_sqrt_bits && label == 0) ) {
    return choose_uniform_confirmed_peer(router->rng, set, min_time);
  }

  assert( skew_style == skew_sqrt_speed ||
          skew_style == skew_sqrt_bits  ||
          skew_style == skew_max_speed     );

  best_peer = 0;
  min_cost = HUGE_VAL;
  i = skew_style != skew_max_speed;

  for (;;) {
    if (skew_style != skew_max_speed) {
      if (i*i > set->n) break;
      peer = choose_uniform_confirmed_peer(router->rng, set, min_time);
      if (peer == 0) break;
    }
    else {
      if (i >= set->n) break;
      peer = check_confirmed_peer(set, i, min_time);
      if (peer == 0) continue;
    }

    ++i;
    cost = skew_style == skew_sqrt_bits ? 1.0 : peer->params.rtt;

    if (label != 0) {
      proximity = label_proximity(peer->label, label);
      assert(proximity > level);
      cost /= proximity - level;
    }

    if (cost < min_cost) {
      best_peer = peer;
      min_cost = cost;
    }
  }

  return best_peer;
}


/* choose_confirmed_set_and_peer(router, level, skew_style) chooses */
/* a confirmed set (if skew_style is not skew_max_speed) starting   */
/* at the given level, flipping coins, and moving down one level    */
/* for each head until a tail comes up.  If that set is empty, we   */
/* we move up to the nearest nonempty set, or down to the nearest   */
/* nonempty set if moving up found nothing, or we return a null     */
/* pointer if there are no confirmed peers.  Returns a peer chosen  */
/* from the chosen set, using a distribution skewed according to    */
/* skew_style (see choose_confirmed_peer() above).  If skew_style   */
/* is skew_max_speed then level must be zero, and we return the     */
/* peer with the minimum round-trip time at any level, or a null    */
/* pointer if there are no confirmed peers.                         */

static struct confirmed_peer *choose_confirmed_set_and_peer(
  struct router *router, int level, enum skew_style skew_style )
{
  double min_time, rtt, min_rtt;
  struct confirmed_peer *peer, *best_peer;
  int i;

  if (skew_style == skew_max_speed) {
    assert(level == 0);
    min_time = confirmed_peer_min_time(router);

    if (router->best_peer->confirmation_time < min_time) {
      min_rtt = HUGE_VAL;
      best_peer = 0;

      for (i = 0;  i < label_size * 8;  ++i) {
        peer = choose_confirmed_peer(router, i, 0, skew_max_speed);
        if (peer == 0) continue;
        rtt = peer->params.rtt;

        if (rtt < min_rtt) {
          best_peer = peer;
          min_rtt = rtt;
        }
      }

      if (best_peer == 0) return 0;
      delete_address(router->best_peer->nmc_addr);
      *router->best_peer = *best_peer;
      router->best_peer->nmc_addr = copy_address(best_peer->nmc_addr, 3);
    }

    return router->best_peer;
  }

  level += rng_bit_run(router->rng);
  peer = choose_confirmed_peer(router, level, 0, skew_style);
  if (peer != 0) return peer;

  for (i = level - 1;  i >= 0;  --i) {
    peer = choose_confirmed_peer(router, i, 0, skew_style);
    if (peer != 0) return peer;
  }

  for (i = level + 1;  i < label_size * 8;  ++i) {
    peer = choose_confirmed_peer(router, i, 0, skew_style);
    if (peer != 0) return peer;
  }

  return 0;
}


/* find_local_resource_version(router,label,timestamp) returns a */
/* pointer to the local resource version matching the label and  */
/* timestamp, or a null pointer if there is no match.            */
 
static struct local_resource_version *find_local_resource_version(
  struct router *router, const unsigned char *label, time_t timestamp )
{
  struct local_resource *resource;
  struct local_resource_version *version;
  struct trie_leaf *leaf;
  struct list_ptrs *item;

  leaf = trie_find_item(router->local_resources, label);
  resource = (struct local_resource *) leaf;

  if (resource == 0) version = 0;
  else {
    item = resource->versions;

    for (;;) {
      version = (struct local_resource_version *) item;
      if (version == 0 || version->params.timestamp == timestamp) break;
      item = item->next;
    }
  }

  return version;
}


/** Functions for sending messages **/


/* send_ping(router, nm_addr) sends a ping message to the given address. */

static INLINE void send_ping(
  struct router *router, const unsigned char *nm_addr )
{
  unsigned int msg_size;
  unsigned char *msg, *p;

  msg_size = router->ping_size;
  p = msg = start_message(router, msg_size);
  p = write_address(router->addresses, 2, p);
  p = write_ping_cookie(router,p);
  *p++ = 1;  /* message type */
  finish_message(router, msg, msg_size, p, nm_addr);
  #if SIMULATION
    ++router->message_tally[simtangle_sent][simtangle_msg_peer];
  #endif
}


/* send_pong(router, nm_addr, cookie, add_cookie) sends a pong        */
/* message to the specified destination, containing the given cookie. */
/* For all three arguments, the first byte contains the number of     */
/* subsequent bytes.  If the add_cookie flag is nonzero, a second     */
/* cookie will be created and included in the pong.                   */

static void send_pong(
  struct router *router, const unsigned char *nm_addr,
  const unsigned char *cookie, int add_cookie          )
{
  struct confirmed_peer *choices[255], *peer;
  int proximity, maxlength, length, i, level;
  unsigned int msg_size, cookie_size;
  unsigned char label[label_size], msg_type;
  unsigned char *p, *msg;
  double uptime;

  if (router->nmc_size == 0) return;
  cookie_size = 1 + cookie[0];

  msg_size = 0
    + (add_cookie ? ping_cookie_size : 0)
    + cookie_size
    + router->nmc_size
    + 1 /* list length */
    /* list items will be added below */
    + node_params_size
    + message_type_size
    + router->group_id_size + 1;

  /* Choose the confirmed peers, adding their sizes as we go. */

  compute_label(label, nm_addr);
  proximity = label_proximity(label, router->label);
  maxlength = router->pong_list_length;
  length = 0;

  for (i = 0;  i < maxlength;  ++i) {
    level = proximity;
    level -= rng_bit_run(router->rng);
    if (level < 0) level = 0;
    peer = choose_confirmed_set_and_peer(router, level, skew_none);

    if (peer != 0) {
      choices[length++] = peer;
      msg_size += sizeof_address(peer->nmc_addr, 2);
    }
  }

  /* Construct the message. */

  p = msg = start_message(router, msg_size);
  msg_type = 2;

  if (add_cookie) {
    ++msg_type;
    p = write_ping_cookie(router,p);
  }

  p = write_cookie(cookie,p);
  p = write_address(router->addresses, 3, p);
  *p++ = length;

  for (i = 0;  i < length;  ++i) {
    p = write_address(choices[i]->nmc_addr, 2, p);
  }

  /* node parameters */
  uptime = scheduler_uptime(router->scheduler) - router->birthtime;
  usf_encode(p,uptime);
  memcpy(p + 2, router->params, node_params_size - 2);
  p += node_params_size;

  *p++ = msg_type;

  finish_message(router, msg, msg_size, p, nm_addr);
  #if SIMULATION
    ++router->message_tally[simtangle_sent][simtangle_msg_peer];
  #endif
}


static void clear_ad_buffer(struct router *, struct ad_buffer *);

/* send_post(router,buffer) sends a post message containing all   */
/* the advertisements in the buffer, and leaves the buffer empty. */

static void send_post(struct router *router, struct ad_buffer *buffer)
{
  int level;
  struct confirmed_peer *peer;
  unsigned int msg_size;
  unsigned char *msg, *p;
  struct list_ptrs *item;
  struct advertisement *ad;

  if (buffer->length == 0) return;

  /* Choose a random peer. */

  level = buffer - router->ad_buffers;
  assert(level >= 0 && level < label_size * 8);
  peer = choose_confirmed_peer(router, level, 0, skew_none);
  if (peer == 0) goto send_post_reset;

  /* Construct and send the message. */

  msg_size = 0
    + 1 + buffer->size
    + message_type_size
    + router->group_id_size + 1;
  p = msg = start_message(router, msg_size);

  /* length byte */
  assert(buffer->length <= 255);
  *p++ = buffer->length;

  for (item = buffer->head;  item != 0;  item = item->next) {
    ad = (struct advertisement *) item;

    /* label */
    memcpy(p, ad->label, label_size);
    p += label_size;

    /* location */
    p = write_location(&ad->location, p);
  }

  *p++ = 6;  /* message type */

  finish_message(router, msg, msg_size, p, peer->nmc_addr);
  #if SIMULATION
    ++router->message_tally[simtangle_sent][simtangle_msg_ad];
  #endif

send_post_reset:
  clear_ad_buffer(router,buffer);
}


static void answer_and_relay_query( struct router *,
  const unsigned char *, size_t, const unsigned char *, int );

/* send_query(router, label, fwd_hint) is *not* for relayed   */
/* queries, only for original queries.  It constructs a query */
/* message and sends it to ourself, without actually going    */
/* through the communication layer.                           */

static INLINE void send_query(
  struct router *router,
  const unsigned char *label, enum fwd_hint fwd_hint )
{
  enum skew_style skew_style;
  unsigned int msg_size;
  unsigned char *msg, *p, *end;

  assert(fwd_hint >= 0 && fwd_hint <= 3);
  skew_style = router->skew_styles[fwd_hint];
  msg_size = router->query_size;
  p = msg = malloc_or_fail(msg_size);

  /* We can't use the shared message buffer because we bypass the */
  /* communication layer, which would have copied the message for */
  /* us.  Instead we will call answer_and_relay_query() directly, */
  /* and it will need the shared buffer for sending reports.      */

  /* node and message addresses */
  p = write_address(router->addresses, 2, p);

  /* label */
  memcpy(p, label, label_size);
  p += label_size;

  /* cookie */
  *p++ = 0;

  /* forwarding hint */
  *p++ = fwd_hint << 6;

  /* message type */
  end = p;
  *p++ = 7;

  finish_message(router, msg, msg_size, p, 0);
  answer_and_relay_query(router, msg, msg_size, end, 1);
  free(msg);
}


/* send_report(router, nm_addr, cookie, label, count, locations,     */
/* locations_size, skew_style) sends an indirect report to nm_addr,  */
/* via a peer chosen according to skew_style, containing the         */
/* cookie (or no cookie if it is a null pointer), the label, and     */
/* the locations.  If and only if the label exactly matches our      */
/* own label, the label argument must be a null pointer.  The        */
/* locations_size must be the total of location_field_size() for all */
/* the locations.  The count of locations must not exceed 255.       */

static void send_report(
  struct router *router, const unsigned char *nm_addr,
  const unsigned char *cookie, const unsigned char *label,
  int count, struct location **locations, unsigned int locations_size,
  enum skew_style skew_style )
{
  const unsigned char *peer_addr;
  struct confirmed_peer *peer;
  unsigned int msg_size;
  unsigned char *p, *msg, msg_type;
  int i;

  assert(count <= 255);

  /* Choose a peer for the indirection. */

  if (same_address(nm_addr, router->addresses, 2)) {
    peer_addr = router->addresses;

    /* In this case we could optimize away the indirection by */
    /* sending a direct report instead of an indirect report, */
    /* but it's a corner case, not worth the extra code.      */

  }
  else {
    peer = choose_confirmed_set_and_peer(router, 0, skew_style);
    if (peer == 0) return;
    peer_addr = peer->nmc_addr;
  }

  /* Construct and send the message. */

  msg_size = 0
    + 1 + sizeof_address(nm_addr, 2) + 1
    + (cookie ? 1 + cookie[0] : 0)
    + label_size
    + 1 + locations_size
    + (label == 0)
    + message_type_size
    + router->group_id_size + 1;

  p = msg = start_message(router, msg_size);
  msg_type = 4;

  /* destination */
  *p++ = 1;
  p = write_address(nm_addr, 2, p);
  *p++ = 0;

  if (cookie != 0) {
    p = write_cookie(cookie,p);
    ++msg_type;
  }

  /* label */
  memcpy(p, label == 0 ? router->label : label, label_size);
  p += label_size;

  *p++ = count;  /* location list length */

  for (i = 0;  i < count;  ++i) {
    p = write_location(locations[i], p);
  }

  if (label == 0) *p++ = 0x80;  /* option */

  *p++ = msg_type;

  finish_message(router, msg, msg_size, p, peer_addr);
  #if SIMULATION
    ++router->message_tally[simtangle_sent][simtangle_msg_lookup];
  #endif
}


/** Functions for modifying router state **/


/* update_scaled_params(router) updates the parameters that are */
/* scaled according to router->depth.  This needs to be called  */
/* whenever the depth changes, but it's safe to call any time.  */

static INLINE void update_scaled_params(struct router *router)
{
  double scale = router->depth <= 1 ? 1.0 : sqrt(router->depth);
  router->ping_period = router->ping_period_0 / scale;
  router->confirmed_peer_age_limit =
    router->confirmed_peer_age_limit_0 * scale;
}


/* remove_unconfirmed_peer(router,level,i) removes */
/* router->unconfirmed_peers[level]->peers[i] and  */
/* updates router->depth, extra_deep, and the      */
/* parameters scaled according to depth.           */

static void remove_unconfirmed_peer(
  struct router *router, int level, int i )
{
  struct unconfirmed_set **upeers = router->unconfirmed_peers;
  const unsigned char **peer;
  int n;

  assert(level < 8 * label_size);
  n = upeers[level]->n;
  assert(i < n);
  peer = upeers[level]->peers + i;

  /* Remove the peer. */

  delete_address(*peer);
  *peer = upeers[level]->peers[n-1];
  n = --upeers[level]->n;

  /* Update depth, extra_deep, and scaled parameters. */

  assert(level != router->depth);

  if (level > router->depth) {
    --router->extra_deep;
    return;
  }

  if (n > 0) return;
  router->depth = level;

  while (++level < 8 * label_size) {
    if (upeers[level] != 0) n += upeers[level]->n;
  }

  router->extra_deep = n;
  update_scaled_params(router);
}


/* add_unconfirmed_peer(router, nm_addr) adds the given node &       */
/* message address to the appropriate unconfirmed set, and updates   */
/* depth, extra_deep, and the parameters scaled according to depth.  */
/* The caller retains ownership of the address buffer; the addresses */
/* are copied.  Nothing is done if the addresses are not valid.      */

static void add_unconfirmed_peer(
  struct router *router, const unsigned char *nm_addr )
{
  unsigned char label[label_size];
  int level, n, i, usetlim = router->unconfirmed_set_limit;
  struct unconfirmed_set **upeers = router->unconfirmed_peers;
  const unsigned char **peers;

  /* Check the validity of the addresses: */

  if ( !communicator_validate_addresses(
          router->communicator,
          nm_addr, nm_addr + 1 + *nm_addr, 0 ) ) return;

  /* Determine which level is appropriate. */

  compute_label(label, nm_addr);
  level = label_proximity(router->label, label);
  if (level >= 8 * label_size) return;

  /* If the set doesn't exist, create it. */

  if (upeers[level] == 0) {
    upeers[level] = malloc_or_fail(
      offsetof(struct unconfirmed_set, peers) +
      usetlim * sizeof (unsigned char *) );
    upeers[level]->n = 0;
  }

  /* If the node address is already present, simply */
  /* replace the old peer with the new peer.        */

  peers = upeers[level]->peers;
  n = upeers[level]->n;

  for (i = 0;  i < n;  ++i) {
    if (same_address(nm_addr, peers[i], 1)) {
      delete_address(peers[i]);
      peers[i] = copy_address(nm_addr, 2);
      return;
    }
  }

  /* If the set is already full, choose one peer to remove. */

  if (n == usetlim) {
    i = rng_uint(router->rng, usetlim);
    if (i == usetlim) return;
    delete_address(peers[i]);
    peers[i] = copy_address(nm_addr, 2);
    return;
  }

  /* Add the new peer at the end. */

  peers[n] = copy_address(nm_addr, 2);
  n = ++upeers[level]->n;

  /* Update depth, extra_deep, and scaled parameters. */

  if (level > router->depth) {
    ++router->extra_deep;
    return;
  }

  if (level < router->depth) {
    assert(n > 1);
    return;
  }

  while (    level < 8 * label_size
          && upeers[level] != 0
          && upeers[level]->n > 0    ) ++level;
  router->depth = level;

  for (n = 0;  level < 8 * label_size;  ++level) {
    if (upeers[level] != 0) n += upeers[level]->n;
  }

  router->extra_deep = n;
  update_scaled_params(router);
}


/* add_confirmed_peer(router, nmc_addr, rtt, node_params, */
/* now) adds the peer to the appropriate confirmed set.   */

static INLINE void add_confirmed_peer(
  struct router *router, const unsigned char *nmc_addr,
  double rtt, const unsigned char *node_params, double now )
{
  unsigned char label[label_size];
  int level, i, csetlim = router->confirmed_set_limit;
  struct confirmed_set **cpeers = router->confirmed_peers, *set;
  struct confirmed_peer *peers, *peer;
  double min_time;

  /* Determine which level is appropriate. */

  compute_label(label, nmc_addr);
  level = label_proximity(router->label, label);

  #if SIMULATION
    if (router->level_target >= 0) {
      if (level == router->level_target) {
        router->level_target = -1;
        if (simrouter_warmup_callback != 0) {
          simrouter_warmup_callback(now, router, router->addresses);
        }
      }
      else assert(level < router->level_target);
    }
  #endif

  if (level >= 8 * label_size) return;

  /* If the set doesn't exist, create it. */

  if (cpeers[level] == 0) {
    cpeers[level] = malloc_or_fail(
      offsetof(struct confirmed_set, peers) +
      csetlim * sizeof (struct confirmed_peer) );
    cpeers[level]->n = 0;
  }

  /* Remove peers that are too old, or that have */
  /* the same node address as the new peer.      */

  /* We don't need to check whether peer discovery has been disabled. */
  /* If it had been, this function could not have been called.        */

  min_time = now - router->confirmed_peer_age_limit;
  set = cpeers[level];
  peers = set->peers;
  i = 0;

  while (i < set->n) {
    peer = peers + i;

    if ( peer->confirmation_time < min_time ||
         same_address(nmc_addr, peer->nmc_addr, 1) ) {
      *peer = peers[--set->n];
    }
    else ++i;
  }

  /* If the set is not full, the new peer will go at the end,      */
  /* otherwise it will replace an existing peer (or not be added). */

  if (set->n < csetlim) peer = peers + set->n++;
  else {
    i = rng_uint(router->rng, csetlim);
    if (i == csetlim) return;
    peer = peers + i;
    delete_address(peer->nmc_addr);
  }

  peer->nmc_addr = copy_address(nmc_addr, 3);
  memcpy(peer->label, label, label_size);
  peer->params.rtt = rtt;
  memcpy(peer->params.node, node_params, node_params_size);
  peer->confirmation_time = now;

  /* Update best_peer. */

  if (peer->params.rtt <= router->best_peer->params.rtt) {
    delete_address(router->best_peer->nmc_addr);
    *router->best_peer = *peer;
    router->best_peer->nmc_addr = copy_address(peer->nmc_addr, 3);
  }
}


static Trigger_Handler handle_location_map_trigger;

/* set_location_map_trigger(router) sets the trigger for timing out  */
/* the oldest instance in the location map queue.  This function     */
/* must not be called if the trigger is already set, and must not be */
/* called if the queue is empty.                                     */

static void set_location_map_trigger(struct router *router)
{
  unsigned char *p;
  struct map_resource_instance *instance;
  double now, delay;

  p = (unsigned char *) list_queue_head(router->location_map_queue);
  assert(p != 0);
  p -= offsetof(struct map_resource_instance, q_ptrs);
  instance = (struct map_resource_instance *) p;
  now = scheduler_uptime(router->scheduler);
  router->location_map_trigger_time = now + location_map_timeout;
  delay = router->location_map_trigger_time - now;
  assert(delay >= 0.0);
  router->location_map_trigger = trigger_set( router->scheduler,
    trigger_none, -1, delay, handle_location_map_trigger, router, 0 );
  if (router->location_map_trigger == 0) {
    failf("unable to create location_map_trigger\n");
  }
}


static Trigger_Handler handle_local_resource_trigger;

/* activate_local_resource_trigger(router) sets a trigger */
/* for advertising local resources, if one is not already */
/* set, and if the set of local resources is not empty,   */
/* and if there is a shipping address.  It is always safe */
/* to call this function.                                 */

static void activate_local_resource_trigger(struct router *router)
{
  double delay;

  if ( router->local_resource_trigger != 0 ||
       router->local_resource_queue_length == 0 ||
       router->nmc_size == 0                       ) return;

  delay = quasi_period( router, router->local_resource_period /
                                router->local_resource_queue_length );
  router->local_resource_trigger = trigger_set( router->scheduler,
    trigger_none, -1, delay, handle_local_resource_trigger, router, 0 );
  if (router->local_resource_trigger == 0) {
    failf("unable to create local_resource_trigger\n");
  }
}


static Trigger_Handler handle_flush_trigger;

/* handle_advertisement(router, label, nmc_addr, params, randomize)   */
/* updates the location map and adds an advertisement to the          */
/* appropriate buffer.  If randomize is nonzero, the buffer is chosen */
/* at random, otherwise it is chosen based on the proximity of the    */
/* label to our own label.                                            */

static void handle_advertisement(
  struct router *router, const unsigned char *label,
  const unsigned char *nmc_addr, struct resource_params params,
  int randomize )
{
  int proximity, inhibit_count, total, newer_count;
  unsigned int newer_size, overhead, delta;
  struct ad_buffer *buffer;
  struct trie_leaf *leaf;
  struct map_resource *resource;
  struct list_ptrs *item, *nextitem;
  struct map_resource_instance *new_instance, *instance = 0;
  struct location *newer_locations[max_location_list_length];
  struct advertisement *ad;
  const unsigned char *report_label;

  /* Find or create the resource. */

  leaf = trie_find_item(router->location_map, label);

  if (leaf != 0) resource = (struct map_resource *) leaf;
  else {
    resource = malloc_or_fail(sizeof (struct map_resource));
    memcpy(resource->leaf.label, label, label_size);
    resource->instances = 0;
    trie_insert_item(&router->location_map, &resource->leaf);
  }

  /* Construct the new instance. */

  new_instance = malloc_or_fail(sizeof (struct map_resource_instance));
  new_instance->resource = resource;
  new_instance->location.nmc_addr = copy_address(nmc_addr, 2);
  new_instance->location.params = params;
  new_instance->timeout_start = scheduler_uptime(router->scheduler);

  /* Scan the known instances, checking for a duplicate, counting  */
  /* the ones that inhibit forwarding, checking for old flags that */
  /* need to be set (which causes report messages to be sent), and */
  /* counting the total number of locations.                       */

  inhibit_count = total = 1;  /* the new location counts as one */
  newer_count = 0;
  newer_size = 0;
  report_label = 0 == memcmp(label, router->label, label_size) ? 0 : label;
  overhead = 0
    + 1 + sizeof_address(nmc_addr, 2) + 1
    + label_size
    + 1 /* list length byte */
    + (report_label == 0) /* option byte */
    + message_type_size
    + router->group_id_size + 1;

  for (item = resource->instances;  item != 0;  item = nextitem) {
    nextitem = item->next;
    instance = (struct map_resource_instance *) item;

    if ( instance->location.params.timestamp == params.timestamp &&
         same_address(instance->location.nmc_addr, nmc_addr, 3)    ) {
      /* remove duplicate */
      list_delete_item(&instance->q_ptrs);
      list_delete_item(&instance->ptrs);
      delete_address(instance->location.nmc_addr);
      free(instance);
      instance = 0;
      item = 0;
      continue;
    }

    ++total;

    if ( instance->location.params.timestamp >= params.timestamp &&
         instance->location.params.completeness >= params.completeness ) {
      ++inhibit_count;
    }

    if ( !instance->location.params.old &&
         params.timestamp > instance->location.params.timestamp ) {
      instance->location.params.old = 1;
      assert(newer_count == 0);
      newer_locations[0] = &new_instance->location;
      send_report( router, instance->location.nmc_addr, 0, report_label, 1,
        newer_locations, location_field_size(newer_locations[0]), skew_none );
    }

    if ( !params.old &&
         instance->location.params.timestamp > params.timestamp ) {
      delta = location_field_size(&instance->location);

      if ( newer_count > 0 && ( newer_count == 255 ||
           overhead + newer_size + delta > message_size_soft_limit ) ) {
        send_report( router, nmc_addr, 0, report_label,
                     newer_count, newer_locations, newer_size, skew_none );
        newer_count = 0;
        newer_size = 0;
      }

      newer_locations[newer_count++] = &instance->location;
      newer_size += delta;
    }
  }

  if (newer_count > 0) {
    send_report( router, nmc_addr, 0, report_label,
                 newer_count, newer_locations, newer_size, skew_none );
  }

  /* Add the new location to the map. */

  list_insert_item(&resource->instances, &new_instance->ptrs);
  list_enqueue(router->location_map_queue, &new_instance->q_ptrs);
  if (router->location_map_trigger == 0) set_location_map_trigger(router);

  /* Enforce the limit on the number of instances. */

  if (total > max_locations) {
    assert(instance != 0);
    assert(total == max_locations + 1);
    list_delete_item(&instance->ptrs);
    list_delete_item(&instance->q_ptrs);
  }

  /* If enough locations are known, don't forward the ad. */

  if (total >= enough_locations) return;

  /* Select the buffer. */

  proximity = randomize ? rng_bit_run(router->rng)
                        : label_proximity(label, router->label);
  if (proximity >= label_size * 8) return;
  buffer = router->ad_buffers + proximity;

  /* Create the advertisement. */

  ad = malloc_or_fail(sizeof (struct advertisement));
  memcpy(ad->label, resource->leaf.label, label_size);
  ad->location.nmc_addr = copy_address(nmc_addr, 3);
  ad->location.params = params;

  /* Add the advertisement to the buffer, flushing */
  /* the buffer if/when it gets too full.          */

  delta = label_size + location_field_size(&ad->location);
  overhead = 1 + message_type_size + router->group_id_size + 1;

  if (overhead + buffer->size + delta > message_size_soft_limit) {
    send_post(router,buffer);
  }

  list_insert_item(&buffer->head, &ad->ptrs);
  ++buffer->length;
  buffer->size += delta;

  if (buffer->length == 255 ||
      overhead + buffer->size >= message_size_soft_limit ) {
    send_post(router,buffer);
  }

  /* If the buffer was empty, schedule a flush trigger. */

  if (buffer->length == 1) {
    assert(buffer->flush_trigger == 0);
    buffer->flush_trigger = trigger_set( router->scheduler, trigger_none,
      -1, ad_buffer_age_limit, handle_flush_trigger, router, buffer );
    if (buffer->flush_trigger == 0) failf("unable to create flush_trigger\n");
  }
}


/* lookup_suspend(lookup) suspends the lookup if it is active. */

static void lookup_suspend(struct router_lookup *lookup)
{
  struct router *router = lookup->router;

  if (lookup->query_trigger != 0) {
    trigger_cancel(router->scheduler, lookup->query_trigger);
    lookup->query_trigger = 0;
  }
}


/* lookup_delete(lookup) deletes the lookup. */

static void lookup_delete(struct router_lookup *lookup)
{
  lookup_suspend(lookup);
  list_delete_item(&lookup->ptrs);
  free(lookup);
}


static Trigger_Handler handle_query_trigger;

/* lookup_resume(lookup) originates a query and schedules the next */
/* query, or does nothing if the next query is already scheduled.  */

static void lookup_resume(struct router_lookup *lookup)
{
  struct router *router = lookup->router;
  double delay;

  if (lookup->query_trigger != 0) return;

  /* Originate the query. */
  send_query(router, lookup->label, lookup->fwd_hint);

  /* Determine the next forwarding hint. */

  if (lookup->phase2_countdown >= 0) {
    if (lookup->fwd_hint == fwd_min_time) lookup->fwd_hint = fwd_reduce_time;
    if (lookup->phase2_countdown > 0) --lookup->phase2_countdown;
    else lookup->fwd_hint =
      lookup->fwd_hint == fwd_reduce_hops ? fwd_uniform : fwd_reduce_hops;
  }

  /* Schedule the next query. */

  delay = quasi_period(router, router->query_period);
  lookup->query_trigger = trigger_set( router->scheduler,
    trigger_none, -1, delay, handle_query_trigger, 0, lookup );
  if (lookup->query_trigger == 0) failf("unable to create query_trigger\n");
}


/* lookup_set_state(lookup,state) puts  */
/* the lookup into the specified state. */

static void lookup_set_state(
  struct router_lookup *lookup, enum router_lookup_state state )
{
  switch (state) {
    case router_lookup_active:   lookup_resume(lookup);  break;
    case router_lookup_inactive: lookup_suspend(lookup); break;
    case router_lookup_dead:     lookup_delete(lookup);  break;
    default: failf("invalid router_lookup_state %d\n", state);
  }
}


/* clear_ad_buffer(router,buffer) resets the buffer to empty. */

static void clear_ad_buffer(struct router *router, struct ad_buffer *buffer)
{
  struct list_ptrs *item, *nextitem;
  struct advertisement *ad;

  for (item = buffer->head;  item != 0;  item = nextitem) {
    nextitem = item->next;
    ad = (struct advertisement *) item;
    delete_address(ad->location.nmc_addr);
    free(item);
  }

  buffer->head = 0;
  buffer->length = 0;
  buffer->size = 0;

  if (buffer->flush_trigger != 0) {
    trigger_cancel(router->scheduler, buffer->flush_trigger);
    buffer->flush_trigger = 0;
  }
}


/* remove_local_resource_version(router,version) removes the */
/* local resource version from the collection and frees it.  */

static void remove_local_resource_version(
  struct router *router, struct local_resource_version *version )
{
  struct local_resource *resource = version->resource;
  list_delete_item(&version->q_ptrs);
  list_delete_item(&version->ptrs);
  free(version);

  if (resource->versions == 0) {
    trie_delete_item(&router->local_resources, resource->leaf.label);
    free(resource);
  }

  --router->local_resource_queue_length;
}


/* location_map_queue_shrink(router) removes the oldest resource     */
/* instance (the one at the head of the queue) from the location map */
/* queue, and updates the trie if necessary.  Returns 1 on success,  */
/* or 0 if the queue was empty.                                      */

static int location_map_queue_shrink(struct router *router)
{
  unsigned char *p;
  struct map_resource_instance *instance;
  struct map_resource *resource;

  p = (unsigned char *) list_dequeue(router->location_map_queue);
  if (p == 0) return 0;
  p -= offsetof(struct map_resource_instance, q_ptrs);
  instance = (struct map_resource_instance *) p;
  resource = instance->resource;
  list_delete_item(&instance->ptrs);
  delete_address(instance->location.nmc_addr);
  free(instance);

  if (resource->instances == 0) {
    trie_delete_item(&router->location_map, resource->leaf.label);
    free(resource);
  }

  return 1;
}


/* The next three functions each halt one of three ongoing            */
/* activities: peer discovery, local resource advertisements,         */
/* and location timeouts.  These functions are called only by         */
/* router_delete(), except in simulations, where they are also called */
/* by simtangle_freeze() and have the following additional side       */
/* effects:  After peer discovery is halted, ping & pong messages     */
/* are ignored (which prevents new peers from entering the peer       */
/* sets), and peer state is not timed out (which prevents old peers   */
/* from leaving the peer sets).  After location timeouts are halted   */
/* (which prevents old locations from leaving the map), incoming      */
/* advertisements are ignored (which prevents new locations from      */
/* entering the map, which in turn prevents the timeouts from being   */
/* resumed).  There is currently no way to resume peer discovery or   */
/* location timeouts, and no need to do so is anticipated.  Local     */
/* resource advertisements will automatically resume if the shipping  */
/* address is set or if a new local resource is added, but that       */
/* presumably won't happen in a simulation.                           */

static void halt_peer_discovery(struct router *router)
{
  #if SIMULATION
  /* Purge the old confirmed peers one last time. */
  {
    int level;

    for (level = 0;  level < label_size * 8;  ++level) {
      (void) choose_confirmed_peer(router, level, 0, skew_max_speed);
    }
  }
  #endif

  if (router->historic_trigger != 0) {
    trigger_cancel(router->scheduler, router->historic_trigger);
    router->historic_trigger = 0;
  }

  if (router->ping_trigger != 0) {
    trigger_cancel(router->scheduler, router->ping_trigger);
    router->ping_trigger = 0;
  }
}

static void halt_local_resource_ads(struct router *router)
{
  if (router->local_resource_trigger != 0) {
    trigger_cancel(router->scheduler, router->local_resource_trigger);
    router->local_resource_trigger = 0;
  }
}

static void halt_location_timeouts(struct router *router)
{
  if (router->location_map_trigger != 0) {
    trigger_cancel(router->scheduler, router->location_map_trigger);
    router->location_map_trigger = 0;
  }
}


/** Functions for receiving messages **/


/* A Message_Type_Handler is a function of the form                   */
/* handle(router,msg,size,end).  When the handler is called, the      */
/* group identifier at the end of the message has already been        */
/* verified to match that of the router, and the type has been        */
/* verified to be appropriate for the handler.  The end argument      */
/* points to the message type byte, so (end - msg) is the number of   */
/* bytes that have not yet been processed.  Because these handlers    */
/* are called by a trigger handler, the concurrency rules for trigger */
/* handlers apply (see the file "conventions").                       */

typedef void Message_Type_Handler(
  struct router *, const unsigned char *, size_t, const unsigned char * );


static void handle_ping( struct router *router,
  const unsigned char *msg, size_t size, const unsigned char *end )
{
  const unsigned char *p, *nm_addr, *cookie;

  #if SIMULATION
    ++router->message_tally[simtangle_received][simtangle_msg_peer];
  #endif

  p = msg;
  nm_addr = p;
  p = read_address(2,p,end);
  if (p == 0) return;
  cookie = p;
  p = read_cookie(p,end);
  if (p == 0) return;
  send_pong(router, nm_addr, cookie, 1);
}


static void handle_pong( struct router *router,
  const unsigned char *msg, size_t size, const unsigned char *end )
{
  const unsigned char *p, *nmc_addr, *cookie, *nm_addr, *node_params;
  double time, now, rtt;
  int length, i;

  #if SIMULATION
    ++router->message_tally[simtangle_received][simtangle_msg_peer];
  #endif

  p = msg;

  if (*end == 2) cookie = 0;
  else {
    cookie = p;
    p = read_cookie(p,end);
    if (p == 0) return;
  }

  /* verify my cookie */
  if (end - p < ping_cookie_size) return;
  if (*p++ != ping_cookie_size - 1) return;
  memcpy(&time, p, sizeof (double));
  p += sizeof (double);
  /* TODO: verify hash, before using time, to prevent traps */
  p += hash_size;
  now = scheduler_uptime(router->scheduler);
  rtt = now - time;
  if (rtt < 0.0 || rtt > max_rtt) return;

  nmc_addr = p;
  p = read_address(3,p,end);
  if (p == 0) return;

  length = *p++;

  for (i = 0;  i < length;  ++i) {
    nm_addr = p;
    p = read_address(2,p,end);
    if (p == 0) return;
    add_unconfirmed_peer(router, nm_addr);
  }

  node_params = p;
  if (end - p < node_params_size) return;

  add_confirmed_peer(router, nmc_addr, rtt, node_params, now);
  if (cookie != 0) send_pong(router, nmc_addr, cookie, 0);
}


static void handle_report( struct router *router,
  const unsigned char *msg, size_t size, const unsigned char *end )
{
  const unsigned char *p, *node_addr, *msg_addr, *cookie, *label, *list,
                      *comm_addr, *inbox_addr;
  unsigned char byte;
  int length, found, i;
  struct router_lookup *lookup;
  struct resource_params params;
  struct list_ptrs *item, *nextitem;
  struct trie_leaf *leaf;
  struct local_resource *resource;
  struct local_resource_version *version;
  time_t timestamp = 0;
  Router_Referral_Handler *handler;
  enum router_lookup_state state;
  void *who, *what;

  #if SIMULATION
    ++router->message_tally[simtangle_received][simtangle_msg_lookup];
  #endif

  p = msg;

  if (end - p < 1) return;
  byte = *p++;

  if (byte == 1) {
    /* indirect report */
    #if SIMULATION
      if (router->relay_bug) return;
    #endif
    node_addr = p;
    p = read_address(1,p,end);
    if (p == 0) return;
    msg_addr = p;
    p = read_address(1,p,end);
    if (p == 0) return;
    if (*p != 0) return;
    (void) communicator_send(
      router->communicator, node_addr, msg_addr, p, size - (p - msg) );
    #if SIMULATION
      ++router->message_tally[simtangle_sent][simtangle_msg_lookup];
    #endif
    return;
  }

  if (byte != 0) return;

  /* direct report */

  if (*end == 4) cookie = 0;
  else {
    cookie = p;
    p = read_cookie(p,end);
    if (p == 0) return;
  }

  label = p;
  if (end - p < label_size) return;
  p += label_size;

  if (end - p < 1) return;
  length = *p++;
  list = p;

  if (cookie != 0) {
    /* answer queries */

    for (item = router->lookups;  item != 0;  item = nextitem) {
      nextitem = item->next;
      lookup = (struct router_lookup *) item;
      if (memcmp(lookup->label, label, label_size) != 0) continue;
      handler = lookup->handler;
      who = lookup->who;
      what = lookup->what;
      p = list;

      if (length == 0) {
        state = handler( label, label_size, 0, 0, 0,
                         p == end || (*p & 0x80) == 0, who, what );
        lookup_set_state(lookup,state);
      }
      else {
        for (i = 0;  i < length;  ++i) {
          p = read_location(&comm_addr, &inbox_addr, &params, p, end);
          if (p == 0) break;
          state = handler( label, label_size, comm_addr, inbox_addr,
                           params.timestamp, params.completeness, who, what );
          lookup_set_state(lookup,state);
          if (state == router_lookup_dead) break;
        }
      }
    }

    return;
  }

  /* no cookie, so this is a report of newer versions */

  handler = router->newer_handler;
  if (handler == 0) return;
  who = router->newer_who;
  what = router->newer_what;

  /* find the oldest matching version not flagged as old */

  leaf = trie_find_item(router->local_resources, label);
  if (leaf == 0) return;
  resource = (struct local_resource *) leaf;
  found = 0;  /* flag */

  for (item = resource->versions;  item != 0;  item = item->next) {
    version = (struct local_resource_version *) item;
    if (version->params.old) continue;

    if (!found || version->params.timestamp < timestamp) {
      timestamp = version->params.timestamp;
    }

    found = 1;
  }

  if (!found) return;

  /* notify of locations newer than timestamp */

  p = list;

  for (i = 0;  i < length;  ++i) {
    p = read_location(&comm_addr, &inbox_addr, &params, p, end);
    if (p == 0) break;
    if (params.timestamp <= timestamp) continue;
    (void) handler( label, label_size, comm_addr, inbox_addr,
                    params.timestamp, params.completeness, who, what );
  }
}


static void handle_post( struct router *router,
  const unsigned char *msg, size_t size, const unsigned char *end )
{
  int length, i;
  const unsigned char *p, *label, *nmc_addr, *c_addr;
  struct resource_params params;

  #if SIMULATION
    ++router->message_tally[simtangle_received][simtangle_msg_ad];
  #endif

  p = msg;
  if (end - p < 1) return;
  length = *p++;

  for (i = 0;  i < length;  ++i) {
    label = p;
    if (end - p < label_size) return;
    p += label_size;
    p = read_location(&nmc_addr, &c_addr, &params, p, end);
    if (p == 0) return;
    handle_advertisement(router, label, nmc_addr, params, 0);
  }
}


/* answer_and_relay_query(router,msg,size,end,fromself) is     */
/* like a Message_Type_Handler, with two exceptions.  First,   */
/* it has an additional argument, fromself, which is a boolean */
/* flag indicating whether the query arrived from ourself or   */
/* not (which affects the policy for choosing the next hop).   */
/* Second, it is not necessarily called by a trigger handler,  */
/* it might be called by a public function, and therefore more */
/* strict limitations apply (see the file "conventions").      */

static void answer_and_relay_query( struct router *router,
  const unsigned char *msg, size_t size, const unsigned char *end,
  int fromself )
{
  const unsigned char *p, *nm_addr, *label, *cookie, *report_label;
  enum skew_style skew_style;
  int proximity, count, total;
  struct confirmed_peer *peer;
  struct trie_leaf *leaf;
  struct map_resource *resource;
  struct map_resource_instance *instance;
  struct list_ptrs *item;
  unsigned int locations_size, overhead, delta;
  struct location *locations[max_location_list_length];

  p = msg;
  nm_addr = p;
  p = read_address(2,p,end);
  if (p == 0) return;
  label = p;
  if (end - p < label_size) return;
  p += label_size;
  cookie = p;
  p = read_cookie(p,end);
  if (p == 0) return;
  skew_style = router->skew_styles[(*p & 0xFF) >> 6];

  /* Choose a closer peer. */

  proximity = label_proximity(label, router->label);
  peer = choose_confirmed_peer(router, proximity, label, skew_style);

  /* Report known locations to requestor, using multiple */
  /* reports if the messages would get too big.          */

  leaf = trie_find_item(router->location_map, label);
  count = 0;           /* length of locations[] */
  locations_size = 0;  /* message bytes needed for locations[] */
  total = 0;           /* total known locations */
  report_label = 0 == memcmp(label, router->label, label_size) ? 0 : label;

  if (leaf != 0) {
    resource = (struct map_resource *) leaf;
    overhead = 0
      + 1 + sizeof_address(nm_addr, 2) + 1
      + label_size
      + 1 + cookie[0]
      + 1 /* list length byte */
      + (report_label == 0) /* option byte */
      + message_type_size
      + router->group_id_size + 1;

    for (item = resource->instances;  item != 0;  item = item->next) {
      ++total;
      instance = (struct map_resource_instance *) item;
      delta = location_field_size(&instance->location);

      if ( count > 0 && ( count == 255 ||
           overhead + locations_size + delta > message_size_soft_limit ) ) {
        send_report( router, nm_addr, cookie, report_label,
                     count, locations, locations_size, skew_style );
        count = 0;
        locations_size = 0;
      }

      locations[count++] = &instance->location;
      locations_size += delta;
    }
  }

  /* If there are locations remaining to be reported, or if we  */
  /* are the closest node and no locations are known, then send */
  /* a report.                                                  */

  if (count > 0 || (total == 0 && peer == 0)) {
    send_report( router, nm_addr, cookie, report_label,
                 count, locations, locations_size, skew_style );
  }

  /* Decide whether to forward the query, and to whom. */

  if (fromself) {
    if (skew_style != skew_max_speed) {
      peer = choose_confirmed_set_and_peer(router, 0, skew_style);
    }
  }
  else {
    #if SIMULATION
      if (router->relay_bug) peer = 0;
    #endif
    if (total >= enough_locations) peer = 0;
  }

  /* Forward the query if appropriate. */

  if (peer != 0) {
    const unsigned char *node_addr = peer->nmc_addr;
    const unsigned char *msg_addr = node_addr + 1 + node_addr[0];
    (void) communicator_send(
      router->communicator, node_addr, msg_addr, msg, size );
    #if SIMULATION
      ++router->message_tally[simtangle_sent][simtangle_msg_lookup];
    #endif
  }
}


static void handle_query( struct router *router,
  const unsigned char *msg, size_t size, const unsigned char *end )
{
  #if SIMULATION
    ++router->message_tally[simtangle_received][simtangle_msg_lookup];
  #endif

  answer_and_relay_query(router,msg,size,end,0);
}


/** Functions for handling triggers **/


/* handle_message_trigger(router,unused)        */
/* handles triggers caused by message arrivals. */

static void handle_message_trigger(void *who, void *what)
{
  struct router *router = who;
  const unsigned char *msg, *p;
  size_t size;
  unsigned int group_id_size, type;
  enum { max_message_type = 7 };
  Message_Type_Handler *handle;
  Message_Type_Handler * const dispatch[max_message_type + 1] = {
    /* type 0 */  0,
    /* type 1 */  handle_ping,
    /* type 2 */  handle_pong,
    /* type 3 */  handle_pong,
    /* type 4 */  handle_report,
    /* type 5 */  handle_report,
    /* type 6 */  handle_post,
    /* type 7 */  handle_query,
  };

  assert(router->guard == 0);
  router->guard = 1;

  router->message_trigger = 0;
  msg = router->message;
  size = router->message_size;
  assert(msg != 0);

  /* Check the group identifier. */

  p = msg + size;
  if (p - msg < 1) goto handle_message_trigger_end;
  group_id_size = *--p;
  if (group_id_size != router->group_id_size) goto handle_message_trigger_end;
  if (p - msg < 1 + group_id_size) goto handle_message_trigger_end;
  p -= group_id_size;

  if (0 != memcmp(p, router->group_id, group_id_size)) {
    goto handle_message_trigger_end;
  }

  /* Dispatch based on the type. */

  type = *--p;
  if (type > max_message_type) goto handle_message_trigger_end;
  handle = dispatch[type];

  #if SIMULATION
    /* side effects of halt_*() */
    if ( (type <= 3 && router->ping_trigger == 0)  ||
         (type == 6 && router->location_map_trigger == 0) ) handle = 0;
  #endif

  if (handle != 0) handle(router,msg,size,p);

handle_message_trigger_end:

  /* Delete the message. */

  message_inbox_clear(router->inbox);
  router->message = 0;

  router->guard = 0;
}


/* handle_query_trigger(unused,lookup) handles the */
/* trigger that evokes an outgoing query message.  */

static void handle_query_trigger(void *who, void *what)
{
  struct router_lookup *lookup = what;
  struct router *router = lookup->router;

  assert(router->guard == 0);
  router->guard = 1;

  lookup->query_trigger = 0;
  lookup_resume(lookup);

  router->guard = 0;
}


/* handle_ping_trigger(router,unused) handles the */
/* trigger that evokes an outgoing ping message.  */

static void handle_ping_trigger(void *who, void *what)
{
  struct router *router = who;
  struct unconfirmed_set **upeers = router->unconfirmed_peers;
  int level, i, set_size, n;
  const unsigned char *nm_addr;
  double delay;

  assert(router->guard == 0);
  router->guard = 1;

  router->ping_trigger = 0;
  n = router->depth;
  if (n == 0 && router->extra_deep > 0) n = 1;

  if (rng_bit_run(router->rng) >= n) {
    /* Use a historic peer. */
    level = -1;  /* signal that we didn't use an unconfirmed peer */
    if (router->historic_n == 0) goto handle_ping_trigger_end;
    i = rng_uint(router->rng, router->historic_n - 1);
    nm_addr = router->historic_peers[i];
  }
  else {
    /* Use an unconfirmed peer.  Notice that we cannot get here */
    /* if there are no unconfirmed peers, because n will be 0.  */

    /* Choose a level.  Levels are handled in a round-robin fashion. */

    level = router->last_ping_level + 1;
    if (level >= router->depth) level = 0;
    /* Usually level < depth, but we can also have level == depth == 0. */

    /* Choose a peer uniformly at random from the set. */

    set_size = upeers[level] == 0 ? 0 : upeers[level]->n;
    if (level >= router->depth - 1) set_size += router->extra_deep;
    assert(set_size > 0);
    i = rng_uint(router->rng, set_size - 1);

    while (upeers[level] == 0 || i >= upeers[level]->n) {
      if (upeers[level] != 0) i -= upeers[level]->n;
      ++level;
      assert(level < label_size * 8);
    }

    nm_addr = upeers[level]->peers[i];
  }

  /* Send the ping. */

  send_ping(router, nm_addr);
  router->last_ping_level = level;

  /* Remove the unconfirmed peer from the set. */

  if (level >= 0) remove_unconfirmed_peer(router,level,i);

  /* Schedule the next ping. */

handle_ping_trigger_end:
  delay = quasi_period(router, router->ping_period);
  router->ping_trigger = trigger_set( router->scheduler,
    trigger_none, -1, delay, handle_ping_trigger, router, 0 );
  if (router->ping_trigger == 0) failf("unable to create ping_trigger\n");

  router->guard = 0;
}


/* handle_historic_trigger(router,unused) handles the trigger */
/* that causes a peer to be added to the historic list.       */

static void handle_historic_trigger(void *who, void *what)
{
  struct router *router = who;
  struct confirmed_peer *peer;
  int i, j;
  double delay;

  assert(router->guard == 0);
  router->guard = 1;

  router->historic_trigger = 0;
  peer = choose_confirmed_set_and_peer(router, 0, skew_none);
  if (peer == 0) goto handle_historic_trigger_end;

  if (router->historic_n < router->historic_threshold) {
    ++router->historic_n;
  }
  else {
    /* choose one to remove */
    i = rng_bit_run(router->rng);

    if (i >= router->historic_n) {
      /* don't remove any, grow the list */
      router->historic_peers = realloc( router->historic_peers,
        ++router->historic_n * sizeof (const unsigned char *)   );
    }
    else {
      j = router->historic_n - 1 - i;  /* count backwards from the end */
      delete_address(router->historic_peers[j]);
      memmove(router->historic_peers + j,
              router->historic_peers + j + 1,
              i * sizeof (const unsigned char *) );
    }
  }

  router->historic_peers[router->historic_n - 1] =
    copy_address(peer->nmc_addr, 2);

  /* TODO: write the list to stable storage */

  /* Schedule the next addition. */

handle_historic_trigger_end:
  delay = quasi_period(router, router->historic_period);
  router->historic_trigger = trigger_set( router->scheduler,
    trigger_none, -1, delay, handle_historic_trigger, router, 0 );
  if (router->historic_trigger == 0) {
    failf("unable to create historic_trigger\n");
  }

  router->guard = 0;
}


/* handle_flush_trigger(router,buffer) handles the   */
/* trigger that evokes the flushing of an ad buffer. */

static void handle_flush_trigger(void *who, void *what)
{
  struct router *router = who;
  struct ad_buffer *buffer = what;

  assert(router->guard == 0);
  router->guard = 1;

  buffer->flush_trigger = 0;
  send_post(router,buffer);

  router->guard = 0;
}


/* handle_location_map_trigger(router,unused) handles the trigger    */
/* that signifies the timeout of the oldest location in the location */
/* map (the one at the head of the queue).                           */

static void handle_location_map_trigger(void *who, void *what)
{
  struct router *router = who;
  unsigned char *p;
  struct map_resource_instance *instance;
  double now, expiration_time;

  assert(router->guard == 0);
  router->guard = 1;

  router->location_map_trigger = 0;
  now = scheduler_uptime(router->scheduler);

  /* If the trigger comes early, we want to go ahead and expire       */
  /* anything that would have been expired if the trigger had come on */
  /* time, rather than risk taking many tiny steps toward the target. */
  /* If the trigger comes late, we need to expire everything that     */
  /* is now too old, so that the next trigger won't want to have a    */
  /* negative delay.  The number of items to be expired could even be */
  /* zero, if the intended item was already deleted for some other    */
  /* reason.                                                          */

  for (;;) {
    p = (unsigned char *) list_queue_head(router->location_map_queue);
    if (p == 0) break;
    p -= offsetof(struct map_resource_instance, q_ptrs);
    instance = (struct map_resource_instance *) p;
    expiration_time = instance->timeout_start + location_map_timeout;
    if ( expiration_time >= now &&
         expiration_time >= router->location_map_trigger_time ) break;
    (void) location_map_queue_shrink(router);
  }

  if (! list_queue_is_empty(router->location_map_queue)) {
    set_location_map_trigger(router);
  }

  router->guard = 0;
}


/* handle_local_resource_trigger(router,unused) handles the       */
/* trigger that causes a local resource version to be advertised. */

static void handle_local_resource_trigger(void *who, void *what)
{
  struct router *router = who;
  struct list_ptrs *item;
  struct local_resource_version *version;
  unsigned char *p;

  assert(router->guard == 0);
  router->guard = 1;

  router->local_resource_trigger = 0;

  if ( router->nmc_size == 0 ||
       router->local_resource_queue_length == 0 ) {
    goto handle_local_resource_trigger_end;
  }

  /* Move the item at the head of the circular queue to the tail. */

  item = list_dequeue(router->local_resource_queue);
  list_enqueue(router->local_resource_queue, item);

  /* Advertise the item. */

  p = (unsigned char *) item;
  p -= offsetof(struct local_resource_version, q_ptrs);
  version = (struct local_resource_version *) p;
  handle_advertisement( router,
    version->resource->leaf.label, router->addresses, version->params, 1 );

  /* Schedule the next advertisement. */

  activate_local_resource_trigger(router);

handle_local_resource_trigger_end:
  router->guard = 0;
}


/** Callback functions **/


/* handle_message(msg,size,router,unused) handles */
/* the entry of a message into the inbox.         */

enum message_action handle_message(
  const unsigned char *msg, size_t size, void *who, void *what )
{
  struct router *router = who;

  assert(router->guard == 0);
  router->guard = 1;

  assert(router->message_trigger == 0);
  router->message = msg;
  router->message_size = size;
  router->message_trigger = trigger_set( router->scheduler,
    trigger_none, -1, 0.0, handle_message_trigger, router, 0 );
  if (router->message_trigger == 0) {
    failf("unable to create message trigger\n");
  }

  router->guard = 0;
  return message_keep;
}


/** Backdoor functions **

see simrouter.h and simtangle.h for interface documentation

*/


void simtangle_set_lookup_style(struct router *router, int lookup_style)
{
  assert(router->guard == 0);
  /* nothing is called, no need to set guard */

  if (lookup_style == simtangle_lookup_default) {
    lookup_style = phase2_countdown_init;
  }

  if (lookup_style >= 0) {
    router->phase2_countdown_init = lookup_style;
    router->fwd_hint_init = fwd_min_time;
    return;
  }

  router->phase2_countdown_init = -1;
  router->fwd_hint_init =
    lookup_style == simtangle_lookup_reduce_time ? fwd_reduce_time :
    lookup_style == simtangle_lookup_reduce_hops ? fwd_reduce_hops :
    lookup_style == simtangle_lookup_uniform     ? fwd_uniform     :
    lookup_style == simtangle_lookup_min_time    ? fwd_min_time    : -1 ;
  assert(router->fwd_hint_init >= 0);
}


void simtangle_freeze(struct router *router)
{
  assert(router->guard == 0);
  router->guard = 1;

  halt_peer_discovery(router);
  halt_local_resource_ads(router);
  halt_location_timeouts(router);

  router->guard = 0;
}


unsigned long simtangle_get_msg_tally(
  struct router *router,
  enum simtangle_msg_direction direction, enum simtangle_msg_class class )
{
  assert(router->guard == 0);
  /* nothing is called, no need to set guard */
  assert(direction == simtangle_sent || direction == simtangle_received);
  assert(class >= 0 && class < simtangle_msg_class_n);
  return router->message_tally[direction][class];
}


/** Public functions **

See router.h for the interface documentation.

*/


struct router *router_new(
  struct scheduler *scheduler,
  struct configurator *configurator,
  struct communicator *communicator,
  const unsigned char *group_id, unsigned int group_id_size )
{
  struct router *router = 0;
  int i;
  const unsigned char *node_address, *message_address;
  unsigned char buffer[256 * 2], *p;
  unsigned int n_size, m_size;

  router = malloc_or_fail(sizeof (struct router));
  router->guard = 1;

  router->rng = rng_new();

  if (!router->rng) {
    free(router);
    return 0;
  }

  router->scheduler = scheduler;
  router->configurator = configurator;
  router->communicator = communicator;
  router->group_id_size = group_id_size;

  if (group_id_size == 0) router->group_id = 0;
  else {
    router->group_id = malloc_or_fail(group_id_size);
    memcpy(router->group_id, group_id, group_id_size);
  }

  /* TODO: Recall the last address we used.    */
  /* For now, we just use a new one each time. */

  router->inbox = message_inbox_new(communicator,0);
  if (!router->inbox) failf("unable to create message inbox\n");
  router->message = 0;
  router->message_size = 0;
  router->message_trigger = 0;

  node_address = communicator_get_address(communicator);
  message_address = message_inbox_get_address(router->inbox);

  /* We copy them twice, first into a temporary buffer, */
  /* so that we can use copy_address() the second time, */
  /* in case it plays any tricks.                       */

  p = buffer;
  n_size = 1 + node_address[0];
  memcpy(p, node_address, n_size);
  p += n_size;
  m_size = 1 + message_address[0];
  memcpy(p, message_address, m_size);
  router->addresses = copy_address(buffer,2);
  router->nm_size = n_size + m_size;
  router->nmc_size = 0;

  router->birthtime = scheduler_uptime(scheduler);
  memset(router->params, 0, node_params_size - 2);
  compute_label(router->label, router->addresses);

  /* TODO: Use Configurator for these three: */
  router->unconfirmed_set_limit = unconfirmed_set_limit;
  router->confirmed_set_limit = confirmed_set_limit;
  router->historic_threshold = historic_threshold;

  #if !SIMULATION
    /* TODO: load historic peers from a file */
    failf("loading historic peers is not yet implemented\n");
  #else /* SIMULATION */

    i = router->historic_n = simrouter_initial_peers_n;
    if (i < router->historic_threshold) i = router->historic_threshold;
    router->historic_peers =
      malloc_or_fail(i * sizeof (const unsigned char *));

    for (i = 0;  i < router->historic_n;  ++i) {
      router->historic_peers[i] =
        sim_nodes[ simrouter_initial_peers[i] ].nmc_addr;
    }

  #endif /* SIMULATION */

  for (i = 0;  i < label_size * 8;  ++i) {
    router->unconfirmed_peers[i] = 0;
    router->confirmed_peers[i] = 0;
    router->ad_buffers[i].head = 0;
    router->ad_buffers[i].length = 0;
    router->ad_buffers[i].size = 0;
    router->ad_buffers[i].flush_trigger = 0;
  }

  router->location_map = 0;
  list_init_queue(&router->location_map_queue);
  router->location_map_trigger = 0;

  router->local_resources = 0;
  list_init_queue(&router->local_resource_queue);
  router->local_resource_queue_length = 0;
  router->local_resource_trigger = 0;

  router->lookups = 0;

  router->newer_handler = 0;
  router->newer_who = 0;
  router->newer_what = 0;

  router->ping_trigger = trigger_set( router->scheduler,
    trigger_none, -1, 0.0, handle_ping_trigger, router, 0 );
  if (router->ping_trigger == 0) {
    failf("unable to create ping_trigger\n");
  }

  router->historic_trigger = trigger_set( router->scheduler,
    trigger_none, -1, 0.0, handle_historic_trigger, router, 0 );
  if (router->historic_trigger == 0) {
    failf("unable to create historic_trigger\n");
  }

  router->depth = 0;
  router->extra_deep = 0;
  router->last_ping_level = -1;

  /* TODO: Use Configurator for these five: */
  router->ping_period_0 = ping_period_0;
  router->ping_period = router->ping_period_0;
  router->historic_period = historic_period;
  router->local_resource_period = local_resource_period;
  router->query_period =
    #if SIMULATION
      simtangle_query_period;
    #else
      query_period;
    #endif
  router->confirmed_peer_age_limit_0 = confirmed_peer_age_factor *
    router->ping_period_0 * router->confirmed_set_limit;
  router->confirmed_peer_age_limit = router->confirmed_peer_age_limit_0;

  /* Outgoing ping messages are all the same size, so precompute it. */

  router->ping_size = 0
    + router->nm_size
    + ping_cookie_size
    + message_type_size
    + router->group_id_size + 1
    ;

  /* Outgoing query messages are all the same size, so precompute it. */

  router->query_size = 0
    + router->nm_size
    + label_size
    + query_cookie_size
    + message_type_size
    + 1 /* option */
    + router->group_id_size + 1
    ;

  #if SIMULATION
    router->msg_buf = shared_message_buffer;
  #else
    router->msg_buf = malloc_or_fail(message_size_soft_limit);
  #endif
  router->msg_buf_size = message_size_soft_limit;

  /* TODO: Use Configurator for these seven: */
  router->pong_list_length = pong_list_length;
  router->fwd_hint_init = fwd_hint_init;
  router->phase2_countdown_init = phase2_countdown_init;
  router->skew_styles[fwd_reduce_time] = skew_sqrt_speed;
  router->skew_styles[fwd_reduce_hops] = skew_sqrt_bits;
  router->skew_styles[fwd_uniform] = skew_none;
  router->skew_styles[fwd_min_time] = skew_max_speed;

  router->best_peer->nmc_addr = copy_address(router->addresses, 1);
    /* Ensures that best_peer->nmc_addr can be passed to delete_address(). */
  router->best_peer->params.rtt = HUGE_VAL;
    /* Ensures that the first confirmed peer is better. */
  router->best_peer->confirmation_time = -HUGE_VAL;
    /* Ensures that the initial garbage best_peer will not be used. */

  #if SIMULATION
  {
    int proximity, max;

    for (i = 0;  i < simtangle_msg_class_n;  ++i) {
      router->message_tally[0][i] = router->message_tally[1][i] = 0;
    }

    for (max = i = 0;  i < sim_nodes_n;  ++i) {
      if (same_address(sim_nodes[i].nmc_addr, router->addresses, 1)) continue;
      proximity = label_proximity(sim_nodes[i].label, router->label);
      if (proximity > max) max = proximity;
    }

    router->level_target = max;
    router->relay_bug = simtangle_relay_bug != 0;
  }
  #endif /* SIMULATION */

  router->msg_lock = 0;
  message_inbox_set_handler(router->inbox, handle_message, router, 0);

  router->guard = 0;
  return router;
}


void router_set_shipping_address(
  struct router *router, const unsigned char *address)
{
  unsigned char buffer[256 * 3], *p = buffer;
  unsigned int nm_size = router->nm_size, c_size;

  assert(router->guard == 0);
  router->guard = 1;

  assert(address != 0);
  assert(router->nmc_size == 0);

  /* We copy twice, first into a temporary buffer, so that we can use */
  /* copy_address() the second time, in case it plays any tricks.     */

  c_size = 1 + address[0];
  memcpy(p, router->addresses, nm_size);
  p += nm_size;
  memcpy(p, address, c_size);
  delete_address(router->addresses);
  router->addresses = copy_address(buffer,3);
  router->nmc_size = nm_size + c_size;

  activate_local_resource_trigger(router);

  router->guard = 0;
}


void router_delete(struct router *router)
{
  int i, j;
  unsigned char *p;
  struct local_resource_version *version;
  struct confirmed_set *cset;
  struct unconfirmed_set *uset;

  assert(router->guard == 0);
  router->guard = 1;

  #if !SIMULATION
    free(router->msg_buf);
    router->msg_buf = 0;
  #endif

  delete_address(router->best_peer->nmc_addr);

  halt_peer_discovery(router);

  while (router->lookups != 0) {
    lookup_delete((struct router_lookup *) router->lookups);
  }

  halt_local_resource_ads(router);

  for (;;) {
    p = (unsigned char *) list_queue_head(router->local_resource_queue);
    if (p == 0) break;
    p -= offsetof(struct local_resource_version, q_ptrs);
    version = (struct local_resource_version *) p;
    remove_local_resource_version(router,version);
  }

  assert(router->local_resource_queue_length == 0);

  halt_location_timeouts(router);
  while (location_map_queue_shrink(router));

  for (i = 0;  i < label_size * 8;  ++i) {
    clear_ad_buffer(router, router->ad_buffers + i);
    cset = router->confirmed_peers[i];
    uset = router->unconfirmed_peers[i];

    if (cset != 0) {
      for (j = 0; j < cset->n; ++j) delete_address(cset->peers[j].nmc_addr);
      free(cset);
      router->confirmed_peers[i] = 0;
    }

    if (uset != 0) {
      for (j = 0; j < uset->n; ++j) delete_address(uset->peers[j]);
      free(uset);
      router->unconfirmed_peers[i] = 0;
    }
  }

  free(router->historic_peers);
  router->historic_peers = 0;

  delete_address(router->addresses);
  router->addresses = 0;

  if (router->message_trigger != 0) {
    trigger_cancel(router->scheduler, router->message_trigger);
    router->message_trigger = 0;
  }

  message_inbox_delete(router->inbox);
  router->inbox = 0;
  router->message = 0;

  if (router->group_id != 0) {
    free(router->group_id);
    router->group_id = 0;
  }

  rng_delete(router->rng);
  router->rng = 0;

  free(router);
}


struct router_lookup *router_lookup_new(
  struct router *router, const unsigned char *label, unsigned int size,
  Router_Referral_Handler *handler, void *who, void *what )
{
  struct router_lookup *lookup;

  if (size != label_size) return 0;

  assert(router->guard == 0);
  router->guard = 1;

  /* Create the lookup structure. */

  lookup = malloc_or_fail(sizeof (struct router_lookup));
  lookup->router = router;
  memcpy(lookup->label, label, label_size);
  lookup->handler = handler;
  lookup->who = who;
  lookup->what = what;
  lookup->query_trigger = 0;
  lookup->fwd_hint = router->fwd_hint_init;
  lookup->phase2_countdown = router->phase2_countdown_init;

  /* Add the lookup to the list. */
  list_insert_item(&router->lookups, &lookup->ptrs);

  /* Start the queries. */
  lookup_resume(lookup);

  router->guard = 0;
  return lookup;
}


void router_lookup_delete(struct router_lookup *lookup)
{
  struct router *router = lookup->router;
  assert(router->guard == 0);
  router->guard = 1;
  lookup_delete(lookup);
  router->guard = 0;
}


void router_lookup_suspend(struct router_lookup *lookup)
{
  struct router *router = lookup->router;
  assert(router->guard == 0);
  router->guard = 1;
  lookup_suspend(lookup);
  router->guard = 0;
}


void router_lookup_resume(struct router_lookup *lookup)
{
  struct router *router = lookup->router;
  assert(router->guard == 0);
  router->guard = 1;
  lookup_resume(lookup);
  router->guard = 0;
}


void router_set_bandwidth(struct router *router, double bandwidth) {
  unsigned char *p = router->params;

  assert(router->guard == 0);
  router->guard = 1;

  usf_encode(p,bandwidth);

  router->guard = 0;
}


void router_set_space(struct router *router, double space) {
  unsigned char *p = router->params + 2;

  assert(router->guard == 0);
  router->guard = 1;

  usf_encode(p,space);

  router->guard = 0;
}


void router_advertise(
  struct router *router, const unsigned char *label, unsigned int size,
  time_t timestamp, unsigned char completeness, unsigned char old       )
{
  struct local_resource *resource;
  struct local_resource_version *version;

  assert(old <= 1);

  if (size != label_size) {
    warnf("router_advertise() called with invalid label size\n");
    return;
  }

  assert(router->guard == 0);
  router->guard = 1;

  version = find_local_resource_version(router,label,timestamp);

  /* If the <label,timestamp> pair already exists, just update the   */
  /* old flag and completeness, and move it to the head of the local */
  /* resource queue (so that it will be advertised soon).            */

  if (version != 0) {
    resource = version->resource;
    version->params.completeness = completeness;
    version->params.old = old;
    list_delete_item(&version->q_ptrs);
    list_insert_item(&router->local_resource_queue.next, &version->q_ptrs);
    goto router_advertise_end;
  }

  /* The resource doesn't already exist, so create it. */

  resource = malloc_or_fail(sizeof (struct local_resource));
  memcpy(resource->leaf.label, label, label_size);
  resource->versions = 0;
  trie_insert_item(&router->local_resources, &resource->leaf);

  /* Add the new version. */

  version = malloc_or_fail(sizeof (struct local_resource_version));
  version->resource = resource;
  version->params.timestamp = timestamp;
  version->params.completeness = completeness;
  version->params.old = old;
  list_insert_item(&resource->versions, &version->ptrs);

  /* Advertise the new version, then insert it */
  /* at the tail of the local resource queue.  */

  if (router->nmc_size != 0) handle_advertisement( router,
    version->resource->leaf.label, router->addresses, version->params, 1 );
  list_enqueue(router->local_resource_queue, &version->q_ptrs);
  ++router->local_resource_queue_length;
  activate_local_resource_trigger(router);

router_advertise_end:
  router->guard = 0;
}


void router_unadvertise(
  struct router *router,
  const unsigned char *label, unsigned int size, time_t timestamp )
{
  struct local_resource_version *version;

  if (size != label_size) {
    warnf("router_unadvertise() called with invalid label size\n");
    return;
  }

  assert(router->guard == 0);
  router->guard = 1;

  version = find_local_resource_version(router,label,timestamp);
  if (version != 0) remove_local_resource_version(router,version);

  router->guard = 0;
}


void router_report_newer( struct router *router,
   Router_Referral_Handler *handler, void *who, void *what )
{
  assert(router->guard == 0);
  router->guard = 1;

  router->newer_handler = handler;
  router->newer_who = who;
  router->newer_what = what;

  router->guard = 0;
}


struct router_peer *router_get_peer(struct router *router)
{
  assert(router->guard == 0);
  router->guard = 1;

  /* TODO */
  failf("router_peer() is not yet implemented\n");
  return 0;  /* quiet warnings */
}


void router_peer_delete(struct router *router, struct router_peer *peer)
{
  assert(router->guard == 0);
  router->guard = 1;

  delete_address(peer->nmc_addr);
  free(peer);

  router->guard = 0;
}


const unsigned char *router_peer_communicator_address(
  struct router *router, struct router_peer *peer )
{
  assert(router->guard == 0);
  /* nothing is called, no need to set guard */
  return peer->nmc_addr;
}


const unsigned char *router_peer_shipping_inbox_address(
  struct router *router, struct router_peer *peer )
{
  const unsigned char *p;

  assert(router->guard == 0);
  /* nothing is called, no need to set guard */
  p = peer->nmc_addr;
  p += 1 + *p;
  p += 1 + *p;
  return p;
}


double router_peer_rtt(struct router *router, struct router_peer *peer)
{
  assert(router->guard == 0);
  /* nothing is called, no need to set guard */
  return peer->params.rtt;
}


double router_peer_uptime(struct router *router, struct router_peer *peer)
{
  double uptime;

  assert(router->guard == 0);
  router->guard = 1;

  uptime = usf_decode(peer->params.node + 0);

  router->guard = 0;
  return uptime;
}


double router_peer_bandwidth(struct router *router, struct router_peer *peer)
{
  double bandwidth;

  assert(router->guard == 0);
  router->guard = 1;

  bandwidth = usf_decode(peer->params.node + 2);

  router->guard = 0;
  return bandwidth;
}


double router_peer_space(struct router *router, struct router_peer *peer)
{
  double space;

  assert(router->guard == 0);
  router->guard = 1;

  space = usf_decode(peer->params.node + 4);

  router->guard = 0;
  return space;
}
